#ifndef CARTRIDGE_H
#define CARTRIDGE_H

enum { CARTRIDGE_LOAD_PUT=0, CARTRIDGE_LOAD_GET };

#define CARTRIDGE_MAXPRG				0x200
#define CARTRIDGE_MAXVROM				0x800
#define CARTRIDGE_MAXVRAM				0x20
#define CARTRIDGE_MAXSIZE				(CARTRIDGE_MAXPRG*0x2000+CARTRIDGE_MAXVROM*0x400)

#define CARTRIDGE_SKIP_TRAINER			FALSE	/* don't apply optional trainer */

#define CARTRIDGE_MIRRORING_HORIZONTAL	BIN8(00001100)
#define CARTRIDGE_MIRRORING_VERTICAL	BIN8(00001010)
#define CARTRIDGE_MIRRORING_ONE_0		0
#define CARTRIDGE_MIRRORING_ONE_1		BIN8(00001111)
#define CARTRIDGE_MIRRORING_4			BIT(7)
#define CARTRIDGE_MIRRORING_CUSTOM		BIT(6)

enum { CARTRIDGE_REGION_USA=0, CARTRIDGE_REGION_JAPAN, CARTRIDGE_REGION_EUROPE };

typedef struct {
	u8* cur_prgbank[4];
	int cur_prgbank_isram[4];
	int cur_prgbank_num[4];
	u8* cur_patbank[8];
	int cur_patbank_isram[8];
	int cur_patbank_num[8];
	
	u8* wram;
	u8* cur_wrambank;
	u8 cur_wrambank_isprg;
	int cur_wrambank_num;
	u8 wram_isenabled;
	u8 wram_writeenable;
	
	u32 prgbanks;
	u32 prgbanks_real;
	u32 vrombanks;
	u32 vrombanks_real;
	u32 mirroring;
	u8 type;
	u8 mode;
	u8 region;
	u8 trainer;

	u8 battery;
	u32 battery_size;
	u8* battery_data;
	
	u8* prg[CARTRIDGE_MAXPRG];
	u8* vrom[CARTRIDGE_MAXVROM];
	u8* vram[CARTRIDGE_MAXVRAM];
	
	u8* dummy_rom;
	u8* dummy_vrom;
	
	u32 crc32;
	u32 crc32_patch;
	char name[STRING_SIZE];
	char filename[STRING_SIZE];
	
} Cartridge;

Cartridge* cartridge;
Cartridge* cartridge_temp;


void cartridge_battery_load(void);
void cartridge_battery_save(int);

int cartridge_load(int,int*,int*,int*,u32*,u32*);
void cartridge_unload(Cartridge*);
void cartridge_init(void);
void cartridge_clean(void);
void cartridge_set_cartridge(void);

#endif /* CARTRIDGE_H */
