#include "global.h"
#include "cont.h"
#include "input.h"
#include "settings.h"
#include "draw.h"

/*
Unemulated devices:
	Arkanoid (dial): Arkanoid, Arkanoid 2(?), Chase HQ(J)
	Konami Laser Scope (some kind of headset with microphone and gun): Laser Invasion/Gun Sight (same as zapper?)
	piano device: Miracle Piano Teaching System
	Power Glove: Super Glove Ball
	R.O.B./Family Robot: Gyromite, Stack-Up
	U-Force: ?
	Zapper/Light Gun: Duck Hunt, etc.

Famicom-only (most on extended port)
	3D system (glasses with 2 screens, alternating frames. press [select] in games to enable 3d): Attack Animal Gakuen, Cosmic Epsilon, Highway Star, JJ
	Battle Box (saves SRAM data): Armadillo, Battle Stadium - Senbatsu Pro Yakyuu
	data recorder (attached to keyboard, for save/load user data): Banana, Excitebike(J), Family Basic, Mach Rider
	Datach (barcode reader): ?
	Exciting Boxing: Exciting Boxing
	Hyper Shot (not the Konami one, this time a gun): Space Shadow?
	Karaoke Studio (microphone too): Karaoke Studio, Karaoke Studio Senyou Cassette Vol 1,2
	keyboard: Family Basic (Subor: different keyboard)
	Mahjong: ?
	modem: ?
	Pachinko: Pachinko Daisakusen, Pachiokun 4, Pachiokun 5
	Party Tap: Casino Derby, Gimmi a Break - Shijou Saikyou no Quiz Ou Ketteisen 1,2, Project Q, 
	piano device (another?): ?(Konami disk game)
	Super Mogura Tataki!! - Pokkun Moguraa: Super Mogura Tataki!! - Pokkun Moguraa
	tablet: Oeka Kids
	Top Rider: Top Rider
	trackball: Moero!! Pro Soccer
	Turbo File: ?
*/

static void cont_zero_structs(void);

/* controller names */
static const char* cont_name[CONT_D_MAX]={
	"NES joypad",
	"Famicom joypad",
	"Power Pad",
	"Family Trainer",
	"HyperShot",
	"Four Score",
	"4 Players Adaptor",
	"Family Basic keyboard",
	
	"nothing"
};

static const char* cont_a_name[CONT_DA_MAX]={
	"Famicom microphone",
	"joypad (port 1)",
	"joypad (port 2)",
	"joypad (port 3)",
	"joypad (port 4)",
	"Power Pad/Family Trainer",
	"HyperShot",
	"Family Basic keyboard"
};

/* ports */
static const int cont_port_default[CONT_P_MAX]={CONT_D_NJ,CONT_D_NJ,CONT_D_NOTHING};
static int cont_port[CONT_P_MAX];

/* trigger info, same numbers as in cont.h */
static const char* cont_t_info[CONT_T_MAX]={
	/* joypad 1 */
	"Famicom microphone",
	
	"joypad 1 A",
	"joypad 1 B",
	"joypad 1 Select",
	"joypad 1 Start",
	"joypad 1 Up",
	"joypad 1 Down",
	"joypad 1 Left",
	"joypad 1 Right",
	
	/* joypad 2 */
	"joypad 2 A",
	"joypad 2 B",
	"joypad 2 Select",
	"joypad 2 Start",
	"joypad 2 Up",
	"joypad 2 Down",
	"joypad 2 Left",
	"joypad 2 Right",
	
	/* joypad 3 */
	"joypad 3 A",
	"joypad 3 B",
	"joypad 3 Select",
	"joypad 3 Start",
	"joypad 3 Up",
	"joypad 3 Down",
	"joypad 3 Left",
	"joypad 3 Right",
	
	/* joypad 4 */
	"joypad 4 A",
	"joypad 4 B",
	"joypad 4 Select",
	"joypad 4 Start",
	"joypad 4 Up",
	"joypad 4 Down",
	"joypad 4 Left",
	"joypad 4 Right",
	
	/* power pad/family trainer */
	"mat button 1 ",
	"mat button 2",
	"mat button 3",
	"mat button 4",
	"mat button 5",
	"mat button 6",
	"mat button 7",
	"mat button 8",
	"mat button 9",
	"mat button 10",
	"mat button 11",
	"mat button 12",
	
	/* hypershot */
	"hypershot I Jump",
	"hypershot I Run",
	"hypershot II Jump",
	"hypershot II Run",
	
	/* keyboard */
	"keyboard 0",
	"keyboard 1",
	"keyboard 2",
	"keyboard 3",
	"keyboard 4",
	"keyboard 5",
	"keyboard 6",
	"keyboard 7",
	"keyboard 8",
	"keyboard 9",
	"keyboard A ",
	"keyboard B ",
	"keyboard C ",
	"keyboard D ",
	"keyboard E ",
	"keyboard F ",
	"keyboard G ",
	"keyboard H ",
	"keyboard I ",
	"keyboard J ",
	"keyboard K ",
	"keyboard L ",
	"keyboard M ",
	"keyboard N ",
	"keyboard O ",
	"keyboard P ",
	"keyboard Q ",
	"keyboard R ",
	"keyboard S ",
	"keyboard T ",
	"keyboard U ",
	"keyboard V ",
	"keyboard W ",
	"keyboard X ",
	"keyboard Y ",
	"keyboard Z ",
	"keyboard F1",
	"keyboard F2",
	"keyboard F3",
	"keyboard F4",
	"keyboard F5",
	"keyboard F6",
	"keyboard F7",
	"keyboard F8",
	"keyboard -",
	"keyboard ^",
	"keyboard Yen",
	"keyboard @",
	"keyboard [",
	"keyboard ]",
	"keyboard ;",
	"keyboard :",
	"keyboard ,",
	"keyboard .",
	"keyboard /",
	"keyboard _",
	"keyboard Spacebar",
	"keyboard Escape",
	"keyboard Enter",
	"keyboard Clear Home",
	"keyboard Insert",
	"keyboard Delete",
	"keyboard Up",
	"keyboard Down",
	"keyboard Left  ",
	"keyboard Right  ",
	"keyboard Stop",
	"keyboard Left Shift",
	"keyboard Right Shift",
	"keyboard Ctrl",
	"keyboard Graph",
	"keyboard Kana"
};

u32 cont_get_port(u32 p) { if (p<CONT_P_MAX) return cont_port[p]; else return CONT_D_NOTHING; }
const char* cont_get_name(u32 i) { if (i>=CONT_D_MAX) return NULL; else return cont_name[i]; }
const char* cont_get_a_name(u32 i) { if (i>=CONT_DA_MAX) return NULL; else return cont_a_name[i]; }
const char* cont_get_trigger_info(u32 i) { if (i>=CONT_T_MAX) return input_get_trigger_info(i-CONT_T_MAX); else return cont_t_info[i]; }

u32 cont_get_da(u32 p)
{
	u32 ret=(CONT_DA_EMPTY)|(CONT_DA_EMPTY<<8)|(CONT_DA_EMPTY<<16)|(CONT_DA_EMPTY<<24);
	
	switch (p) {
		
		/* port 1 */
		case CONT_P1:
			switch (cont_get_port(CONT_P1)) {
				case CONT_D_NJ: case CONT_D_FJ: ret=(ret&0xffffff00)|CONT_DA_J1; break; /* standard joypad */
				case CONT_D_FS: ret=(CONT_DA_J1)|(CONT_DA_J2<<8)|(CONT_DA_J3<<16)|(CONT_DA_J4<<24); break; /* four score */
				case CONT_D_PP: ret=(ret&0xffffff00)|CONT_DA_PPFT; break; /* powerpad */
				default: break;
			}
			break;
		
		/* port 2 */
		case CONT_P2:
			switch (cont_get_port(CONT_P2)) {
				case CONT_D_NJ: case CONT_D_FJ: ret=(ret&0xffffff00)|CONT_DA_J2; break; /* standard joypad */
				case CONT_D_FS: ret=(CONT_DA_J1)|(CONT_DA_J2<<8)|(CONT_DA_J3<<16)|(CONT_DA_J4<<24); break; /* four score */
				case CONT_D_PP: ret=(ret&0xffffff00)|CONT_DA_PPFT; break; /* powerpad */
				default: break;
			}
			break;
		
		/* extended port */
		case CONT_PEXT:
			switch (cont_get_port(CONT_PEXT)) {
				case CONT_D_FT: ret=(ret&0xffffff00)|CONT_DA_PPFT; break; /* familytrainer */
				case CONT_D_HS: ret=(ret&0xffffff00)|CONT_DA_HS; break; /* hypershot */
				case CONT_D_4P: ret=(ret&0x0000ffff)|(CONT_DA_J3<<16)|(CONT_DA_J4<<24); break; /* 4 players adaptor */
				case CONT_D_FK: ret=(ret&0xffffff00)|CONT_DA_KEY; break; /* family basic keyboard */
				default: break;
			}
			break;
		
		default: break;
	}
	
	return ret;
}

/* handlers */
/* write */
#define _CONTWRITE(x) static void __fastcall _cont_write_##x(int data,int port)
_CONTWRITE(nothing); _CONTWRITE(nothing) { return; }
typedef void __fastcall(*_fcontwrite)(int,int);
static _fcontwrite _contwrite[CONT_P_MAX];
#define SET_CONTWRITE(x) _contwrite[port]=_cont_write_##x

/* read */
#define _CONTREAD(x) static int __fastcall _cont_read_##x(int address,int port)
_CONTREAD(nothing); _CONTREAD(nothing) { return 0; }
typedef int __fastcall(*_fcontread)(int,int);
static _fcontread _contread[CONT_P_MAX];
#define SET_CONTREAD(x) _contread[port]=_cont_read_##x

/* init */
#define _FCONTINIT(x) _cont_init_##x
#define _CONTINIT(x) static void _FCONTINIT(x) (int port)
_CONTINIT(joypad);
_CONTINIT(powerpad);
_CONTINIT(familytrainer);
_CONTINIT(hypershot);
_CONTINIT(fourplayer);
_CONTINIT(familybasic);
_CONTINIT(nothing); _CONTINIT(nothing) { return; }
typedef void (*_fcontinit)(int);
static const _fcontinit _continit[CONT_D_MAX]={
	_FCONTINIT(joypad),			/* nes joypad */
	_FCONTINIT(joypad),			/* famicom joypad */
	_FCONTINIT(powerpad),		/* powerpad */
	_FCONTINIT(familytrainer),	/* family trainer */
	_FCONTINIT(hypershot),		/* hypershot */
	_FCONTINIT(joypad),			/* four score */
	_FCONTINIT(fourplayer),		/* 4 players adaptor */
	_FCONTINIT(familybasic),	/* family basic keyboard */
	
	_FCONTINIT(nothing)
};

/* clean */
#define _CONTCLEAN(x) static void _cont_clean_##x(int port)
_CONTCLEAN(nothing); _CONTCLEAN(nothing) { return; }
typedef void (*_fcontclean)(int);
static _fcontclean _contclean[CONT_P_MAX];
#define SET_CONTCLEAN(x) _contclean[port]=_cont_clean_##x

/* reset */
#define _CONTRESET(x) static void _cont_reset_##x(int port)
_CONTRESET(nothing); _CONTRESET(nothing) { return; }
typedef void (*_fcontreset)(int);
static _fcontreset _contreset[CONT_P_MAX];
#define SET_CONTRESET(x) _contreset[port]=_cont_reset_##x

/* update */
#define _CONTUPDATE(x) static void __fastcall _cont_update_##x(int port)
_CONTUPDATE(nothing); _CONTUPDATE(nothing) { return; }
typedef void __fastcall(*_fcontupdate)(int);
static _fcontupdate _contupdate[CONT_P_MAX];
#define SET_CONTUPDATE(x) _contupdate[port]=_cont_update_##x

/* shortcuts */
void cont_init_cont(void) { cont_zero_structs(); _continit[cont_port[CONT_P1]](CONT_P1); _continit[cont_port[CONT_P2]](CONT_P2); _continit[cont_port[CONT_PEXT]](CONT_PEXT); }
void cont_clean_cont(void) { _contclean[CONT_P1](CONT_P1); _contclean[CONT_P2](CONT_P2); _contclean[CONT_PEXT](CONT_PEXT); }
void cont_reset(void) { _contreset[CONT_P1](CONT_P1); _contreset[CONT_P2](CONT_P2); _contreset[CONT_PEXT](CONT_PEXT); }
void __fastcall cont_update(void) { _contupdate[CONT_P1](CONT_P1); _contupdate[CONT_P2](CONT_P2); _contupdate[CONT_PEXT](CONT_PEXT); }
u8 __fastcall cont_read_port(u16 address) { return 0x40|_contread[address&1](address&1,address&1)|_contread[CONT_PEXT](address&1,CONT_PEXT); }
void __fastcall cont_write_port(u8 data) { _contwrite[CONT_P1](data&1,CONT_P1); _contwrite[CONT_P2](data&1,CONT_P2); _contwrite[CONT_PEXT](data&7,CONT_PEXT); }


void cont_init(void)
{
	int port;
	
	/* get inserted controllers settings */
	for (port=CONT_P1;port<CONT_P_MAX;port++) {
		int setport=0;
		char* c=NULL;
		
		switch (port) {
			case CONT_P1: setport=SETTINGS_PORT1; break;
			case CONT_P2: setport=SETTINGS_PORT2; break;
			case CONT_PEXT: setport=SETTINGS_PORTEXT; break;
			default: break;
		}
		
		SETTINGS_GET_STRING(settings_info(setport),&c);
		cont_port[port]=cont_port_default[port];
		
		if (c&&strlen(c)) {
			int cont;
			for (cont=0;cont<CONT_D_MAX;cont++) if (stricmp(c,cont_name[cont])==0) break;
			switch (port) {
				/* port 1,2 */
				case CONT_P1: case CONT_P2:
					switch (cont) {
						case CONT_D_FK: case CONT_D_HS: case CONT_D_MAX: cont=cont_port_default[port]; break; /* keyboard,hypershot disallowed */
						case CONT_D_FT: cont=CONT_D_PP; break; /* familytrainer->powerpad */
						case CONT_D_4P: cont=CONT_D_FS; break; /* 4 players adaptor->four score */
						default: break;
					}
					break;
				
				/* extended port */
				case CONT_PEXT:
					switch (cont) {
						case CONT_D_NJ: case CONT_D_FJ: case CONT_D_MAX: cont=cont_port_default[port]; break; /* joypad disallowed */
						case CONT_D_PP: cont=CONT_D_FT; break; /* powerpad->familytrainer */
						case CONT_D_FS: cont=CONT_D_4P; break; /* four score->4 players adaptor */
						default: break;
					}
					break;
				
				default: break;
			}
			cont_port[port]=cont;
		}
		MEM_CLEAN(c);
	}
	
	/* 1 standard port=four score -> both are four score */
	if (cont_port[CONT_P1]==CONT_D_FS||cont_port[CONT_P2]==CONT_D_FS) cont_port[CONT_P1]=cont_port[CONT_P2]=CONT_D_FS;
	
	/* functionpointers to nothing */
	for (port=CONT_P1;port<CONT_P_MAX;port++) {
		SET_CONTWRITE(nothing);
		SET_CONTREAD(nothing);
		SET_CONTUPDATE(nothing);
		SET_CONTCLEAN(nothing);
		SET_CONTRESET(nothing);
	}
}

void cont_clean(void)
{
	;
}

void cont_settings_save(void)
{
	/* called from input.c input_settings_save */
	SETTINGS_PUT_STRING(NULL,"\n; Valid for standard port: ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_NJ)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_FJ)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_PP)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_FS)); SETTINGS_PUT_STRING(NULL,",\n; -- or ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_NOTHING)); SETTINGS_PUT_STRING(NULL,".\n");
	
	SETTINGS_PUT_STRING(NULL,"; Valid for "); SETTINGS_PUT_STRING(NULL,settings_info(SETTINGS_PORTEXT)); SETTINGS_PUT_STRING(NULL,": ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_FT)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_HS)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_4P)); SETTINGS_PUT_STRING(NULL,",\n; -- ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_FK)); SETTINGS_PUT_STRING(NULL,", or ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_NOTHING)); SETTINGS_PUT_STRING(NULL,".\n\n");
	
	SETTINGS_PUT_STRING(settings_info(SETTINGS_PORT1),cont_name[cont_port[CONT_P1]]);
	SETTINGS_PUT_STRING(settings_info(SETTINGS_PORT2),cont_name[cont_port[CONT_P2]]);
	SETTINGS_PUT_STRING(settings_info(SETTINGS_PORTEXT),cont_name[cont_port[CONT_PEXT]]);
}


#include "cont_joypad.c"	/* NES/Famicom joypad, four score */
#include "cont_keyboard.c"	/* family basic keyboard */
#include "cont_powerpad.c"	/* powerpad, family trainer */
#include "cont_hypershot.c"	/* HyperShot */


static void cont_zero_structs(void)
{
	#define CONT_ZERO_STRUCT(x) memset(&x,0,sizeof(x))
	
	CONT_ZERO_STRUCT(_cont_joypad);
	CONT_ZERO_STRUCT(_cont_keyboard);
	CONT_ZERO_STRUCT(_cont_powerpad);
	CONT_ZERO_STRUCT(_cont_hypershot);
}
