/*

standard controller: used by almost every game

 u
l r  select start   B A
 d

The original Famicom (not AV) had 2 controllers stuck to the console, the 2nd controller
didn't have start/select buttons, having a built-in microphone with volume-slider in place:
	Doraemon, Family Basic, Legend of Zelda (FDS), PlayBox Basic(?),
	Raid on Bungeling Bay, Takeshi no Chousenjou, Kid Icarus(?), Famicom Test Cart

The NES Advantage (joystick with player switch and rapid fire), the NES Max (joypad with
rapid fire), and 3rd party joysticks/joypads are standard and their special functions can
be simulated. Some devices can be stick onto the original joypads (eg. rapid fire devices, or
stick-on joysticks for Crazy Climber), and likewise can be simulated too (using trigger rules).

Four Score (or NES Sattelite: infra red), uses both controller ports
(currently only supporting standard joypad devices on it for now):
	Bomberman II, Danny Sullivan's Indy Heat, Gauntlet II, Greg Norman's Golf Power,
	Harlem Globetrotters, Ivan Ironman Stewart's Super Off-Road, Kings of the Beach,
	M.U.L.E., Magic Johnson's Fast Break, Monster Truck Rally, NES Play Action Football,
	A Nightmare on Elm Street, R.C. Pro-Am 2, Roundball - 2-on-2 Challenge, Super Jeopardy!,
	Swords and Serpents, Top Players' Tennis - Featuring Chris Evert & Ivan Lendl,
	Nintendo World Cup, Super Spike V'Ball, (Super Spike V'Ball + Nintendo World Cup),
	Spot, Quattro Sports/Super Sports Challenge, Smash TV

4 Players Adaptor (Famicom, by Hori), similar to the four score, except it uses the extended port:
	Kunio Kun no Nekketsu Soccer League, Nekketsu Koukou Dodgeball Bu, Bikkuri Nekketsu
	Shin Kiroku! - Harukanaru Kin Medal, Nekketsu Kakutou Densetsu, Nekketsu!
	Street Basket - Ganbare Dunk Heroes, Downtown - Nekketsu Koushin Kyoku -
	Soreyuke Daiundoukai, Bomberman II, (some 2-joypad games are also compatible with it: 1=3,2=4)

*/


static struct {
	int strobe;
	int shift;
	int shiftmax;
	int mic;
	int btn;
	int btn_upd;
	
	int four_strobe;
	int four_shift[2];
	int four_btn[2];
	int four_btn_upd[2];
	
} _cont_joypad[CONT_P_MAX];

#define cont_joypad _cont_joypad[port]


/* std, fourscore */

_CONTREAD(joypad)
{
	/* 1st port famicom joypad microphone state needs constant update */
	if (cont_joypad.mic&0x40) input_need_update();
	
	if (cont_joypad.strobe) {
		input_need_update();
		cont_joypad.btn=cont_joypad.btn_upd;
		return (cont_joypad.btn&1)|cont_joypad.mic;
	}
	else if (cont_joypad.shift<cont_joypad.shiftmax) {
		int ret=(cont_joypad.btn&1)|cont_joypad.mic;
		cont_joypad.shift++; cont_joypad.btn>>=1;
		return ret;
	}
	else return 1|cont_joypad.mic;
}

_CONTWRITE(joypad)
{
	if (data) cont_joypad.shift=0;
	else if (cont_joypad.strobe) { /* 1 -> 0 */
		input_need_update();
		cont_joypad.btn=cont_joypad.btn_upd;
	}
	cont_joypad.strobe=data;
}

_CONTUPDATE(joypad)
{
	int i,j=0;
	cont_joypad.btn_upd=0;
	cont_joypad.mic=0;
	
	switch (port) {
		case CONT_P1:
			/* serial button status in bit 0: a,b,se,st,u,d,l,r */
			for (i=CONT_T_J1_BEGIN;i<CONT_T_J1_END;i++) cont_joypad.btn_upd|=(input_trigger_held(i)<<j++); /* std */
			if (cont_port[CONT_P1]==CONT_D_FJ&&cont_port[CONT_P2]==CONT_D_FJ) cont_joypad.mic=input_trigger_held(CONT_T_MIC)<<2|0x40; /* 2 famicom pads: microphone: 4016 bit 2, not serial */
			else if (cont_port[port]==CONT_D_FS) {
				/* four score: player 3 in bits 8-15, signature (bit 19 set) in bits 16-23 */
				for (i=CONT_T_J3_BEGIN;i<CONT_T_J3_END;i++) cont_joypad.btn_upd|=(input_trigger_held(i)<<j++);
				cont_joypad.btn_upd|=BIT(19);
			}
			break;
		case CONT_P2:
			for (i=CONT_T_J2_BEGIN;i<CONT_T_J2_END;i++) cont_joypad.btn_upd|=(input_trigger_held(i)<<j++); /* std */
			if (cont_port[CONT_P1]==CONT_D_FJ&&cont_port[CONT_P2]==CONT_D_FJ) cont_joypad.btn_upd&=~0xc; /* 2 famicom pads: disable 2nd pad start and select */
			else if (cont_port[port]==CONT_D_FS) {
				/* four score: player 4 in bits 8-15, signature (bit 18 set) in bits 16-23 */
				for (i=CONT_T_J4_BEGIN;i<CONT_T_J4_END;i++) cont_joypad.btn_upd|=(input_trigger_held(i)<<j++); /* std */
				cont_joypad.btn_upd|=BIT(18);
			}
			break;
		
		default: break;
	}
	
	/* switch left/right */
	if (draw_is_flip_x()) cont_joypad.btn_upd=(cont_joypad.btn_upd&0xff3f3f)|(cont_joypad.btn_upd<<1&0x8080)|(cont_joypad.btn_upd>>1&0x4040);
}

_CONTRESET(joypad)
{
	_contwrite[port](1,port);
	_contwrite[port](0,port);
}

_CONTINIT(joypad)
{
	SET_CONTREAD(joypad);
	SET_CONTWRITE(joypad);
	SET_CONTUPDATE(joypad);
	SET_CONTRESET(joypad);
	cont_joypad.shiftmax=(cont_port[port]==CONT_D_FS)<<4|8; /* normal: 8, four score: 24 */
	cont_reset();
}


/* 4 Players Adaptor: exactly the same as normal, but using 2nd bit for p3/4 */

_CONTREAD(fourplayer)
{
	if (cont_joypad.four_strobe) {
		input_need_update();
		cont_joypad.four_btn[address]=cont_joypad.four_btn_upd[address];
		return cont_joypad.four_btn[address]&3;
	}
	else if (cont_joypad.four_shift[address]<8) {
		int ret=cont_joypad.four_btn[address]&3;
		cont_joypad.four_shift[address]++; cont_joypad.four_btn[address]>>=2;
		return ret;
	}
	else return 3;
}

_CONTWRITE(fourplayer)
{
	if (data&1) cont_joypad.four_shift[0]=cont_joypad.four_shift[1]=0;
	else if (cont_joypad.four_strobe) {
		input_need_update();
		cont_joypad.four_btn[0]=cont_joypad.four_btn_upd[0];
		cont_joypad.four_btn[1]=cont_joypad.four_btn_upd[1];
	}
	cont_joypad.four_strobe=data&1;
}

_CONTUPDATE(fourplayer)
{
	int i,j;
	cont_joypad.four_btn_upd[0]=cont_joypad.four_btn_upd[1]=0;
	j=0; for (i=CONT_T_J1_BEGIN;i<CONT_T_J1_END;i++,j+=2) cont_joypad.four_btn_upd[0]|=(input_trigger_held(i)<<j);
	j=1; for (i=CONT_T_J3_BEGIN;i<CONT_T_J3_END;i++,j+=2) cont_joypad.four_btn_upd[0]|=(input_trigger_held(i)<<j);
	j=0; for (i=CONT_T_J2_BEGIN;i<CONT_T_J2_END;i++,j+=2) cont_joypad.four_btn_upd[1]|=(input_trigger_held(i)<<j);
	j=1; for (i=CONT_T_J4_BEGIN;i<CONT_T_J4_END;i++,j+=2) cont_joypad.four_btn_upd[1]|=(input_trigger_held(i)<<j);
	
	/* switch left/right */
	if (draw_is_flip_x()) {
		cont_joypad.four_btn_upd[0]=(cont_joypad.four_btn_upd[0]&0xfff)|(cont_joypad.four_btn_upd[0]<<2&0xc000)|(cont_joypad.four_btn_upd[0]>>2&0x3000);
		cont_joypad.four_btn_upd[1]=(cont_joypad.four_btn_upd[1]&0xfff)|(cont_joypad.four_btn_upd[1]<<2&0xc000)|(cont_joypad.four_btn_upd[1]>>2&0x3000);
	}
}

_CONTINIT(fourplayer)
{
	SET_CONTREAD(fourplayer);
	SET_CONTWRITE(fourplayer);
	SET_CONTUPDATE(fourplayer);
}
