#ifndef CPU_H
#define CPU_H

#define CPU_EMPTY_RAM		FALSE /* ram is not entirely empty on first boot.. due to lockout chip? */


#define INTERRUPT_NONE		0
#define INTERRUPT_BRK		BIT(0)
#define INTERRUPT_IRQ		BIT(1)
#define INTERRUPT_NMI		BIT(2)
#define INTERRUPT_RES		BIT(3)

#define IRQ_FRAME			BIT(4)
#define IRQ_DMC				BIT(5)
#define IRQ_MAPPER			BIT(6)
#define IRQ_MAPPER2			BIT(7)

#define INT_EDGE_NMI		BIT(0)
#define INT_EDGE_IRQ		BIT(1)

/* when calling with set_interrupt */
#define INT_EDGE_NMI_EX		(INT_EDGE_NMI<<8)
#define INT_EDGE_IRQ_EX		(INT_EDGE_IRQ<<8)


enum { CPU_FPT_NORMAL=0, CPU_FPT_WRAM_READ, CPU_FPT_WRAM_WRITE, CPU_FPT_PRG1A_READ, CPU_FPT_PRG1B_READ, CPU_FPT_PRG2A_READ, CPU_FPT_PRG2B_READ };

typedef __fastcall void(*fp_mapper_irq)(void);

void cpu_set_fpt(int);
void cpu_set_mapper_irq(fp_mapper_irq);
void __fastcall cpu_set_interrupt(int);
void __fastcall cpu_acknowledge_interrupt(int);
void cpu_set_overflow(void);
int* cpu_get_cycles_ptr(void);
int __fastcall cpu_get_cycles(void);
int cpu_get_halt(void);
u16 cpu_get_pc(void);
void cpu_new_frame(void);
int cpu_execute(const int);
void cpu_execute_halted(const int);
void cpu_reset(int);
void cpu_sync_reset_time(void);
void cpu_init(void);
void cpu_clean(void);
void cpu_fill_op_cycles_lookup(void);
void cpu_adjust_rel_cycles(void);
int __fastcall cpu_get_rel_cycles(void);
void __fastcall cpu_sprite_dma_transfer(u8);
void __fastcall cpu_sprite_dma_transfer_continue(void);
int __fastcall cpu_set_apu_fs_next_step(int);
int __fastcall cpu_test_apu_fs_next_step(int);
void __fastcall cpu_reset_apu_fs_next_step(void);
int __fastcall cpu_update_apu_fs_onclock(void);
int __fastcall cpu_set_apu_dmc_next_step(void);
int __fastcall cpu_test_apu_dmc_next_cycle(void);
void __fastcall cpu_reset_apu_dmc_freq(int);
u8 __fastcall cpu_dmc_dma_transfer(u16);
void cpu_set_nmi(int);
void cpu_do_vblank_nmi(void);
void cpu_kill_cycle(void);

#endif /* CPU_H */
