/* 1 tab = 4 spaces! */

#include <windows.h>
#include <commctrl.h>
#include <time.h>

#include "global.h"
#include "version.h"
#include "resource.h"
#include "main.h"
#include "nes.h"
#include "draw.h"
#include "file.h"
#include "settings.h"
#include "sound.h"
#include "input.h"
#include "crystal.h"
#include "cont.h"
#include "mapper.h"
#include "help.h"
#include "mt19937ar.h"


/* msg callback */
static LRESULT CALLBACK WndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
		/* for debug stuff */
		#if 0
		case WM_LBUTTONDOWN:
			return FALSE;
		case WM_LBUTTONUP:
			return FALSE;
		case WM_RBUTTONDOWN:
			return FALSE;
		case WM_RBUTTONUP:
			return FALSE;
		#endif
		
		
		case WM_DROPFILES: nes_drop(wParam); return FALSE;
		
		case WM_PAINT: case WM_SIZE: case WM_SIZING: case WM_MOVE: case WM_MOVING:
			 draw_repaint(); return FALSE;
		
		case WM_ACTIVATEAPP: MAIN->foreground=wParam; draw_repaint(); return FALSE;
		case WM_ENTERMENULOOP: case WM_ENTERSIZEMOVE: MAIN->dialog=TRUE; draw_repaint(); return FALSE;
		case WM_EXITMENULOOP: case WM_EXITSIZEMOVE: MAIN->dialog=FALSE; draw_repaint(); return FALSE;
		case WM_SYSKEYDOWN:
			if (!(nes_get_paused()&1)&&input_menukey(wParam)) return FALSE;
			break;
		
		case WM_CLOSE: nes_stop(FALSE); KillTimer(MAIN->window,0); DestroyWindow(hwnd); MAIN->window=NULL; return FALSE;
		case WM_DESTROY: PostQuitMessage(0); return FALSE;
		
		case WM_SYSCOMMAND:
			switch (wParam&0xfff0) {
				case SC_KEYMENU:
					if (!(nes_get_paused()&1)&&input_menukey(VK_MENU)&&lParam!=0) return FALSE;
					break;
				
				default: break;
			}
			break;
		
		case WM_LBUTTONDBLCLK: input_doubleclick(); return FALSE;
		
		/* the menu, bon appetit */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDM_OPENGAME: nes_open(); break;
				case IDM_RELOADGAME: nes_reload(); break;
				case IDM_CLOSEGAME: nes_stop(TRUE); sound_silence(FALSE); draw_repaint(); break;
				case IDM_DIRECTORIES: MAIN->dialog=TRUE; DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_DIRECTORIES),MAIN->window,file_directories_dialog); MAIN->dialog=FALSE; break;
				case IDM_EXITEMU: PostMessage(hwnd,WM_CLOSE,0,0); break;
				
				case IDM_RESETNES: nes_wait(); nes_set_soft_reset(); nes_wait_done(); break;
				case IDM_PAUSEEMU: nes_set_paused(nes_get_paused()^1); main_menu_check(IDM_PAUSEEMU,nes_get_paused()&1); main_menu_enable(IDM_RESETNES,(nes_get_paused()&1)^1); break;
				case IDM_SOUND: nes_wait(); sound_set_enabled(sound_get_enabled()^1); nes_wait_done(); break;
				
				case IDM_ZOOM1: nes_wait(); draw_set_zoom(1); nes_wait_done(); break;
				case IDM_ZOOM2: nes_wait(); draw_set_zoom(2); nes_wait_done(); break;
				case IDM_ZOOM3: nes_wait(); draw_set_zoom(3); nes_wait_done(); break;
				case IDM_ZOOM4: nes_wait(); draw_set_zoom(4); nes_wait_done(); break;
				case IDM_FULLSCREEN: nes_wait(); draw_switch_screenmode(); nes_wait_done(); break;
				case IDM_CORRECTASPECT: nes_wait(); draw_set_correct_aspect(draw_get_correct_aspect()^1); draw_set_repaint(2); nes_wait_done(); break;
				case IDM_STRETCHFIT: nes_wait(); draw_set_stretch(draw_get_stretch()^1); draw_set_repaint(2); nes_wait_done(); break;
				
				case IDM_RENDERD3D: if (draw_get_rendertype()!=DRAW_RENDERTYPE_D3D) { draw_set_rendertype(DRAW_RENDERTYPE_D3D); draw_set_recreate(); } break;
				case IDM_RENDERDD: if (draw_get_rendertype()!=DRAW_RENDERTYPE_DD) { draw_set_rendertype(DRAW_RENDERTYPE_DD); draw_set_recreate(); } break;
				case IDM_SOFTRENDER: draw_set_softrender(draw_get_softrender()^1); draw_set_recreate(); break;
				case IDM_SHELLMSG: nes_wait(); draw_set_text(draw_get_text()^1); nes_wait_done(); break;
				case IDM_BFILTER: draw_set_bfilter(draw_get_bfilter()^1); break;
				case IDM_FLIP_X: draw_set_flip_x(draw_get_flip_x()^1); break;
				case IDM_VSYNC: draw_set_vsync(draw_get_vsync()^1); if (draw_get_rendertype()==DRAW_RENDERTYPE_D3D) draw_set_recreate(); break;
				case IDM_VIDEOFORMAT: nes_wait(); nes_videoformat(); nes_wait_done(); break;
				
				case IDM_VSMONOCHROME: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_MONOCHROME); break;
				case IDM_VSCOMPOSITE: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_COMPOSITE); break;
				case IDM_VSSVIDEO: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_SVIDEO); break;
				case IDM_VSRGB: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_RGB); break;
				case IDM_VSCUSTOM: MAIN->dialog=TRUE; DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_VIDEOSIGNAL),MAIN->window,draw_change_vidsignal_custom); MAIN->dialog=FALSE; break;
				
				case IDM_HELP: MAIN->dialog=TRUE; DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_HELP),MAIN->window,help_help); MAIN->dialog=FALSE; break;
				case IDM_ABOUT: MAIN->dialog=TRUE; DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_ABOUT),MAIN->window,help_about); MAIN->dialog=FALSE; break;
				
				
				case IDM_CPUSPEEDPLUS: nes_wait(); crystal_set_cpuspeed(crystal->overclock+1); nes_wait_done(); break;
				case IDM_CPUSPEEDMIN: nes_wait(); crystal_set_cpuspeed(crystal->overclock-1); nes_wait_done(); break;
				case IDM_CPUSPEEDNORMAL: nes_wait(); crystal_set_cpuspeed(100); nes_wait_done(); break;
				
				case IDM_CARTSWITCHUP: nes_wait(); mapper_set_switches(MAPSWITCH_UP); nes_wait_done(); break;
				case IDM_CARTSWITCHDOWN: nes_wait(); mapper_set_switches(MAPSWITCH_DOWN); nes_wait_done(); break;
				
				default: break;
			}
			return FALSE;
		
		/* timer timeout */
		case WM_TIMER:
			WaitForSingleObject(MAIN->mutex,INFINITE);
			
			/* workaround: WM_ACTIVATEAPP should be enough to determine window priority, but for some reason it gets rarely ignored */
			MAIN->foreground=MAIN->window==GetForegroundWindow();
			
			input_mouse_cursor_update();
			sound_close_sleepthread();
			draw_check_reset(); /* resetting/recreating the renderer device is the responsibility of the thread that created it */
			
			nes_set_main_interrupt();
			ReleaseMutex(MAIN->mutex);
			return FALSE;
		
		default: break;
	}
	
	return DefWindowProc(hwnd,msg,wParam,lParam);
}

/* program entry */
int WINAPI WinMain(HINSTANCE inst,HINSTANCE prev,LPSTR arg,int show)
{
	WNDCLASSEX wc;
	MSG msg;
	RECT rect;
	HGDIOBJ bgc_start=GetStockObject(WHITE_BRUSH);
	
	memset(&wc,0,sizeof(WNDCLASSEX));
	
	InitCommonControls();
	
	init_genrand((unsigned long)time(NULL));
	log_init();
	crystal_init();
	draw_setzero();
	draw_create_text_mutex();
	draw_set_text_max_ticks_default();
	
	if ((CoInitializeEx(NULL,COINIT_APARTMENTTHREADED))!=S_OK) {
		LOG(LOG_MISC|LOG_ERROR,"can't initialize COM!\n");
		exit(1);
	}
	
	timeBeginPeriod(1);
	
	MEM_CREATE(MAIN,sizeof(Main));
	MAIN->module=inst;
	
	#if DEBUG_PROFILER
	PROFILER_INIT();
	#endif
	
	file_init();
	settings_init();
	file_directories();
	
	if ((MAIN->mutex=(CreateMutex(NULL,FALSE,NULL)))==NULL) { LOG(LOG_MISC|LOG_ERROR,"can't create main mutex!\n"); exit(1); }
	
	/* register the window class */
	wc.cbSize=sizeof(WNDCLASSEX);				/* struct size */
	wc.lpfnWndProc=WndProc;						/* pointer to wndproc */
	wc.hInstance=MAIN->module;					/* handle to app instance */
	wc.hIconSm=wc.hIcon=(HICON)LoadImage(MAIN->module,MAKEINTRESOURCE(ID_ICON),IMAGE_ICON,0,0,LR_DEFAULTSIZE); /* 32x32 icon (alt-tab), and small icon (top left, file) */
	wc.hCursor=(HCURSOR)LoadImage(NULL,MAKEINTRESOURCE(OCR_NORMAL),IMAGE_CURSOR,0,0,LR_SHARED); /* cursor when it's over the window */
	wc.hbrBackground=(HBRUSH)bgc_start;			/* bg colour */
	wc.lpszMenuName=MAKEINTRESOURCE(ID_MENU);	/* name of menu resource */
	wc.lpszClassName=version_get(VERSION_NAME);	/* class name */
	wc.style=CS_DBLCLKS;						/* style */
	
	if(!RegisterClassEx(&wc)) { LOG(LOG_MISC|LOG_ERROR,"can't register window!\n"); exit(1); }
	
	
	#define WIN_STYLE	WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN
	#define WIN_EX_STYLE	WS_EX_CLIENTEDGE|WS_EX_ACCEPTFILES
	
	/* get (window size)-(client area size) */
	rect.left=rect.top=rect.right=rect.bottom=0;
	AdjustWindowRectEx(
		&rect,			/* location of rect struct */
		WIN_STYLE,		/* window style */
		TRUE,			/* has menu? */
		WIN_EX_STYLE	/* extended style */
	);
	MAIN->x_plus=rect.right-rect.left;
	MAIN->y_plus=rect.bottom-rect.top;
	
	MAIN->menu=LoadMenu(MAIN->module,MAKEINTRESOURCE(ID_MENU));
	
	/* create window */
	MAIN->window=CreateWindowEx (
			WIN_EX_STYLE,				/* extended style */
			wc.lpszClassName,			/* class name */
			version_get(VERSION_NAME),	/* window title */
			WIN_STYLE,					/* window style */
			CW_USEDEFAULT,				/* x placement */
			CW_USEDEFAULT,				/* y placement */
			MAIN->x_plus+256*2,			/* width */
			MAIN->y_plus+224*2,			/* height */
			NULL,						/* parent window handle */
			MAIN->menu,					/* menu handle */
			MAIN->module,				/* app instance handle */
			NULL						/* pointer to window creation data */
	);
	
	if(MAIN->window==NULL) { LOG(LOG_MISC|LOG_ERROR,"can't create window!\n"); exit(1); }
	draw_init(); main_parent_window(MAIN->window,0,0,0,0,0);
	crystal_settings_load();
	switch (draw_get_vidformat()) {
		case DRAW_VIDFORMAT_AUTO: case DRAW_VIDFORMAT_NTSC: crystal_set_mode(0); break;
		case DRAW_VIDFORMAT_PAL: crystal_set_mode(1); break;
		default: break;
	}
	
	ShowWindow(MAIN->window,show);
	UpdateWindow(MAIN->window);
	
	LOG(LOG_MISC|LOG_COLOUR(LC_PINK),"Welcome to %s %s,  %s\n",version_get(VERSION_NAME),version_get(VERSION_NUMBER),version_get(VERSION_AUTHOR));
	
	if (draw_get_fullscreen_set()) draw_switch_screenmode();
	draw_draw();
	SetClassLongPtr(MAIN->window,GCL_HBRBACKGROUND,0);
	DeleteObject(bgc_start);
	
	sound_init();
	nes_init();
	input_init();
	cont_init();
	
	SetTimer(MAIN->window,0,15,NULL);
	
	if (strlen(arg)) {
		file_setfile(NULL,arg,NULL,"nes");
		nes_start();
	}
	
	
	/* messageloop */
	while(GetMessage(&msg,NULL,0,0)>0) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	
	settings_save();
	
	nes_clean();
	draw_clean();
	sound_clean();
	cont_clean();
	input_clean();
	mapper_clean_info();
	settings_clean();
	file_clean();
	
	timeEndPeriod(1);
	
	CoUninitialize();
	
	DestroyMenu(MAIN->menu);
	DestroyIcon(wc.hIconSm);
	DestroyIcon(wc.hIcon);
	CloseHandle(MAIN->mutex);
	
	UnregisterClass(wc.lpszClassName,MAIN->module);
	
	MEM_CLEAN(MAIN);
	/* LOG_ERROR unallowed from here */
	LOG(LOG_VERBOSE,"main cleaned\n");
	crystal_clean();
	
	return msg.wParam;
}

void main_menu_enable(UINT id,int en) { EnableMenuItem(MAIN->menu,id,en==FALSE?MF_GRAYED:MF_ENABLED); DrawMenuBar(MAIN->window); }
void main_menu_check(UINT id,int en) { CheckMenuItem(MAIN->menu,id,en==FALSE?MF_UNCHECKED:MF_CHECKED); DrawMenuBar(MAIN->window); }
void main_menu_radio(UINT id,UINT first,UINT last) { CheckMenuRadioItem(MAIN->menu,first,last,id,MF_BYCOMMAND); DrawMenuBar(MAIN->window); }

void main_menu_caption(int get,UINT id,char* c)
{
	MENUITEMINFO m;
	memset(&m,0,sizeof(MENUITEMINFO));
	m.cbSize=sizeof(MENUITEMINFO);
	m.fMask=MIIM_STRING;
	m.dwTypeData=c;
	if (get) { m.cch=STRING_SIZE; GetMenuItemInfo(MAIN->menu,id,FALSE,&m); }
	else { SetMenuItemInfo(MAIN->menu,id,FALSE,&m); DrawMenuBar(MAIN->window); } /* drawmenubar doesn't seem to have effect */
}


/* center window on its parent */
void main_parent_window(HWND window,UINT xloc,UINT yloc,int xoff,int yoff,int desk)
{
	RECT rect_parent,rect_window,rect;
	int x=0,y=0,xs=0,ys=0;
	
	HWND parent=GetParent(window);
	memset(&rect_parent,0,sizeof(RECT));
	memset(&rect_window,0,sizeof(RECT));
	memset(&rect,0,sizeof(RECT));
	
	if (parent==NULL||desk) { parent=GetDesktopWindow(); GetClientRect(parent,&rect_parent); }
	else GetWindowRect(parent,&rect_parent);
	GetWindowRect(window,&rect_window);
	CopyRect(&rect,&rect_parent);
	
	OffsetRect(&rect_window,-rect_window.left,-rect_window.top);
	OffsetRect(&rect,-rect.left,-rect.top);
	OffsetRect(&rect,-rect_window.right,-rect_window.bottom);
	
	xs=rect_window.right-rect_window.left;
	ys=rect_window.bottom-rect_window.top;
	
	/* set location */
	switch (xloc) {
		case MAIN_PW_OUTERL: x=rect_parent.left-xs; break;
		case MAIN_PW_LEFT: x=rect_parent.left; break;
		case MAIN_PW_RIGHT: x=rect_parent.right-xs; break;
		case MAIN_PW_OUTERR: x=rect_parent.right; break;
		
		default: x=rect_parent.left+(rect.right/2.0); break; /* center */
	}
	switch (yloc) {
		case MAIN_PW_OUTERL: y=rect_parent.top-ys; break;
		case MAIN_PW_LEFT: y=rect_parent.top; break;
		case MAIN_PW_RIGHT: y=rect_parent.bottom-ys; break;
		case MAIN_PW_OUTERR: y=rect_parent.bottom; break;
		
		default: y=rect_parent.top+(rect.bottom/2.0); break; /* center */
	}
	
	x+=xoff; y+=yoff;
	
	/* check out of bounds */
	if (!SystemParametersInfo(SPI_GETWORKAREA,0,&rect,0)) {
		rect.left=rect.top=0;
		rect.right=GetSystemMetrics(SM_CXFULLSCREEN);
		rect.bottom=GetSystemMetrics(SM_CYFULLSCREEN);
	}
	if (x>(rect.right-xs)) x=rect.right-xs;
	if (x<rect.left) x=rect.left;
	if (y>(rect.bottom-ys)) y=rect.bottom-ys;
	if (y<rect.top) y=rect.top;
	
	SetWindowPos(window,NULL,x,y,0,0,SWP_NOSIZE|SWP_NOZORDER);
}

/* set titlebar text */
void main_titlebar(const char* text)
{
	char title[STRING_SIZE]={0};
	
	if (text==NULL) SetWindowText(MAIN->window,version_get(VERSION_NAME));
	else {
		sprintf(title,"%s - %s",version_get(VERSION_NAME),text);
		SetWindowText(MAIN->window,title);
	}
}
