/* AVE (American Video Entertainment) mappers */

static struct {
	/* d-1012 */
	int d1012_r[2];
	
} map_ave;


/* --- D-1012 (234) ---
	Maxi-15
		F-15 City War, Puzzle, Blackjack, Tiles of Fate, Krazy Kreatures, Death Race, Dudes With Attitude,
		Venice Volleyball, Stakk 'm, Deathbots, Rad Racket Tennis, Chiller, Solitaire, Menace Beach, Shockwave
*/

static __inline__ void ave_d1012_bs(void);

/* soft reset */
_MAPRESET(ave_d1012)
{
	map_ave.d1012_r[0]=0;
	map_ave.d1012_r[1]=0;
	MAPIO_FUN(ave_d1012)(0xff80,0,0);
}

/* IO */
_MAPIO(ave_d1012)
{
	switch (address&0xfff8) {
		/* $ff80-$ff9f: R1 */
		case 0xff80: case 0xff88: case 0xff90: case 0xff98:
			if (map_ave.d1012_r[0]==0) {
				MIR_VH(data&BIT(7)) /* bit 7: set mirroring */
				map_ave.d1012_r[0]=data;
				ave_d1012_bs();
			}
			break;
		
		/* $ffe8-$fff7: R2 */
		case 0xffe8: case 0xfff0:
			map_ave.d1012_r[1]=data;
			ave_d1012_bs();
			break;
		
		/* ( $ffc0-$ffdf: R3 (lockout chip related) ) */
		
		default: break;
	}
	
}

/* reads are also considered IO! */
_MAPIO_R(ave_d1012)
{
	u8 bus=cartridge->cur_prgbank[3][address&0x1fff];
	MAPIO_FUN(ave_d1012)(address,bus,bus);
	return bus;
}

/* bankswitch */
static __inline__ void ave_d1012_bs(void)
{
	u16 prg,chr;
	
	if (map_ave.d1012_r[0]&BIT(6)) {
		prg=((map_ave.d1012_r[0]<<2&BIN8(00111000))|(map_ave.d1012_r[1]<<2&BIN8(00000100)))&MAXPRG;
		chr=(((map_ave.d1012_r[0]&BIN8(00001110))<<5)|(map_ave.d1012_r[1]>>1&BIN8(00111000)))&MAXVROM;
	}
	else {
		prg=(map_ave.d1012_r[0]<<2&BIN8(00111100))&MAXPRG;
		chr=(((map_ave.d1012_r[0]&BIN8(00001111))<<5)|(map_ave.d1012_r[1]>>1&BIN8(00011000)))&MAXVROM;
	}
	
	PRG32_BS(prg);
	CHR8_BS_V(chr);
}

/* init */
_MAPINIT(ave_d1012)
{
	SET_MAPIO8(nothing,nothing,nothing,ave_d1012);
	SET_MAPIO8_R(nothing,nothing,nothing,ave_d1012);
	SET_MAPRESET(ave_d1012);
	
	INIT_MIR_V(); /* R1=0 */
	
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
}





/* --- Nina-001 (iNES: 34) ---
	Impossible Mission 2
*/

_MAPIO(nina1)
{
	int bank;
	
	WRAM_WRITE(data);
	
	switch(address) {
		
		/* select 32k prg bank */
		case 0x7ffd: bank=data<<2&MAXPRG; PRG32_BS(bank); break;
		
		/* chr bankswitch */
		/* select 4k chr bank at 0x0 */
		case 0x7ffe: bank=data<<2&MAXVROM; CHR4_BS_V(0,bank); break;
		/* select 4k chr bank at 0x1000 */
		case 0x7fff: bank=data<<2&MAXVROM; CHR4_BS_V(4,bank); break;
		
		default: break;
	}
}

_MAPINIT(nina1)
{
	SET_MAPIO6(nina1);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
	DEF_6000_WRAM();
}





/* --- Nina-003/006 (iNES: 79/113) ---
	AV Hanafuda Club, Blackjack, Deathbots, Double Strike, Dudes with Attitude, F-15 City War,
	Krazy Kreatures, Mahjong Companion, Mermaids of Atlantis, Metal Fighter, Othello(NROM?),
	Papillion, Poke Block, Rad Racket - Deluxe Tennis II, Sidewinder, Solitaire, Tiles of Fate,
	Trolls on Treasure Island, Ultimate League Soccer/AV Soccer, Venice Beach Volleyball
	
   --- HES X-in-1 (same as above, only with mirroring switch) (iNES: no/113) ---
	HES 6-in-1,
		Cosmos Cop, Magic Carpet 1001, Balloon Monster, Adam and Eve, Porter, Bookyman (same games as Caltron 6-in-1)
	Mind Blower Pak,
		The Penguin Seal, Magic Mathematic, Jackpot, Galactic Crusader (all crap unlicensed games)
	Total Funpak (HES)
		Sidewinder, Othello, Duck, Pac-Man (unlicensed, except pacman; bootleg or licensed from tengen/namco?)
*/

/*
	(16 Mahjong [p1][!] has been assigned to this mapper, and doesn't work,
	though it does if swapping the 2 8k chr banks, and running it on cnrom)
*/

void __fastcall (*nina3_mirroring)(int);
static void __fastcall nina3_mirroring_no(int m) { return; }
static void __fastcall nina3_mirroring_hes(int m) { MIR_HV(m) }

_MAPIO(nina3)
{
	if ((address&0xe100)==0x4100) {
		/* bits 0,1,2,6: select 8k chr */
		int bank=((data<<3&BIN8(00111000))|(data&BIT(6)))&MAXVROM;
		CHR8_BS_V(bank);
		
		/* bits 3,4,5: select 32k prg */
		bank=data>>1&~3&MAXPRG;
		PRG32_BS(bank);
		
		/* bit 7: set mirroring (only on hes X-in-1) */
		(*nina3_mirroring)(data&0x80);
	}
}

_MAPINIT(nina3)
{
	SET_MAPIO4(nina3);
	SET_MAPIO8_ALL(nina3);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
	
	switch (maptype) {
		case MAPTYPE_AVE_NINA_003:
			nina3_mirroring=&nina3_mirroring_no;
			break;
		
		case MAPTYPE_HES_X_IN_1:
			nina3_mirroring=&nina3_mirroring_hes;
			INIT_MIR_H();
			break;
		
		default: break;
	}
}
