/* Bandai mappers */

static struct {
	/* bandai */
	int prg_hi;
	int prg_lo;
	int irq_isenabled;
	int irq_latch;
	int irq_counter;
	
} map_bandai;


/* --- Bandai LZ93D50 (iNES: 16) ---
	Akuma-kun - Makai no Wana, Crayon Shin-Chan - Ora to Poi Poi, Dragon Ball 3 - Gokuu Den,
	Dragon Ball - Dai Maou Fukkatsu, Dragon Ball Z - Kyoushuu! Saiya Jin, Dragon Ball Z Gaiden -
	Saiya Jin Zetsumetsu Keikaku, Dragon Ball Z II - Gekishin Freeza!!, Dragon Ball Z III -
	Ressen Jinzou Ningen, Famicom Jump - Eiyuu Retsuden, Famicom Jump II - Saikyou no 7 Nin,
	Magical Taruruuto Kun - Fantastic World!!, Magical Taruruuto Kun 2 - Mahou Daibouken,
	Meimon! Dai San Yakyuu Bu, Nishimura Kyoutarou Mystery - Blue Train Satsujin Jiken,
	Rokudenashi Blues, Sakigake!! Otoko Juku - Shippuu Ichi Gou Sei, SD Gundam Gaiden -
	Knight Gundam Monogatari, SD Gundam Gaiden - Knight Gundam Monogatari 2 - Hikari no Kishi,
	SD Gundam Gaiden - Knight Gundam Monogatari 3 - Densetsu no Kishi Dan, Datach -
	Battle Rush - Build Up Robot Tournament, Datach - Crayon Shin Chan - Ora to Poi Poi,
	Datach - Dragon Ball Z - Gekitou Tenkaichi Budou Kai, Datach - J League Super Top Players,
	Datach - SD Gundam - Gundam Wars, Datach - Ultraman Club - Supokon Fight!, Datach -
	Yuu Yuu Hakusho - Bakutou Ankoku Bujutsu Kai
*/

/* TODO:
	- IRQ handling is correct?
	- EPROM, games use it for saving (Rokudenashi Blues won't work at all)
*/

/* prg bankswitch */
static __inline__ void bandai_prg_bs(void)
{
	int bank=(map_bandai.prg_hi<<4|map_bandai.prg_lo)<<1&MAXPRG; PRG16_BS(0,bank);
	bank=(map_bandai.prg_hi<<4|0xf)<<1&MAXPRG; PRG16_BS(2,bank);
}

/* io */
_MAPIO(bandai)
{
	switch (address&0xf) {
		/* chr 1k bankswitch, also, bit 0=prg high bank */
		case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
			if (MAXVROM) { CHR1_BS_V(address&7,data&MAXVROM); }
			map_bandai.prg_hi=data&1; bandai_prg_bs();
			break;
		
		/* prg 16k bankswitch */
		case 8: map_bandai.prg_lo=data&0xf; bandai_prg_bs(); break;
		
		/* mirroring */
		case 9: MIR_VH01(data&3) break;
		
		/* irq */
		case 0xa:
			map_bandai.irq_isenabled=data&1;
			if (data&1) map_bandai.irq_counter=(map_bandai.irq_latch+1)*crystal->rel_cycle;
			else cpu_acknowledge_interrupt(~IRQ_MAPPER);
			break;
		case 0xb: map_bandai.irq_latch=(map_bandai.irq_latch&0xff00)|data; break; /* l */
		case 0xc: map_bandai.irq_latch=(map_bandai.irq_latch&0xff)|(data<<8); break; /* h */
		
		/* eprom */
		case 0xd:
			/* printf("%x ",data); */
			break;
		
		default: break;
	}
}

/* irq counter, based on cpu cycles */
_MAPIRQ(bandai)
{
	if (map_bandai.irq_isenabled&&(map_bandai.irq_counter=map_bandai.irq_counter-(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr))<=0) {
		cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(map_bandai.irq_counter==0?INT_EDGE_IRQ_EX:0));
		map_bandai.irq_counter+=(0x10000*crystal->rel_cycle);
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

/* init */
_MAPINIT(bandai)
{
	memset(cartridge->wram,0xff,0x2000); /* Famicom Jump II expects WRAM to be flushed with $FF */
	if (!cartridge->wram_isenabled) { SET_MAPIO6(bandai); }
	SET_MAPIO8_ALL(bandai);
	INIT_CHR_FIRST();
	INIT_MIR_V();
	SET_MAPIRQ(bandai);
	SET_MAPNF(updateoldcycles);
	bandai_prg_bs();
}





/* --- simple P16/C8/M2(MPPPCCCC) (iNES: 70) ---
	Kamen Rider Club, Space Shadow, Family Trainer - Manhattan Police, Family Trainer - Meiro Daisakusen *** all Japanese
	
   --- simple P16/C8/M1(MPPPCCCC) (iNES: 152) ---
	Arkanoid 2, Gegege no Kitarou 2 - Youkai Gundan no Chousen, Pocket Zaurus - Juu Ouken no Nazo, Saint Seiya - Ougon Densetsu *** all Japanese
*/

void __fastcall (*bandai_l1_mirroring)(int);
static void __fastcall bandai_l1_mirroring_m1(int m) { MIR_01(m) }
static void __fastcall bandai_l1_mirroring_m2(int m) { MIR_VH(m) }

_MAPIO(bandai_l1)
{
	/* low nibble: swap 8k vrom */
	int bank=data<<3&MAXVROM;
	CHR8_BS_V(bank);
	
	/* bits 4,5,6: swap 16k rom in $8000 */
	bank=(data&BIN8(01110000))>>3&MAXPRG;
	PRG16_BS(0,bank);
	
	/* bit 7: set mirroring */
	(*bandai_l1_mirroring)(data&0x80);
}

_MAPINIT(bandai_l1)
{
	SET_MAPIO8_ALL(bandai_l1);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	
	switch (maptype) {
		/* M1 */
		case MAPTYPE_BANDAI_SIMPLE_M1:
			bandai_l1_mirroring=&bandai_l1_mirroring_m1;
			INIT_MIR_0();
			break;
		
		/* M2 */
		case MAPTYPE_BANDAI_SIMPLE_M2:
			bandai_l1_mirroring=&bandai_l1_mirroring_m2;
			INIT_MIR_V();
			break;
		
		default: break;
	}
}
