/* Irem mappers */

static struct {
	/* g-101 */
	int g101_bsb;
	int g101_prg;
	int g101_smir; /* alternate mirroring */
	
	/* h-3001 */
	int h3001_irq_isenabled;
	int h3001_irq_latch;
	int h3001_irq_counter;

	/* simple (napoleon senki) */
	u8* l2_nt[4];
	u8* l2_vram[8];

} map_irem;


/* --- G101 (iNES: 32) ---
	Ai Sensei no Oshiete - Watashi no Hoshi, Paaman 1,2, Kaiketsu Yanchamaru 2, Image Fight, Meikyuu Shima *** all Japanese
	
   --- with alternate mirroring (iNES: no/32) ---
	Major League (Japanese)
*/

/* io */
_MAPIO(irem_g101)
{
	switch (address>>12&3) {
		
		#define IREM_G101_BS() \
		if (map_irem.g101_bsb) { PRG8_BS(0,MAXPRG-1); PRG8_BS(2,map_irem.g101_prg); } \
		else { PRG8_BS(0,map_irem.g101_prg); PRG8_BS(2,MAXPRG-1); }
		
		/* $8000: select 8k prg bank at $8000 or $c000 */
		case 0: map_irem.g101_prg=data&MAXPRG; IREM_G101_BS() break;
		
		/* $9000: set mirroring and switch register */
		case 1:
			map_irem.g101_bsb=data&2;
			IREM_G101_BS()
			MIR_VH(data&1)
			break;
		
		/* $a000: select 8k prg bank at $a000 */
		case 2: PRG8_BS(1,data&MAXPRG); break;
		
		/* $b000: chr bankswitch, subtype: bit 6: set custom mirroring */
		case 3:
			CHR1_BS_V(address&7,data&MAXVROM); /* $b000: chr bankswitch */
			if (data&map_irem.g101_smir) { MIR_C(~address&1); } /* subtype: bit 6: set custom mirroring */
			break;
	}
}

/* init */
_MAPINIT(irem_g101)
{
	SET_MAPIO8(irem_g101,irem_g101,nothing,nothing);
	INIT_PRG_FIRSTLAST();
	/* no chrbanks init */
	DEF_6000_WRAM(); /* Image Fight */
	
	switch (maptype) {
		case MAPTYPE_IREM_G_101:
			INIT_MIR_V();
			break;
		
		case MAPTYPE_IREM_G_101_ALTERNATE_MIRRORING:
			map_irem.g101_smir=0x40;
			INIT_MIR(0);
			break;
		
		default: break;
	}
}





/* H-3001 (iNES: 65) ---
	Daiku no Gen San 2, Kaiketsu Yanchamaru 3, Spartan X 2 *** all Japanese
*/

/* TODO:
	$9000?: Spartan X 2 writes $40 to it, other games 0
	$9002?
	IRQ handling is correct? (Spartan X 2 intro)
*/

/* io */
_MAPIO(irem_h3001)
{
	switch (address) {
		
		/* select 8k prg bank at $8000 */
		case 0x8000: PRG8_BS(0,data&MAXPRG); break;
		
		/* set mirroring (bit 7) */
		case 0x9001: MIR_VH(data&0x80) break;
		
		/* irq */
		case 0x9003: map_irem.h3001_irq_isenabled=data&0x80; if (map_irem.h3001_irq_isenabled==0) cpu_acknowledge_interrupt(~IRQ_MAPPER); break; /* enable/disable irqs */
		case 0x9004: map_irem.h3001_irq_counter=map_irem.h3001_irq_latch*crystal->rel_cycle; break; /* set irq counter */
		case 0x9005: map_irem.h3001_irq_latch=(map_irem.h3001_irq_latch&0xff)|(data<<8); break; /* set irq latch: high byte */
		case 0x9006: map_irem.h3001_irq_latch=(map_irem.h3001_irq_latch&0xff00)|data; break; /* set irq latch: low byte */
		
		/* select 8k prg bank at $a000 */
		case 0xa000: PRG8_BS(1,data&MAXPRG); break;
		
		/* select 1k chr bank */
		case 0xb000: case 0xb001: case 0xb002: case 0xb003: case 0xb004: case 0xb005: case 0xb006: case 0xb007:
			CHR1_BS_V(address&7,data&MAXVROM); break;
		
		/* select 8k prg bank at $c000 */
		case 0xc000: PRG8_BS(2,data&MAXPRG); break;
		
		default:
			#if DEBUG_MAPPER
			LOG(LOG_MAPPER,"%x %x\n",address,data);
			#endif
			break;
	}
}

/* irq handler, based on cpu cycles */
_MAPIRQ(irem_h3001)
{
	if (map_irem.h3001_irq_isenabled) {
		map_irem.h3001_irq_counter-=(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr);
		if (map_irem.h3001_irq_counter<=0) {
			cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(map_irem.h3001_irq_counter==0?INT_EDGE_IRQ_EX:0));
			map_irem.h3001_irq_isenabled=FALSE;
		}
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

/* init */
_MAPINIT(irem_h3001)
{
	SET_MAPIRQ(irem_h3001);
	SET_MAPNF(updateoldcycles);
	SET_MAPIO8_ALL(irem_h3001);
	INIT_PRG_FIRSTLAST();
	/* no chrbanks init */
	INIT_MIR_V();
}





/* SIMPLE TYPES */


/* --- simple P16u/M2(MMxxPPPP) (iNES: 97) ---
	Kaiketsu Yanchamaru (Japanese version of Kid Niki)
*/

/* TODO:
	- (startup) mirroring?
*/

_MAPIO(irem_l1)
{
	/* swap 16k prgbank in $c000 */
	int bank=data<<1&MAXPRG;
	PRG16_BS(2,bank);
	
	/* bits 6,7: set mirroring */
	if ((data<<1^data)&0x80) { MIR_HV(data&0x80) }
}

_MAPINIT(irem_l1)
{
	SET_MAPIO8_ALL(irem_l1);
	INIT_MIR_U();
	INIT_PRG_LASTFIRST();
	INIT_CHR_FIRST(); /* 8k vram */
}



/* --- simple P32/C2(CCCCxxPP) (iNES: 77) ---
	Napoleon Senki (Japanese)
*/

_MAPIO(irem_l2)
{
	int bank;
	
	/* low nibble: select 32k prg bank */
	bank=data<<2&MAXPRG;
	PRG32_BS(bank);
	
	/* high nibble: select 2k chr bank in ppu $0000 */
	bank=data>>3&BIN8(00011110)&MAXVROM;
	CHR2_BS_V(0,bank);
}

_MAPCLEAN(irem_l2)
{
	int i;
	
	for (i=2;i<8;i++) { MEM_CLEAN(map_irem.l2_vram[i]); }
	for (i=0;i<4;i++) { MEM_CLEAN(map_irem.l2_nt[i]); }
}

_MAPINIT(irem_l2)
{
	int i;
	
	SET_MAPIO8_ALL(irem_l2);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
	INIT_MIR_4();
	SET_MAPCLEAN(irem_l2);
	
	/* 4-screen mirroring, 6k vram at ppu $0800-$1fff */
	for (i=2;i<8;i++) {
		MEM_CREATE(map_irem.l2_vram[i],0x400);
		cartridge->cur_patbank_isram[i]=TRUE;
		cartridge->cur_patbank[i]=map_irem.l2_vram[i];
	}
	for (i=0;i<4;i++) {
		MEM_CREATE(map_irem.l2_nt[i],0x400);
		mapper.ppu_memmap_cur_name_ptr[i]=map_irem.l2_nt[i];
	}
}



/* --- simple P16/C8/M2(CCCCMPPP) (similar to one of Jaleco's JF mappers) (iNES: 78) ---
	Holy Diver (Japanese)
*/

_MAPIO(irem_l3)
{
	/* bits 0,1,2: swap 16k prgbank in $8000 */
	int bank=data<<1&MAXPRG;
	PRG16_BS(0,bank);
	
	/* bit 3: set mirroring */
	MIR_HV(data&BIT(3))
	
	/* high nibble: swap 8k vrombank */
	bank=(data&0xf0)>>1&MAXVROM;
	CHR8_BS_V(bank);
}

_MAPINIT(irem_l3)
{
	INIT_MIR_H();
	SET_MAPIO8_ALL(irem_l3);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
}
