/* Jaleco mappers */

/* SS 88005 is like Konami VRC I */

static struct {
	/* ss88006 */
	int ss_prg[4];
	int ss_chr[8];
	int ss_irq_latch;
	int ss_irq_counter;
	int ss_irq_hit;
	int ss_irq_isenabled;
	
	int ss_sn;
	
	/* simple (tennis/pinball) */
	int s3_chr;
	int s3_prg;
	
} map_jaleco;


/* --- SS 88006 (iNES: 18) ---
	Goal!!, Jajamaru Gekimaden - Maboroshi no Kinmajou, Lord of King, Magic John, Mezase Top Pro - Green ni Kakeru Yume,
	*Moe Pro! - Saikyou Hen, *Moe Pro! '90 - Kandou Hen, Ninja Jajamaru - Ginga Daisakusen, Pizza Pop!, Plasma Ball,
	Saiyuuki World 2 - Tenjoukai no Majin, *Shin Moero!! Pro Yakyuu, *Terao no Dosukoi Oozumou, Toukon Club,
	Tsuru Pika Hagemaru - Mezase! Tsuru Seko no Akashi, USA Ice Hockey in FC *** all Japanese
	
	*: games using the add-on sound
*/

/* TODO:
	- $9002/3?
	- $f003, correct extra sound emulation (UPD7755C or UPD7756C)
	- IRQ handling is correct? (earlier irq will cause Toukon Club to screenshake)
*/

/* io */
_MAPIO(jaleco_ss88006)
{
	switch (address&0xf003) {
		/* select 8k rom in $8/a/c000 */
		case 0x8000: case 0x8001: case 0x8002: case 0x8003: case 0x9000: case 0x9001: {
			int a=(address>>11&BIT(1))|(address>>1&1);
			map_jaleco.ss_prg[a]=(map_jaleco.ss_prg[a]&~(0xf<<(address<<2&BIT(2))))|((data&0xf)<<(address<<2&BIT(2)));
			PRG8_BS(a,map_jaleco.ss_prg[a]&MAXPRG);
			break;
		}
		
		case 0x9002: break; /* ? (is not the wram_isenabled flag) */
		case 0x9003: break; /* ? */
		
		/* select 1k vrom */
		case 0xa000: case 0xa001: case 0xa002: case 0xa003: case 0xb000: case 0xb001: case 0xb002: case 0xb003:
		case 0xc000: case 0xc001: case 0xc002: case 0xc003: case 0xd000: case 0xd001: case 0xd002: case 0xd003: {
			int a=(((address^BIT(13))>>11)&6)|(address>>1&1);
			map_jaleco.ss_chr[a]=(map_jaleco.ss_chr[a]&~(0xf<<(address<<2&BIT(2))))|((data&0xf)<<(address<<2&BIT(2)));
			CHR1_BS_V(a,map_jaleco.ss_chr[a]&MAXVROM);
			break;
		}
		
		/* set irq latch */
		case 0xe000: case 0xe001: case 0xe002: case 0xe003:
			map_jaleco.ss_irq_latch=(map_jaleco.ss_irq_latch&~(0xf<<((address&3)<<2)))|((data&0xf)<<((address&3)<<2));
			break;
		
		/* load irq counter */
		case 0xf000:
			map_jaleco.ss_irq_counter=map_jaleco.ss_irq_latch;
			break;
		
		/* irq control */
		case 0xf001:
			/* bit 0: enable/disable irqs */
			map_jaleco.ss_irq_isenabled=data&1;
			if (!map_jaleco.ss_irq_isenabled) cpu_acknowledge_interrupt(~IRQ_MAPPER);
			
			/* bits 3,2,1: select hit value */
			if (data&BIT(3)) map_jaleco.ss_irq_hit=0xf;
			else if (data&BIT(2)) map_jaleco.ss_irq_hit=0xff;
			else if (data&BIT(1)) map_jaleco.ss_irq_hit=0xfff;
			else map_jaleco.ss_irq_hit=0xffff;
			
			break;
		
		/* set mirroring, bit 1 set?:0, standard with bit 0 otherwise */
		case 0xf002:
			MIR_0(data&2) else { MIR_HV(data&1) }
			break;
		
		/* sound, bit 0:? bit 1:init sample?. bits 2-6=sample number */
		case 0xf003:
			if (((map_jaleco.ss_sn&BIN8(01111100))==(data&BIN8(01111100)))&(map_jaleco.ss_sn>>1&(~data>>1))) {
				#if DEBUG_MAPPER
				LOG(LOG_MAPPER,"%02x --- %2d\n",data,data>>2&0x1f);
				#endif
				sample_play(data>>2&0x1f);
			}
			map_jaleco.ss_sn=data;
			break;
		
		default: break;
	}
}

/* irq handler, based on cpu cycles, an irq is fired on 4,8,12,or 16 bit overflow */
_MAPIRQ(jaleco_ss88006)
{
	if (map_jaleco.ss_irq_isenabled) {
		
		int cpu_cycles;
		
		for (cpu_cycles=*mapper.cpu_cycles_ptr;mapper.cpu_cycles_old>cpu_cycles;mapper.cpu_cycles_old-=crystal->rel_cycle) {
			if (!(map_jaleco.ss_irq_counter&map_jaleco.ss_irq_hit)) {
				cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|((mapper.cpu_cycles_old-cpu_cycles)>crystal->rel_cycle?0:INT_EDGE_IRQ_EX));
			}
			map_jaleco.ss_irq_counter--;
		}
	}
	
	else mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

/* init */
_MAPINIT(jaleco_ss88006)
{
	int i;
	
	map_jaleco.ss_prg[0]=0; map_jaleco.ss_prg[1]=1; map_jaleco.ss_prg[2]=MAXPRG-1; map_jaleco.ss_prg[3]=MAXPRG;
	for (i=0;i<8;i++) map_jaleco.ss_chr[i]=i;
	map_jaleco.ss_irq_hit=0xffff;
	map_jaleco.ss_sn=1;
	
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	DEF_6000_WRAM();
	INIT_MIR_H();
	SET_MAPIO8_ALL(jaleco_ss88006);
	SET_MAPIRQ(jaleco_ss88006);
	SET_MAPNF(updateoldcycles);
	
	sample_load();
	SET_MAPSTREAM(sample);
}





/* SIMPLE TYPES (JF-xx: xx is most likely the game ID, not the mapper type ID) */


/* --- Jaleco(/Konami) simple C8(xxxxxxCC) (iNES: 87) ---
	TwinBee, Argus, Choplifter, Goonies, City Connection, Ninja Jajamaru Kun, Hyper Olympic (Genteiban!),
	Urusei Yatsura - Lum no Wedding Bell (good dump) *** all Japanese
*/

_MAPIO(jaleco_l1)
{
	/* bit 1,0(reversed): select 8k vrom bank */
	data=((data<<2&BIT(3))|(data<<4&BIT(4)))&MAXVROM;
	CHR8_BS_V(data);
}

_MAPINIT(jaleco_l1)
{
	SET_MAPIO6(jaleco_l1);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
}



/* --- simple P32/C8(xCPPxxCC) (iNES: 86/101) ---
	Urusei Yatsura - Lum no Wedding Bell (bad dump), Moero!! Pro Yakyuu (Red), Moero!! Pro Yakyuu (Black) *** all Japanese
*/

/* TODO:
	correct sound emulation with 256KB UPD7756C internal ROM dump instead of samples
*/

static __fastcall void jaleco_l2_bs(u8); static __fastcall void jaleco_l2_sound(u8);
typedef __fastcall void(*fpt_jaleco_l2_io)(u8);
static const fpt_jaleco_l2_io jaleco_l2_io[2]= { jaleco_l2_bs, jaleco_l2_sound };

_MAPIO(jaleco_l2) { jaleco_l2_io[address>>12&1](data); }

/* $6000: bankswitch */
static __fastcall void jaleco_l2_bs(u8 data)
{
	int bank;
	
	/* bits 4,5: select 32k prg bank */
	bank=data>>2&BIN8(00001100)&MAXPRG;
	PRG32_BS(bank);
	
	/* bits 0,1,6: select 8k chr bank */
	bank=((data&3)|(data>>4&BIT(2)))<<3&MAXVROM;
	CHR8_BS_V(bank);
}

/* $7000: sound */
static __fastcall void jaleco_l2_sound(u8 data)
{
	/* bit 5: init sample?, bit 4: stop/play? (to play a sample, it usually writes $2x, then $3x (x=sample number)) */
	if (data>>1&~data&BIT(4)) sample_play(data&0xf);
	#if DEBUG_MAPPER
	LOG(LOG_MAPPER,"%02x --- %2d\n",data,data&0xf);
	#endif
}

_MAPINIT(jaleco_l2)
{
	SET_MAPIO6(jaleco_l2);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
	
	sample_load();
	SET_MAPSTREAM(sample);
}



/* --- simple P16/C8(PCSSBBBB) (iNES: 72) ---
	Moero!! Juudou Warriors, *Moero!! Pro Tennis, Pinball Quest *** all Japanese
	
   --- simple P16u/C8(PCSSBBBB) (iNES: 92) ---
	Moero!! Pro Soccer, *Moero!! Pro Yakyuu '88 - Ketteiban *** both Japanese
	
	*: games using the add-on sound
*/

/* TODO:
	correct sound emulation with 256KB UPD7756C internal ROM dump instead of samples
*/

void __fastcall (*jaleco_l3_bs)(int);
static void __fastcall jaleco_l3_bs0(int bank) { PRG16_BS(0,bank); }
static void __fastcall jaleco_l3_bs2(int bank) { PRG16_BS(2,bank); }

_MAPIO(jaleco_l3)
{
	/* bits 0,1,2,3: bank select */
	if (data&BIT(7)) map_jaleco.s3_prg=data<<1&MAXPRG;	/* bit 7: select 16k prg (subtype 1: low bank, subtype 2: high bank) */
	if (data&BIT(6)) map_jaleco.s3_chr=data<<3&MAXVROM;	/* bit 6: select 8k chr */
	
	/* sound: bit 5: 0=playing, 1=stop?. bit 4: 0=init sample, 1=?. address bits 0-5=sample number */
	if (data>>1&~data&BIT(4)) {
		#if DEBUG_MAPPER
		LOG(LOG_MAPPER,"%02x --- %2d\n",address,address&0x1f);
		#endif
		sample_play(address&0x1f);
	}
	
	/* bits 6 and 7 are 0: bankswitch */
	if ((data&BIN8(11000000))==0) {
		int bank=map_jaleco.s3_chr;
		CHR8_BS_V(bank);
		(*jaleco_l3_bs)(map_jaleco.s3_prg);
	}
}

_MAPINIT(jaleco_l3)
{
	SET_MAPIO8_ALL(jaleco_l3);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	
	sample_load();
	SET_MAPSTREAM(sample);
	
	switch (maptype) {
		/* subtype 1 */
		case MAPTYPE_JALECO_JF_SIMPLE_P16S:
			jaleco_l3_bs=&jaleco_l3_bs0;
			break;
		
		/* subtype 2 */
		case MAPTYPE_JALECO_JF_SIMPLE_P16US:
			map_jaleco.s3_prg=MAXPRG-1;
			jaleco_l3_bs=&jaleco_l3_bs2;
			break;
		
		default: break;
	}
}



/* --- simple P32/C8(PPPPCCCC) (similar to Nintendo GNROM, except using IO at $6000) (iNES: 66/140) ---
	Bio Senshi Dan - Increaser Tono Tatakai, Youkai Kurabu, Mississippi Satsujin Jiken *** all Japanese
*/

_MAPIO(jaleco_l4)
{
	/* low nibble: select 8k vrom bank */
	int bank=data<<3&BIN8(01111000)&MAXVROM;
	if (MAXVROM) { CHR8_BS_V(bank); }
	
	/* high nibble: select 32k prg bank */
	bank=data>>2&BIN8(00111100)&MAXPRG;
	PRG32_BS(bank);
}

_MAPINIT(jaleco_l4)
{
	SET_MAPIO6(jaleco_l4);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
}



/* --- simple P16/C8/M1(CCCCMPPP) (similar to one of Irem's simple mappers) (iNES: 78) ---
	Uchuusen - Cosmo Carrier (Japanese)
*/

_MAPIO(jaleco_l5)
{
	/* bits 0,1,2: swap 16k prgbank in $8000 */
	int bank=data<<1&MAXPRG;
	PRG16_BS(0,bank);
	
	/* bit 3: set mirroring */
	MIR_01(data&BIT(3))
	
	/* high nibble: swap 8k vrombank */
	bank=(data&0xf0)>>1&MAXVROM;
	CHR8_BS_V(bank);
}

_MAPINIT(jaleco_l5)
{
	INIT_MIR_0();
	SET_MAPIO8_ALL(jaleco_l5);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
}
