/* Konami mappers */

#include "ym2413.h" /* vrc7 sound */

static struct {
	/* vrcx */
	int d4;
	int ab[2];
	int prg[4];
	int chr[8];
	
	int prg_mask[4][3];
	int chr_mask[8][3];
	int mir_mask[1][3];
	int swap_mask[1][3];	/* vrc4 only */
	int irq_latch_mask[1][3];
	int irq_control_mask[1][3];
	int irq_ack_mask[1][3];
	
	int irq_counter;
	int irq_latch;
	int irq_divider;
	int irq_isenabled;
	int irq_restart;
	int irq_cycles;
	int irq_hit;		/* vrc3 only */
	
	int chr_shift;		/* vrc2 only */
	int vrc1_chr_hi;	/* vrc1 only */
	
	int mir_inv;		/* vrc6 only */
	int mir_and;
	int mir_shift;		/* vrc6 only */
	
	int wram_and;
	int prg_swap;		/* vrc4 only */
	
	int* dac;			/* vrc6 only */
	
	/* vrc7 sound */
	int cycles;
	float factor;
	signed short* stream;
	int bc;
	int bcb;
	float bcbf;
	int bcs;
	
} map_konami;

/* Konami VRC hardware *** all Japanese */

/* --- VRC I (and Jaleco SS 88005) (iNES: 75) ---
	Exciting Boxing, Ganbare Goemon! - Karakuri Douchuu, Jajamaru Ninpou Chou, King Kong 2 - Ikari no Megaton Punch,
	Moero!! Junior Basket - Two on Two, Tetsuwan Atom
   --- VRC IIa (iNES: 22) ---
	Ganbare Pennant Race!, TwinBee 3 - Poko Poko Dai Maou
   --- VRC IIb/VRC IVe (iNES: 23) ---
	Akumajou Special - Boku Dracula-kun, Contra, Crisis Force, Dragon Scroll - Yomigaerishi Maryuu, Ganbare Goemon 2,
	Getsufuu Maden, Jarinko Chie - Bakudan Musume no Shiawase Sagashi, Parodius da!, Tiny Toon Adventures, Wai Wai World
   --- VRC III (iNES: 73) ---
	Salamander
   --- VRC IVa/VRC IVc (iNES: 21) ---
	Ganbare Goemon Gaiden 2 - Tenka no Zaihou, Wai Wai World 2 - SOS!! Paseri Jou
   --- VRC IVb/VRC IVd (iNES: 25) ---
	Bio Miracle Bokutte Upa, Ganbare Goemon Gaiden - Kieta Ougon Kiseru, Ganbare Pennant Race! (also on VRC2), Gradius II,
	Racer Mini Yonku - Japan Cup, Teenage Mutant Ninja Turtles, Teenage Mutant Ninja Turtles II
	
	(The Turtles games are 2 and 3 outside Japan, part 1=Gekikame Ninja Den (MMC1) )
	
   --- VRC VIa (iNES: 24) ---
	Akumajou Densetsu
   --- VRC VIb (iNES: 26) ---
	Esper Dream 2 - Aratanaru Tatakai, Mouryou Senki Madara
   --- VRC VIIa/VRC VIIb (iNES: 85) ---
	Lagrange Point, Tiny Toon Adventures 2 - Montana Land e Youkoso
*/

/* TODO:
	- IRQ handling is correct? (screen bumps when going through doors in in Akumajou Densetsu, not in MMC5 port)
	- China Forever (Unl) (aka World Hero) uses iNES mapper 27? maybe it's a hacked version of the undumped Japanese version of TMNTTF..
	- vrc5 exists?
*/

static int __fastcall konami_io_nothing(u16 address,u8 data) { return FALSE; }

static __inline__ int konami_prg(u16,u8);
static __inline__ void konami_prg_bs(void);

int __fastcall (*konami_chr)(u16,u8);
static int __fastcall konami_chr_vrc1(u16,u8);
static int __fastcall konami_chr_vrc2467(u16,u8);
static __inline__ void konami_chr_bs_vrc1(void);

int __fastcall (*konami_irq_latch)(u16,u8);
int __fastcall (*konami_irq_control)(u16,u8);
int __fastcall (*konami_irq_ack)(u16,u8);
static int __fastcall konami_irq_latch_vrc3(u16,u8);
static int __fastcall konami_irq_latch_vrc467(u16,u8);
static int __fastcall konami_irq_control_vrc3467(u16,u8);
static int __fastcall konami_irq_ack_vrc3467(u16,u8);

int __fastcall (*konami_mir)(u16,u8);
static int __fastcall konami_mir_vrc12467(u16,u8);

int __fastcall (*konami_swap)(u16,u8);
static int __fastcall konami_swap_vrc4(u16,u8);
static __inline__ void konami_wram_vrc46(u8);

int __fastcall (*konami_sound)(u16,u8);
static int __fastcall konami_sound_vrc6(u16,u8);
static int __fastcall konami_sound_vrc7(u16,u8);


/* fill masks */
static void konami_fill_prg_mask(int,int);
static void konami_fill_chr_mask(int,int,int);

#define KONAMI_SET_MASK(x,y,z,c) \
	switch(c) { \
		case 0: map_konami.x##_mask[y][z]=0; break; \
		case 1: map_konami.x##_mask[y][z]=map_konami.ab[z]; break; \
		default: map_konami.x##_mask[y][z]=-1; break; \
	}
#define _KONAMI_FILL_MASK(f) \
static void konami_fill_##f##_mask(int,int,int); \
static void konami_fill_##f##_mask(int hi,int b1,int b0) \
{ \
	int i; \
	int b[2]; b[0]=b0; b[1]=b1; \
	map_konami.f##_mask[0][2]=hi; \
	for (i=0;i<2;i++) { KONAMI_SET_MASK(f,0,i,b[i]) } \
}
static void konami_fill_prg_mask(int hi,int b1)
{
	int i=4,j=0;
	
	while (i--) {
		map_konami.prg_mask[j][2]=(hi>>(i<<2))&0xf;
		KONAMI_SET_MASK(prg,j,1,(b1>>(i<<1))&3)
		KONAMI_SET_MASK(prg,j,0,-1)
		j++;
	}
}
static void konami_fill_chr_mask(int hi,int b1,int b0)
{
	int i=8,j=0,k;
	int b[2]; b[0]=b0; b[1]=b1;
	
	while (i--) {
		map_konami.chr_mask[j][2]=(hi>>(i<<2))&0xf;
		for (k=0;k<2;k++) { KONAMI_SET_MASK(chr,j,k,(b[k]>>(i<<1))&3) }
		j++;
	}
}
_KONAMI_FILL_MASK(mir)
_KONAMI_FILL_MASK(swap)
_KONAMI_FILL_MASK(irq_latch)
_KONAMI_FILL_MASK(irq_control)
_KONAMI_FILL_MASK(irq_ack)



/* io */
_MAPIO(konami)
{
	if (konami_prg(address,data)) return;
	else if ((*konami_chr)(address,data)) return;
	else if ((*konami_irq_latch)(address,data)) return;
	else if ((*konami_irq_control)(address,data)) return;
	else if ((*konami_irq_ack)(address,data)) return;
	else if ((*konami_mir)(address,data)) return;
	else if ((*konami_swap)(address,data)) return;
	else (*konami_sound)(address,data);
}

#define KONAMI_MASK(x,y) (((address>>12)==map_konami.x##_mask[y][2])&((map_konami.x##_mask[y][1]!=0)?((address&map_konami.x##_mask[y][1])!=0):((address&map_konami.ab[1])==0))&((map_konami.x##_mask[y][0]!=0)?((address&map_konami.x##_mask[y][0])!=0):((address&map_konami.ab[0])==0)))


/* prg */

/* select prg bank */
static __inline__ int konami_prg(u16 address,u8 data)
{
	int i;
	
	/* select 16k prg in $8000 */
	if (KONAMI_MASK(prg,0)) {
		i=data<<1&MAXPRG;
		map_konami.prg[0]=i;
		map_konami.prg[1]=i|1;
		
		konami_prg_bs();
		return TRUE;
	}
	
	else
	
	/* select 8k prg in $8000, $a000, or $c000 */
	for (i=3;i>0;i--)
		if (KONAMI_MASK(prg,i)) {
			map_konami.prg[i-1]=data&MAXPRG;
			
			konami_prg_bs();
			return TRUE;
		}
	return FALSE;
}
/* prg bankswitch */
static __inline__ void konami_prg_bs(void)
{
	PRG8_BS(map_konami.prg_swap,map_konami.prg[0]);
	PRG8_BS(1,map_konami.prg[1]);
	PRG8_BS(2^map_konami.prg_swap,map_konami.prg[2]);
}



/* chr */

/* vrc1: select chr bank */
static int __fastcall konami_chr_vrc1(u16 address,u8 data)
{
	switch (address&0xf000) {
		/* bit 1: high chr bit for ppu lower half, bit 2: upper half */
		case 0x9000:
			map_konami.vrc1_chr_hi=data;
			konami_chr_bs_vrc1();
			
			konami_mir_vrc12467(0x9000,data);
			return TRUE;
		
		/* select 4k chr */
		case 0xe000: case 0xf000: {
			int bank=data<<2&BIN8(00111100);
			int loc=address>>10&4;
			int i=4; while (i--) map_konami.chr[loc|i]=bank|i;
			konami_chr_bs_vrc1();
			return TRUE;
		}
		
		default: break;
	}
	return FALSE;
}
/* vrc1: chr bankswitch */
static __inline__ void konami_chr_bs_vrc1(void)
{
	int j=map_konami.vrc1_chr_hi<<5&BIT(6);
	int i=4; while (i--) { CHR1_BS_V(i,(j|map_konami.chr[i])&MAXVROM); }
	j=map_konami.vrc1_chr_hi<<4&BIT(6);
	i=4; while (i--) { CHR1_BS_V(i|4,(j|map_konami.chr[i|4])&MAXVROM); }
}
/* select 1k chr, 2*4 or 8 bits, vrom, or vram (vrc7) */
static int __fastcall konami_chr_vrc2467(u16 address,u8 data)
{
	int i=8;
	
	while (i--) {
		if (KONAMI_MASK(chr,i)) {
			if (MAXVROM) {
				if (map_konami.d4)
					if (address&map_konami.ab[0]) map_konami.chr[i]=((map_konami.chr[i]&0x0f)|((data>>map_konami.chr_shift&0xf)<<4))&MAXVROM;
					else map_konami.chr[i]=((map_konami.chr[i]&0xf0)|(data>>map_konami.chr_shift&0xf))&MAXVROM;
				else map_konami.chr[i]=data&MAXVROM;
				CHR1_BS_V(i,map_konami.chr[i]);
			}
			else { CHR1_BS(i,data&7,vram); }
			return TRUE;
		}
	}
	return FALSE;
}



/* irq */

/* vrc3: $8/9/a/b000: set irq latch, 4*4bits */
static int __fastcall konami_irq_latch_vrc3(u16 address,u8 data)
{
	switch (address&0xf000) {
		case 0x8000: case 0x9000: case 0xa000: case 0xb000: {
			int i=address>>10&BIN8(00001100);
			map_konami.irq_latch=(map_konami.irq_latch&~(0xf<<i))|((data&0xf)<<i);
			return TRUE;
		}
		
		default: break;
	}
	return FALSE;
}
/* set irq latch, either 2*4 bits, or 8 bits */
static int __fastcall konami_irq_latch_vrc467(u16 address,u8 data)
{
	if (KONAMI_MASK(irq_latch,0)) {
		if (map_konami.d4)
			if (address&map_konami.ab[0]) map_konami.irq_latch=(map_konami.irq_latch&0x0f)|(data<<4&0xf0);
			else map_konami.irq_latch=(map_konami.irq_latch&0xf0)|(data&0x0f);
		else map_konami.irq_latch=data;
		return TRUE;
	}
	return FALSE;
}
/* irq control */
static int __fastcall konami_irq_control_vrc3467(u16 address,u8 data)
{
	if (KONAMI_MASK(irq_control,0)) {
		/* bit 1: enable irqs */
		map_konami.irq_isenabled=data>>1&1;
		
		/* bit 2: vrc3: select irq hit, other: set scanline or CPU cycle counter (never seen any game set this bit though) */
		if (data&BIT(2)) { map_konami.irq_hit=0xff; map_konami.irq_divider=crystal->rel_cycle; }
		else { map_konami.irq_hit=0xffff; map_konami.irq_divider=341*crystal->ppu_cycle; }
		
		/* bit 0: enable irq restarts */
		map_konami.irq_restart=data&1;
		
		if (map_konami.irq_isenabled) {
			map_konami.irq_counter=map_konami.irq_latch;
			map_konami.irq_cycles=map_konami.irq_divider;
		}
		cpu_acknowledge_interrupt(~IRQ_MAPPER);
		return TRUE;
	}
	return FALSE;
}
/* acknowledge/restart irq */
static int __fastcall konami_irq_ack_vrc3467(u16 address,u8 data)
{
	if (KONAMI_MASK(irq_ack,0)) {
		map_konami.irq_isenabled=map_konami.irq_restart;
		cpu_acknowledge_interrupt(~IRQ_MAPPER);
		return TRUE;
	}
	return FALSE;
}

/* vrc4/6/7 irq handler, based on cpu cycles, clocked every 113 2/3 cycles (1 scanline), an irq is fired on 8 bit counter overflow */
_MAPIRQ(konami_vrc467)
{
	if (map_konami.irq_isenabled) {
		map_konami.irq_cycles-=(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr);
		if (map_konami.irq_cycles<=0) {
			if (++map_konami.irq_counter==256) {
				map_konami.irq_counter=map_konami.irq_latch;
				cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(map_konami.irq_cycles>-crystal->rel_cycle?INT_EDGE_IRQ_EX:0));
			}
			map_konami.irq_cycles+=map_konami.irq_divider;
		}
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}
/* vrc3 irq handler, based on cpu cycles, 16 bit or 8 bit */
_MAPIRQ(konami_vrc3)
{
	if (map_konami.irq_isenabled) {
		
		int cpu_cycles;
		
		for (cpu_cycles=*mapper.cpu_cycles_ptr;mapper.cpu_cycles_old>cpu_cycles;mapper.cpu_cycles_old-=crystal->rel_cycle) {
			map_konami.irq_counter++;
			if (!(map_konami.irq_counter&map_konami.irq_hit)) {
				map_konami.irq_counter=map_konami.irq_latch;
				cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(mapper.cpu_cycles_old-cpu_cycles?0:INT_EDGE_IRQ_EX));
			}
		}
	}
	
	else mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}



/* mirroring */
static int __fastcall konami_mir_vrc12467(u16 address,u8 data)
{
	if (KONAMI_MASK(mir,0)) {
		/* 2 or 1 bits selection */
		MIR_VH01(data>>map_konami.mir_shift&map_konami.mir_and)
		
		/* vrc6: bit 5 inverts mirroring? results in garbage in Esper Dream 2 in-game menu */
		/*if (data&map_konami.mir_inv) { MIR_C(~cartridge->mirroring); }*/
		
		/* vrc6: bit 7: enable/disable wram */
		if (map_konami.wram_and&BIT(7)) konami_wram_vrc46(data);
		return TRUE;
	}
	return FALSE;
}


/* vrc4: swap prg0 with prg2 */
static int __fastcall konami_swap_vrc4(u16 address,u8 data)
{
	if (KONAMI_MASK(swap,0)) {
		/* bit 2: set swap */
		map_konami.prg_swap=data&2;
		konami_prg_bs();
		
		/* bit 0: enable/disable wram */
		if (map_konami.wram_and&BIT(0)) konami_wram_vrc46(data);
		return TRUE;
	}
	return FALSE;
}


/* enable/disable wram */
static __inline__ void konami_wram_vrc46(u8 data)
{
	if (data&map_konami.wram_and) cartridge->wram_isenabled=TRUE;
	else cartridge->wram_isenabled=FALSE;
}



/* vrc6 sound, used by all (3) vrc6 games */
/* 2 square channels, one saw channel .. (sound test in Akumajou Densetsu=a|b+start on titlescreen. sound test in madara=d,d,u,u,r,l,r,l,a,b on titlescreen, then up/down, b/a. ) */

static void konami_sound_vrc6_init(void);

typedef struct {
	int cycles;
	int* buffer;
	int buffer_counter;
	int sample;
	int timer_raw;
	int timer;
	int enabled;
	int duty;			/* or phase for saw */
	int duty_counter;	/* or phase counter for saw */
	int volume;
	int gate;
	u8 a;				/* saw accumulator */
	
} konami_vrc6_channel;

static konami_vrc6_channel* konami_vrc6_square0;
static konami_vrc6_channel* konami_vrc6_square1;
static konami_vrc6_channel* konami_vrc6_saw;

static void __fastcall konami_vrc6_square0_update(void);
static void __fastcall konami_vrc6_square1_update(void);
static void __fastcall konami_vrc6_saw_update(void);

static int konami_vrc6_saw_volume_lookup[32];

/* duty waveform for square channels (>=%1000=digitized mode) */
static const int konami_vrc6_duty_waveform[8][16]={
	{ 0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0 }, /* 1 */
	{ 0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0 }, /* 2 */
	{ 0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0 }, /* 3 */
	{ 0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0 }, /* 4 */
	{ 0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0 }, /* 5 */
	{ 0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0 }, /* 6 */
	{ 0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0 }, /* 7 */
	{ 0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0 }, /* 8 */
};

/* timer registers. $9/a/b, bit 0 or 1 set (depending on addressline configuration). low=8 bits, high=lower 4 bits, bit 7=channel enable */
#define VRC6_TIMER_LOW_REGISTER(x) x##_update(); x->timer_raw=(x->timer_raw&0xf00)|data
#define VRC6_TIMER_HIGH_REGISTER(x) x##_update(); x->timer_raw=(x->timer_raw&0xff)|(data<<8&0xf00); x->enabled=data>>7

/* duty/volume registers, squares only. $9/a, bit 0 and 1 clear */
/* lower nibble=volume, upper nibble=duty (last bit being the gate) */
#define VRC6_DUTY_REGISTER(x) x##_update(); x->duty=data>>4&7; x->gate=data>>7; x->volume=data&0xf

/* $b000: phase register, saw only. bits 0-5: set phase constant add value */
#define VRC6_PHASE_REGISTER(x) x##_update(); x->duty=data&BIN8(00111111);

/* IO */
static int __fastcall konami_sound_vrc6(u16 address,u8 data)
{
	/* handle addresslines this way instead; mapper 26 and 24 bits (0 and 1) are crossed */
	address=(address&~3)|(((address&BIT(0))<<(map_konami.ab[0]>>1))|((address&BIT(1))>>(map_konami.ab[0]>>1)));
	
	switch (address&0xf003) {
		
		/* squares */
		case 0x9000: VRC6_DUTY_REGISTER(konami_vrc6_square0); break;
		case 0xa000: VRC6_DUTY_REGISTER(konami_vrc6_square1); break;
		case 0x9001: VRC6_TIMER_LOW_REGISTER(konami_vrc6_square0); break;
		case 0xa001: VRC6_TIMER_LOW_REGISTER(konami_vrc6_square1); break;
		case 0x9002: VRC6_TIMER_HIGH_REGISTER(konami_vrc6_square0); break;
		case 0xa002: VRC6_TIMER_HIGH_REGISTER(konami_vrc6_square1); break;
		
		/* saw */
		case 0xb000: VRC6_PHASE_REGISTER(konami_vrc6_saw); break;
		case 0xb001: VRC6_TIMER_LOW_REGISTER(konami_vrc6_saw); break;
		case 0xb002: VRC6_TIMER_HIGH_REGISTER(konami_vrc6_saw); break;
		
		default: break;
	}
	
	return FALSE; /* don't care */
}


/* reset */
_MAPRESET(konami_vrc6)
{
	konami_vrc6_square0->enabled=konami_vrc6_square1->enabled=konami_vrc6_saw->enabled=FALSE; /* silence */
}

/* new frame */
_MAPNF(konami_vrc6)
{
	mapper_new_frame_updateoldcycles(); /* std */
	konami_vrc6_square0->cycles=konami_vrc6_square1->cycles=konami_vrc6_saw->cycles=cpu_get_cycles();
}

/* end frame */
_MAPEF(konami_vrc6)
{
	int i;
	
	konami_vrc6_square0_update();
	konami_vrc6_square1_update();
	konami_vrc6_saw_update();
	
	/* fill dac */
	i=konami_vrc6_saw->buffer_counter;
	while (i--) map_konami.dac[i]=konami_vrc6_saw_volume_lookup[konami_vrc6_saw->buffer[i]]+ac_square_volume_lookup[konami_vrc6_square0->buffer[i]+konami_vrc6_square1->buffer[i]];
}

_MAPKILLCYCLE(konami_vrc6)
{
	konami_vrc6_square0->cycles-=crystal->ppu_cycle;
	konami_vrc6_square1->cycles-=crystal->ppu_cycle;
	konami_vrc6_saw->cycles-=crystal->ppu_cycle;
}

_MAPSOUNDBC(konami_vrc6) { konami_vrc6_square0->buffer_counter=konami_vrc6_square1->buffer_counter=konami_vrc6_saw->buffer_counter=bc; }


/* square channels clocking */
#define KONAMI_VRC6_SQUARE_UPDATE(x) \
	int cpu_cycles; \
	\
	for (cpu_cycles=cpu_get_cycles();x->cycles>cpu_cycles;x->cycles-=crystal->rel_cycle) { \
		if (!--x->timer) { \
			x->timer=x->timer_raw+1; \
			x->duty_counter=(x->duty_counter+1)&0xf; \
			x->sample=((x->gate|((x->timer_raw>8)&konami_vrc6_duty_waveform[x->duty][x->duty_counter]))&x->enabled)*x->volume; \
		} \
		x->buffer[x->buffer_counter++]=x->sample; \
	}

static void __fastcall konami_vrc6_square0_update(void) { KONAMI_VRC6_SQUARE_UPDATE(konami_vrc6_square0) }
static void __fastcall konami_vrc6_square1_update(void) { KONAMI_VRC6_SQUARE_UPDATE(konami_vrc6_square1) }

/* saw channel clocking */
static void __fastcall konami_vrc6_saw_update(void)
{
	int cpu_cycles;
	
	for (cpu_cycles=cpu_get_cycles();konami_vrc6_saw->cycles>cpu_cycles;konami_vrc6_saw->cycles-=crystal->rel_cycle) {
		if (!--konami_vrc6_saw->timer) {
			konami_vrc6_saw->timer=konami_vrc6_saw->timer_raw+1;
			konami_vrc6_saw->a+=konami_vrc6_saw->duty;
			if (!--konami_vrc6_saw->duty_counter) {
				konami_vrc6_saw->duty_counter=14;
				konami_vrc6_saw->a=0;
			}
			konami_vrc6_saw->sample=(konami_vrc6_saw->a>>3)*(konami_vrc6_saw->enabled&(konami_vrc6_saw->timer_raw>6));
		}
		konami_vrc6_saw->buffer[konami_vrc6_saw->buffer_counter++]=konami_vrc6_saw->sample;
	}
}


/* clean */
_MAPCLEAN(konami_vrc6)
{
	MEM_CLEAN(konami_vrc6_square0);
	MEM_CLEAN(konami_vrc6_square1);
	MEM_CLEAN(konami_vrc6_saw);
}

/* init */
static void konami_sound_vrc6_init(void)
{
	int i;
	
	#define KONAMI_VRC6_CREATE_CHANNEL(x) MEM_CREATE(x,sizeof(konami_vrc6_channel)); x->buffer=sound_create_channel(); x->timer=1
	KONAMI_VRC6_CREATE_CHANNEL(konami_vrc6_square0);
	KONAMI_VRC6_CREATE_CHANNEL(konami_vrc6_square1);
	KONAMI_VRC6_CREATE_CHANNEL(konami_vrc6_saw); konami_vrc6_saw->duty_counter=14;
	
	konami_sound=&konami_sound_vrc6;
	SET_MAPCLEAN(konami_vrc6);
	SET_MAPRESET(konami_vrc6);
	SET_MAPNF(konami_vrc6);
	SET_MAPEF(konami_vrc6);
	SET_MAPKILLCYCLE(konami_vrc6);
	SET_MAPSOUNDBC(konami_vrc6);
	map_konami.dac=sound_create_dac();
	ac_fill_volume_lookup(); /* square channels */
	
	for (i=0;i<32;i++) konami_vrc6_saw_volume_lookup[i]=i*72;
}



/*
	vrc7 sound, a stripped down YM2413; 6 channels, no rhythm, and with a different instrument set
	too bad that it's only used in 1 game: Lagrange Point
	
	sound updating is different, as it doesn't write 1 sample per cpu cycle
*/

static void konami_sound_vrc7_init(void);
static void konami_vrc7_reset_factor(void);
static void __fastcall konami_vrc7_sound_update(void);

static void konami_vrc7_reset_factor(void)
{
	map_konami.factor=(((float)crystal->sub)/((float)crystal->base))+((crystal->slice[0]<crystal->slice[1])?crystal->slice[0]:crystal->slice[1]);
}

/* IO */
static int __fastcall konami_sound_vrc7(u16 address,u8 data)
{
	switch (address&0xf030) {
		
		/* index */
		case 0x9010:
			data&=0x3f;
			switch (data) { /* not all registers are available */
				case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7: /* custom */
				case 0x10: case 0x11: case 0x12: case 0x13: case 0x14: case 0x15: /* frequency */
				case 0x20: case 0x21: case 0x22: case 0x23: case 0x24: case 0x25: /* etc */
				case 0x30: case 0x31: case 0x32: case 0x33: case 0x34: case 0x35: /* instrument/volume */
					konami_vrc7_sound_update();
					ym2413_write_address(data);
					break;
				
				default: break;
			}
			break;
		
		/* data */
		case 0x9030:
			konami_vrc7_sound_update();
			ym2413_write_data(data);
			break;
		
		default: break;
	}
	
	return FALSE; /* don't care */
}

_MAPRESET(konami_vrc7) { ym2413_reset(); }
_MAPMODE(konami_vrc7) { konami_vrc7_reset_factor(); }
_MAPKILLCYCLE(konami_vrc7) { map_konami.cycles-=crystal->ppu_cycle; }

/* new frame */
_MAPNF(konami_vrc7)
{
	mapper_new_frame_updateoldcycles(); /* std */
	map_konami.cycles=cpu_get_cycles();
	
	map_konami.bcbf=map_konami.factor;
	map_konami.bcb=map_konami.bcbf+0.5;
	map_konami.bc=0;
}

/* update */
static void __fastcall konami_vrc7_sound_update(void)
{
	const int cpu_cycles=cpu_get_cycles();
	const int bcs_old=map_konami.bcs;
	
	for (;map_konami.cycles>cpu_cycles;map_konami.cycles-=crystal->rel_cycle) {
		map_konami.bc++;
		if (map_konami.bc>map_konami.bcb) {
			map_konami.bcbf+=map_konami.factor;
			map_konami.bcb=map_konami.bcbf+0.5;
			map_konami.bcs++;
		}
	}
	
	if (map_konami.bcs-bcs_old) ym2413_update(&map_konami.stream[bcs_old],map_konami.bcs-bcs_old);
}

/* sound stream */
_MAPSTREAM(konami_vrc7)
{
	int i=length;
	
	if (i==0) return;
	
	if ((i-map_konami.bcs)>0) {
		ym2413_update(&map_konami.stream[map_konami.bcs],i-map_konami.bcs);
		map_konami.bcs=0;
	}
	else {
		map_konami.bcs-=length;
		#if DEBUG_MAPPER
		LOG(LOG_MISC,"o");
		#endif
	}
	
	/* update stream */
	while (i--)
		stream[i]=stream[i]+map_konami.stream[i]*2;
	
	/* on overflow, copy last part back to start */
	if (map_konami.bcs)
		for (i=length;i<(map_konami.bcs+length);i++)
			map_konami.stream[i-length]=map_konami.stream[i];
}

/* clean */
_MAPCLEAN(konami_vrc7)
{
	ym2413_clean();
	MEM_CLEAN(map_konami.stream);
}

/* init */
static void konami_sound_vrc7_init(void)
{
	konami_sound=&konami_sound_vrc7;
	ym2413_init();
	konami_vrc7_reset_factor();
	
	MEM_CREATE(map_konami.stream,sound_get_buffer_size()); /* more than enough */
	
	SET_MAPMODE(konami_vrc7);
	SET_MAPSTREAM(konami_vrc7);
	SET_MAPCLEAN(konami_vrc7);
	SET_MAPRESET(konami_vrc7);
	SET_MAPNF(konami_vrc7);
	SET_MAPKILLCYCLE(konami_vrc7);
}



/* main init */
_MAPINIT(konami)
{
	int i;
	map_konami.prg[0]=0; map_konami.prg[1]=1; map_konami.prg[2]=MAXPRG-1; map_konami.prg[3]=MAXPRG;
	for (i=0;i<8;i++) map_konami.chr[i]=i;
	
	DEF_6000_WRAM();
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	INIT_MIR_V();
	SET_MAPIO8_ALL(konami);
	
	konami_swap=konami_chr=konami_mir=konami_irq_latch=konami_irq_control=konami_irq_ack=konami_sound=&konami_io_nothing;
	
	/* addresslines */
	#define KONAMI_AB(x,y) map_konami.ab[0]=x; map_konami.ab[1]=y
	switch (maptype) {
		case MAPTYPE_KONAMI_VRC4A_VRC4C:		KONAMI_AB(BIT(1)|BIT(6),BIT(2)|BIT(7));	break;
		case MAPTYPE_KONAMI_VRC2A:				KONAMI_AB(BIT(1),BIT(0));				break;
		case MAPTYPE_KONAMI_VRC2B_VRC4E:		KONAMI_AB(BIT(0)|BIT(2),BIT(1)|BIT(3));	break;
		case MAPTYPE_KONAMI_VRC6A:				KONAMI_AB(BIT(0),BIT(1));				break;
		case MAPTYPE_KONAMI_VRC4B_VRC4D:		KONAMI_AB(BIT(3)|BIT(1),BIT(2)|BIT(0));	break;
		case MAPTYPE_KONAMI_VRC6B:				KONAMI_AB(BIT(1),BIT(0));				break;
		case MAPTYPE_KONAMI_VRC3: 														break;
		case MAPTYPE_KONAMI_VRC1_JALECO_SS_88005: 										break;
		case MAPTYPE_KONAMI_VRC7A_VRC7B:		KONAMI_AB(BIT(5),BIT(4)|BIT(3));		break;
		
		default: break;
	}
	
	/* prg, swap, wram, chr, mir, irq, (sound) */
	#define KONAMI_INIT_CHR(a,b,c) konami_fill_chr_mask(a,b,c); konami_chr=&konami_chr_vrc2467
	#define KONAMI_INIT_MIR(a,b,c) konami_fill_mir_mask(a,b,c); konami_mir=&konami_mir_vrc12467; map_konami.mir_and=3
	#define KONAMI_INIT_IRQ(a1,a2,a3,b1,b2,b3,c1,c2,c3) SET_MAPIRQ(konami_vrc467); SET_MAPNF(updateoldcycles); konami_irq_latch=&konami_irq_latch_vrc467; konami_irq_control=&konami_irq_control_vrc3467; konami_irq_ack=&konami_irq_ack_vrc3467; konami_fill_irq_latch_mask(a1,a2,a3); konami_fill_irq_control_mask(b1,b2,b3); konami_fill_irq_ack_mask(c1,c2,c3)
	switch (maptype) {
		/* VRC I */
		case MAPTYPE_KONAMI_VRC1_JALECO_SS_88005:
			konami_fill_prg_mask(0x08ac,0xaa);
			konami_chr=&konami_chr_vrc1;
			KONAMI_INIT_MIR(9,-1,-1); map_konami.mir_and=1;
			break;
		
		/* VRC II */
		case MAPTYPE_KONAMI_VRC2A:
			konami_fill_prg_mask(0x08a0,0xaa);
			KONAMI_INIT_CHR(0xbbccddee,0x1111,0xaaaa); map_konami.d4=TRUE; map_konami.chr_shift=1;
			KONAMI_INIT_MIR(9,0,0);
			break;
		
		/* VRC III */
		case MAPTYPE_KONAMI_VRC3:
			konami_fill_prg_mask(0xf000,0xaa);
			KONAMI_INIT_IRQ(0,-1,-1, 0xc,-1,-1, 0xd,-1,-1); SET_MAPIRQ(konami_vrc3); konami_irq_latch=&konami_irq_latch_vrc3; map_konami.irq_hit=0xffff;
			break;
		
		/* VRC IV */
		case MAPTYPE_KONAMI_VRC4A_VRC4C: case MAPTYPE_KONAMI_VRC2B_VRC4E: case MAPTYPE_KONAMI_VRC4B_VRC4D:
			konami_fill_prg_mask(0x08a0,0xaa);
			konami_swap=&konami_swap_vrc4; konami_fill_swap_mask(9,1,0);
			map_konami.wram_and=BIT(0);
			KONAMI_INIT_CHR(0xbbccddee,0x1111,0xaaaa); map_konami.d4=TRUE;
			KONAMI_INIT_MIR(9,0,0);
			KONAMI_INIT_IRQ(0xf,0,-1, 0xf,1,0, 0xf,1,1);
			break;
		
		/* VRC VI */
		case MAPTYPE_KONAMI_VRC6A: case MAPTYPE_KONAMI_VRC6B:
			konami_fill_prg_mask(0x800c,0xaa);
			map_konami.wram_and=BIT(7);
			KONAMI_INIT_CHR(0xddddeeee,0x0505,0x1111);
			KONAMI_INIT_MIR(0xb,1,1); map_konami.mir_shift=2; map_konami.mir_inv=BIT(5);
			KONAMI_INIT_IRQ(0xf,0,0, 0xf,0,1, 0xf,1,0);
			konami_sound_vrc6_init();
			break;
		
		/* VRC VII */
		case MAPTYPE_KONAMI_VRC7A_VRC7B:
			konami_fill_prg_mask(0x0889,0x84);
			KONAMI_INIT_CHR(0xaabbccdd,0x1111,0xaaaa);
			KONAMI_INIT_MIR(0xe,0,-1);
			KONAMI_INIT_IRQ(0xe,1,-1, 0xf,0,-1, 0xf,1,-1);
			konami_sound_vrc7_init();
			break;
		
		default: break;
	}
}
