/* Miscellaneous mappers */

static struct {
	/* caltron */
	int caltron8_enabled;
	int caltron_chr;
	
} map_misc;


/* --- Caltron 6-in-1 (accordingly, it's licensed) (iNES: 41) ---
	Caltron 6 in 1
		Cosmos Cop, Magic Carpet 1001, Balloon Monster, Adam and Eve, Porter, Bookyman
*/

static __inline__ void caltron_chr_bs(void)
{
	int bank=map_misc.caltron_chr<<3&MAXVROM;
	CHR8_BS_V(bank);
}

/* $6000-$67ff */
_MAPIO(caltron6)
{
	if (address<0x6800) {
		int bank;
		
		/* address bits 0,1,2: select 32k prg */
		bank=address<<2&MAXPRG;
		PRG32_BS(bank);
		
		/* address bit 2: enable $8000 register */
		map_misc.caltron8_enabled=address&BIT(2);
		
		/* address bits 3,4: upper 2 bits of chr bank */
		map_misc.caltron_chr&=3;
		map_misc.caltron_chr|=(address>>1&BIN8(00001100));
		caltron_chr_bs();
		
		/* address bit 5: mirroring */
		MIR_VH(address&BIT(5))
	}
}

/* $8000-$ffff */
_MAPIO(caltron8)
{
	if (map_misc.caltron8_enabled) {
		/* bits 0,1: lower 2 bits of chr bank */
		map_misc.caltron_chr&=BIN8(00001100);
		map_misc.caltron_chr|=(data&3);
		caltron_chr_bs();
	}
}

/* reset: registers are set to 0 */
_MAPRESET(caltron)
{
	map_misc.caltron8_enabled=TRUE;
	MAPIO_FUN(caltron8)(0x8000,0,0);
	MAPIO_FUN(caltron6)(0x6000,0,0);
}

_MAPINIT(caltron)
{
	SET_MAPIO6(caltron6);
	SET_MAPIO8_ALL(caltron8);
	SET_MAPRESET(caltron); /* when reset, it initially sets chr and prg */
}





/* SIMPLE TYPES */


/* --- Capcom simple P16(xxxPPPxx) (iNES: 94) ---
	Senjou no Ookami (Japanese version of Commando)
*/

_MAPIO(capcom_l1)
{
	/* swap 16k prgbank in $8000 (from bit 2) */
	data=((data>>1)&BIN8(11111110))&MAXPRG;
	PRG16_BS(0,data);
}

_MAPINIT(capcom_l1)
{
	SET_MAPIO8_ALL(capcom_l1);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
}



/* --- Nichibutsu simple P16u(xxxxxPPP) (iNES: 180) ---
	Crazy Climber (Japanese)
*/

_MAPIO(nichibutsu_l1)
{
	/* swap 16k prgbank in $c000 */
	int bank=data<<1&MAXPRG;
	PRG16_BS(2,bank);
}

_MAPINIT(nichibutsu_l1)
{
	SET_MAPIO8_ALL(nichibutsu_l1);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
}
