/* Nintendo MMC1 mappers */

static struct {
	/* all */
	int reg[4];
	int serial;
	int serial_counter;
	int regnum_old;
	int writecycle;
	
	/* sorom/sxrom */
	u8* wram;
	
	/* nes-event */
	int event_dips;
	int event_isenabled;
	int event_irq_isenabled;
	int event_irq_counter;
	int event_irq_hit;
	
} map_mmc1;


/* --- MMC1:MMC1A/MMC1B1/MMC1B2/MMC1C, NES-SAROM/NES-SBROM/NES-SCROM/NES-SEROM/NES-SFROM/NES-SGROM/NES-SHROM/NES-SJROM/NES-SKROM/NES-SLROM/NES-SL1ROM/NES-SL2ROM/NES-SL3ROM/NES-SLRROM/NES-SNROM/NES-SUROM(512K PRG) (iNES: 1) ---
	Legend of Zelda 1,2, Metroid, Turtles 1, Tetris, Megaman 2, Final Fantasy, Snake Rattle n Roll,
	Double Dragon, Chip n Dale 1,2, Castlevania 2, Ninja Gaiden, Dragon Warrior 1,2,3,4, etc.
	
   --- MMC1A (without WRAM protection) (iNES: 155)
	Tatakae!! Rahmen Man - Sakuretsu Choujin 102 Gei, The Money Game
	
   --- NES-SOROM(2*8K WRAM) (iNES: no(but possible with 16K RAM pages detection)/1) ---
	Aoki Ookami to Shiroki Mejika - Genghis Khan (J)/Genghis Khan (U), Nobunaga no Yabou - Zenkoku Han (J)/Nobunaga's Ambition (U),
	Sangokushi (J)/Romance of the Three Kingdoms (U)
	
   --- HVC-SXROM(32K WRAM,512K PRG) (iNES: no(but possible with 32K RAM pages detection)/1) ---
	Final Fantasy I & II, Best Play Pro Yakyuu Special
	
   --- NES-EVENT (iNES: 105)
	Nintendo World Championships 1990
*/

/* TODO:
	- SUROM high bank/SOROM WRAM bank can change on PPU addressbus accesses when used bits are unequal to eachother.
	  I'll add it when I see it's used somewhere (I doubt it), it'll cause a big speed decrease.
*/

static void __fastcall mmc1_control(void); static void __fastcall mmc1_pat0bs(void); static void __fastcall mmc1_pat1bs(void); static void __fastcall mmc1_prgbs(void);
static void __fastcall mmc1_sorom_reg1(void);	static void __fastcall mmc1_sorom_reg2(void);
static void __fastcall mmc1_surom_reg1(void);	static void __fastcall mmc1_surom_reg2(void);
static void __fastcall mmc1_sxrom_reg1(void);	static void __fastcall mmc1_sxrom_reg2(void);
static void __fastcall mmc1_event_reg0(void);	static void __fastcall mmc1_event_reg1(void);	static void __fastcall mmc1_event_reg3(void);
static void __fastcall mmc1_nothing(void);
typedef void __fastcall (*fpt_mmc1_io)(void);
static const fpt_mmc1_io mmc1_io_std[4]= { mmc1_control, mmc1_pat0bs, mmc1_pat1bs, mmc1_prgbs };
static fpt_mmc1_io mmc1_io[4];

void __fastcall (*mmc1_wram_protection)(void);
static void __fastcall mmc1_wram_protection_no(void) { cartridge->wram_isenabled=TRUE; }
static void __fastcall mmc1_wram_protection_yes(void) { cartridge->wram_isenabled=map_mmc1.reg[3]>>4^1; }

/* IO, fill serial data */
_MAPNF(mmc1)
{
	mapper_new_frame_updateoldcycles(); /* std (actually only for irq) */
	map_mmc1.writecycle=crystal->frame<<1; /* invalidate writecycle */
}

_MAPIO(mmc1)
{
	if (*mapper.cpu_cycles_ptr==map_mmc1.writecycle) return; /* ignore rmw 2nd write when reset */
	address=address>>13&3;
	if (map_mmc1.regnum_old!=address) map_mmc1.serial_counter=0;
	map_mmc1.regnum_old=address;
	map_mmc1.serial=map_mmc1.serial>>1|(data<<4&BIT(4));
	
	if (data&0x80) {
		/* reset */
		map_mmc1.writecycle=*mapper.cpu_cycles_ptr;
		map_mmc1.serial_counter=0;
		map_mmc1.reg[0]|=BIN8(00001100);
		mmc1_io[0]();
	}
	else if (++map_mmc1.serial_counter==5) {
		map_mmc1.serial_counter=0;
		map_mmc1.reg[address]=map_mmc1.serial;
		mmc1_io[address]();
	}
}

/* registers */

static void __fastcall mmc1_nothing(void) { return; }

/* register 0: control register: $8000-$9fff */
#define MMC1_MIRRORING() MIR_01VH(map_mmc1.reg[0]&3)
static void __fastcall mmc1_control(void)
{
	MMC1_MIRRORING()
	mmc1_io[3](); mmc1_io[1](); mmc1_io[2]();
}

/* register 1: pattern page 0 switch: $a000-$bfff */
static void __fastcall mmc1_pat0bs(void)
{
	int bank;
	
	#define MMC1_PAT0BS(b,v) \
	/* 4k */ \
	if (map_mmc1.reg[0]&BIT(4)) { \
		bank=map_mmc1.reg[1]<<2&b; \
		CHR4_BS(0,bank,v); \
	} \
	/* 8k */ \
	else { \
		bank=map_mmc1.reg[1]<<2&~7&b; \
		CHR8_BS(bank,v); \
	}
	
	if (MAXVROM) { MMC1_PAT0BS(MAXVROM,vrom); } /* vrom */
	else { MMC1_PAT0BS(7,vram); } /* vram */
}

/* register 2: pattern page 1 switch: $c000-$dfff */
static void __fastcall mmc1_pat1bs(void)
{
	int bank;
	
	#define MMC1_PAT1BS(b,v) bank=map_mmc1.reg[2]<<2&b; CHR1_BS(4,bank,v); CHR1_BS(5,bank|1,v); CHR1_BS(6,bank|2,v); CHR1_BS(7,bank|3,v)
	
	if (map_mmc1.reg[0]&BIT(4)) { /* 4k only */
		if (MAXVROM) { MMC1_PAT1BS(MAXVROM,vrom); } /* vrom */
		else { MMC1_PAT1BS(7,vram); } /* vram */
	}
}

/* register 3: prg switch: $e000-$ffff */
static void __fastcall mmc1_prgbs(void)
{
	int bank,hi;
	(*mmc1_wram_protection)();
	
	hi=(map_mmc1.reg[1]|(map_mmc1.reg[0]&map_mmc1.reg[2]))<<1&BIT(5);
	
	if (map_mmc1.reg[0]&BIT(3)) { /* 16k */
		bank=((map_mmc1.reg[3]<<1&0x1f)|hi)&MAXPRG;
		if (map_mmc1.reg[0]&BIT(2)) { /* at $8000 */
			PRG8_BS(2,(0x1e|hi)&MAXPRG); PRG8_BS(3,(0x1f|hi)&MAXPRG);
			PRG16_BS(0,bank);
		}
		else { /* at $c000 */
			PRG8_BS(0,hi&MAXPRG); PRG8_BS(1,(1|hi)&MAXPRG);
			PRG16_BS(2,bank);
		}
	}
	else { /* 32k */
		bank=((map_mmc1.reg[3]<<1&BIN8(00011100))|hi)&MAXPRG;
		PRG32_BS(bank);
	}
}

/* misc std */

_MAPCLEAN(mmc1_wram)
{
	MEM_CLEAN(map_mmc1.wram);
}


/* MMC1 NES-SOROM */

static __inline__ void mmc1_sorom_wram_bs(void)
{
	/* bit 3 defines WRAM bank */
	cartridge->cur_wrambank=map_mmc1.wram+((map_mmc1.reg[1]|(map_mmc1.reg[0]>>1&map_mmc1.reg[2]))<<10&0x2000);
}

_MAPBATSAVE(mmc1_sorom)
{
	cartridge->cur_wrambank=map_mmc1.wram+0x2000;
	return FALSE;
}

/* registers */
static void __fastcall mmc1_sorom_reg1(void) { mmc1_pat0bs(); mmc1_sorom_wram_bs(); }
static void __fastcall mmc1_sorom_reg2(void) { mmc1_pat1bs(); mmc1_sorom_wram_bs(); }


/* MMC1 NES-SUROM */

/* registers */
static void __fastcall mmc1_surom_reg1(void) { mmc1_prgbs(); mmc1_pat0bs(); }
static void __fastcall mmc1_surom_reg2(void) { mmc1_prgbs(); mmc1_pat1bs(); }


/* MMC1 HVC-SXROM */

static __inline__ void mmc1_sxrom_wram_bs(void)
{
	/* bits 2,3 define WRAM bank */
	if (map_mmc1.reg[0]&BIT(4)) cartridge->cur_wrambank=map_mmc1.wram+((map_mmc1.reg[1]|map_mmc1.reg[2])<<11&0x6000); /* 4k */
	else cartridge->cur_wrambank=map_mmc1.wram+(map_mmc1.reg[1]<<11&0x6000); /* 8k */
}

_MAPBATSAVE(mmc1_sxrom)
{
	cartridge->cur_wrambank=map_mmc1.wram;
	return FALSE;
}

/* registers */
static void __fastcall mmc1_sxrom_reg1(void) { mmc1_prgbs(); mmc1_pat0bs(); mmc1_sxrom_wram_bs(); }
static void __fastcall mmc1_sxrom_reg2(void) { mmc1_prgbs(); mmc1_pat1bs(); mmc1_sxrom_wram_bs(); }


/* MMC1 NES-EVENT */

/* registers */
static void __fastcall mmc1_event_reg0(void) { MMC1_MIRRORING() mmc1_event_reg3(); }
static void __fastcall mmc1_event_reg1(void)
{
	if (map_mmc1.reg[1]&BIT(4)) {
		/* enable the cartridge: write 1, then 0 (bit 4) */
		if (map_mmc1.event_isenabled==0) map_mmc1.event_isenabled=TRUE;
		
		/* reset irq */
		else if (map_mmc1.event_isenabled==TRUE) {
			cpu_acknowledge_interrupt(~IRQ_MAPPER);
			map_mmc1.event_irq_isenabled=FALSE;
			map_mmc1.event_irq_counter=0;
		}
	}
	else {
		if (map_mmc1.event_isenabled==2) map_mmc1.event_isenabled=0;
		else if (map_mmc1.event_isenabled==TRUE) map_mmc1.event_irq_isenabled=TRUE; /* enable irq */
	}
	
	if (!(map_mmc1.reg[1]&BIT(3))) {
		/* bankswitch lower bank */
		int bank=0;
		if (map_mmc1.event_isenabled==TRUE) bank=map_mmc1.reg[1]<<1&BIN8(00001100)&MAXPRG;
		
		PRG32_BS(bank);
	}
	else mmc1_event_reg3();
}

static void __fastcall mmc1_event_reg3(void)
{
	if (map_mmc1.event_isenabled==TRUE&&map_mmc1.reg[1]&BIT(3)) {
		/* bankswitch upper bank */
		int bank,i=4;
		mmc1_prgbs();
		
		while (i--) { bank=(cartridge->cur_prgbank_num[i]|0x10)&MAXPRG; PRG8_BS(i,bank); }
	}
	else (*mmc1_wram_protection)();
}

/* irq: based on cpu cycles, 30 bit counter */
_MAPIRQ(mmc1_event)
{
	if (map_mmc1.event_irq_isenabled) {
		
		int cpu_cycles;
		
		for (cpu_cycles=*mapper.cpu_cycles_ptr;mapper.cpu_cycles_old>cpu_cycles;mapper.cpu_cycles_old-=crystal->rel_cycle) {
			map_mmc1.event_irq_counter=(map_mmc1.event_irq_counter+1)&0x3fffffff;
			if ((map_mmc1.event_irq_counter&map_mmc1.event_irq_hit)==map_mmc1.event_irq_hit) cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(mapper.cpu_cycles_old-cpu_cycles?0:INT_EDGE_IRQ_EX));
		}
	}
	
	else mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

static void mmc1_event_reset_irq_hit(void) { map_mmc1.event_irq_hit=(mapper_switches.nwc_dips|0x10)<<25; }

/* dips */
_MAPSWITCHINIT(mmc1_event)
{
	int i=BIN8(00000100); /* the default was ~6mins 15 seconds */
	char* c=NULL;
	
	/* (copy/pasted from Kevin Horton's document)
		
		Dip switch settings...
		
		switch|time (mins)
		------------------
		OOOO - 5.001
		OOOC - 5.316
		OOCO - 5.629
		OOCC - 5.942
		OCOO - 6.254
		OCOC - 6.567
		OCCO - 6.880
		OCCC - 7.193
		COOO - 7.505
		COOC - 7.818
		COCO - 8.131
		COCC - 8.444
		CCOO - 8.756
		CCOC - 9.070
		CCCO - 9.318
		CCCC - 9.695
		
		O = switch open ("off"), C = switch closed ("on")
	*/
	
	SETTINGS_GET_STRING(settings_info(SETTINGS_NWCDIPS),&c);
	if (c&&strlen(c)==4&&((c[0]=='0')|(c[0]=='1'))&&((c[1]=='0')|(c[1]=='1'))&&((c[2]=='0')|(c[2]=='1'))&&((c[3]=='0')|(c[3]=='1')))
		i=((c[0]=='1')<<3)|((c[1]=='1')<<2)|((c[2]=='1')<<1)|(c[3]=='1');
	
	mapper_switches.nwc_dips=i;
	
	MEM_CLEAN(c);
}

_MAPSWITCHSAVE(mmc1_event)
{
	char c[10]; int i=mapper_switches.nwc_dips;
	sprintf(c,"%d%d%d%d",i>>3&1,i>>2&1,i>>1&1,i&1);
	SETTINGS_PUT_STRING(settings_info(SETTINGS_NWCDIPS),c);
}

_MAPSWITCHSET(mmc1_event)
{
	switch (set&7) {
		case MAPSWITCH_UP: set=mapper_switches.nwc_dips+(mapper_switches.nwc_dips!=0xf); break;
		case MAPSWITCH_DOWN: set=mapper_switches.nwc_dips-(mapper_switches.nwc_dips!=0); break;
		default: return;
	}
	
	if (set!=mapper_switches.nwc_dips) {
		LOG(LOG_MISC|LOG_COLOUR(LC_GREEN)|LOG_TYPE(LT_CARTSWITCH),"DIP switches: %d%d%d%d",set>>3&1,set>>2&1,set>>1&1,set&1);
		mapper_switches.nwc_dips=set;
		mmc1_event_reset_irq_hit(); /* only effective when nes is running of course */
	}
}



/* init */
_MAPINIT(mmc1)
{
	int i;
	SET_MAPNF(mmc1);
	mmc1_wram_protection=&mmc1_wram_protection_yes;
	for (i=0;i<4;i++) mmc1_io[i]=mmc1_io_std[i];
	SET_MAPIO8_ALL(mmc1);
	INIT_CHR_FIRST();
	
	switch (maptype) {
		
		/* standard, but without wram protection */
		case MAPTYPE_NINTENDO_MMC1A_SxROM:
			mmc1_wram_protection=&mmc1_wram_protection_no;
			/* continue to std */
		/* standard */
		case MAPTYPE_NINTENDO_MMC1_SxROM:
			if (MAXPRG>=63) { mmc1_io[1]=mmc1_surom_reg1; mmc1_io[2]=mmc1_surom_reg2; } /* SUROM (large prg) */
			break;
		
		/* sorom (2*8k wram) */
		case MAPTYPE_NINTENDO_MMC1_SOROM:
			mmc1_io[1]=mmc1_sorom_reg1; mmc1_io[2]=mmc1_sorom_reg2;
			SET_MAPCLEAN(mmc1_wram);
			MEM_CREATE(map_mmc1.wram,0x4000);
			cartridge->battery_size=0x2000; /* upper 8k battery backed */
			cartridge->cur_wrambank=map_mmc1.wram+0x2000;
			SET_MAPBATSAVE(mmc1_sorom);
			break;
		
		/* sxrom (32k wram) */
		case MAPTYPE_NINTENDO_MMC1_SXROM:
			mmc1_io[1]=mmc1_sxrom_reg1; mmc1_io[2]=mmc1_sxrom_reg2;
			SET_MAPCLEAN(mmc1_wram);
			MEM_CREATE(map_mmc1.wram,0x8000);
			cartridge->battery_size=0x8000;
			cartridge->cur_wrambank=map_mmc1.wram;
			SET_MAPBATSAVE(mmc1_sxrom);
			break;
		
		/* NES-EVENT */
		case MAPTYPE_NINTENDO_MMC1_WORLD_CHAMPIONSHIPS:
			map_mmc1.event_isenabled=2;
			SET_MAPSWITCHSET(mmc1_event);
			SET_MAPIRQ(mmc1_event);
			mmc1_event_reset_irq_hit();
			mmc1_io[0]=mmc1_event_reg0; mmc1_io[1]=mmc1_event_reg1; mmc1_io[2]=mmc1_nothing; mmc1_io[3]=mmc1_event_reg3;
			mmc1_io[1](); /* 1st 32k of lower 128k ROM (reg1) */
			break;
		
		default: break;
	}
	
	map_mmc1.reg[0]=BIN8(00001111); /* tested by blargg */
	INIT_MIR_H();
	mmc1_io[3](); /* init prgbanks here instead */
}
