/* Nintendo MMC2 and MMC4 mappers */

static struct {
	int latch[2];
	int reg[2][2];
	
} map_mmc2;


/* --- MMC2(NES-PNROM/NES-PEEOROM) (iNES: 9) ---
	Punch Out (any version)
	
   --- MMC4(HVC-FJROM): same as MMC2, except it can switch 16k prg instead of 8k, and has wram (iNES: 10) ---
	Famicom Wars, Fire Emblem, Fire Emblem Gaiden *** all Japanese
*/

static void __fastcall mmc2_nothing(u8 data) { return; } /* won't be called */
static void __fastcall mmc2_bs0fd(u8); static void __fastcall mmc2_bs0fe(u8); static void __fastcall mmc2_bs1fd(u8); static void __fastcall mmc2_bs1fe(u8); static void __fastcall mmc2_mirroring(u8);
static void __fastcall mmc2_prgswitch(u8); static void __fastcall mmc4_prgswitch(u8);
static __inline__ void mmc2_bankswitch(int);
typedef void __fastcall (*fpt_mmc2_io)(u8);
static fpt_mmc2_io mmc2_io[8]= { mmc2_nothing, mmc2_nothing, mmc2_nothing, mmc2_bs0fd, mmc2_bs0fe, mmc2_bs1fd, mmc2_bs1fe, mmc2_mirroring };

/* ppu memmap overrides; same as normal, but including latch check */
#define MMC2_GET_PATDATA(x) \
	int ret=cartridge->cur_patbank[x][address&0x3ff]; \
	address&=BIN8(11111000); \
	if ((address==0xd8)|(address==0xe8)) { \
		map_mmc2.latch[x>>2&1]=address; mmc2_bankswitch(x>>2&1); \
	} \
	return ret
_MAPPPU_R(mmc2,pt0d) { MMC2_GET_PATDATA(3); }
_MAPPPU_R(mmc2,pt1d) { MMC2_GET_PATDATA(7); }

/* shortcuts: call one of the underneath functions */
_MAPIO(mmc2) { mmc2_io[address>>12&7](data); }

/* $a000: PRG select */
static void __fastcall mmc2_prgswitch(u8 data) { PRG8_BS(0,data&MAXPRG); } /* switch 8k prgbank in $8000 */
static void __fastcall mmc4_prgswitch(u8 data) { PRG8_BS(0,data<<1&MAXPRG); PRG8_BS(1,(1|data<<1)&MAXPRG); } /* switch 16k prgbank in $8000 */

/* VROM select */
#define MMC2_SETLATCH(r,l) map_mmc2.reg[r][l]=data<<2&MAXVROM; mmc2_bankswitch(l)
static void __fastcall mmc2_bs0fd(u8 data) { MMC2_SETLATCH(0,0); } /* $b000: select chr bank for $fd latch 0 */
static void __fastcall mmc2_bs0fe(u8 data) { MMC2_SETLATCH(1,0); } /* $c000: select chr bank for $fe latch 0 */
static void __fastcall mmc2_bs1fd(u8 data) { MMC2_SETLATCH(0,1); } /* $d000: select chr bank for $fd latch $1000 */
static void __fastcall mmc2_bs1fe(u8 data) { MMC2_SETLATCH(1,1); } /* $e000: select chr bank for $fe latch $1000 */
/* CHR bankswitch */
static __inline__ void mmc2_bankswitch(int latch)
{
	int b=latch<<2;
	int reg=map_mmc2.reg[map_mmc2.latch[latch]>>5&1][latch];
	
	CHR4_BS_V(b,reg);
}

/* $f000, bit 0: mirroring */
static void __fastcall mmc2_mirroring(u8 data) { MIR_VH(data&1) }


/* init */
_MAPINIT(mmc2)
{
	ppu_set_fpt(PPU_FPT_CUSTOM,0x0f,mapper_mmc2_ppu_read_pt0d,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
	ppu_set_fpt(PPU_FPT_CUSTOM,0x1f,mapper_mmc2_ppu_read_pt1d,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
	map_mmc2.latch[0]=map_mmc2.latch[1]=0xe8;
	SET_MAPIO8(nothing,mmc2,mmc2,mmc2);
	INIT_PRG_FIRSTLAST();
	/* no chrbanks init */
	INIT_MIR_V();
	
	switch (maptype) {
		case MAPTYPE_NINTENDO_MMC2_PXROM:
			PRG8_BS(1,(MAXPRG-2)&MAXPRG); /* prgbanks: first 8k into $8000, last 3 from $a000 are fixed */
			mmc2_io[2]=mmc2_prgswitch;
			break;
		
		case MAPTYPE_NINTENDO_MMC4_FXROM:
			mmc2_io[2]=mmc4_prgswitch;
			DEF_6000_WRAM();
			break;
		
		default: break;
	}
}
