/* Namco(t) mappers */
/* Namco is often referred to as Namcot, especially in Japan, don't know what the deal is with that */

/* (Namcot 108 and Namcot 118 are like MMC3) */

static struct {
	int n106_chr[8];
	int n106_chr_isrom[2];
	int n106_nt_isram[4];
	int n106_irq_isenabled;
	int n106_irq_counter;
	int n106_has_vram;
	u8* n106_battery_data;
	
	/* sound */
	int n106_sound_address;
	int n106_sound_address_autoinc;
	int n106_numchans;
	u8 n106_exram[0x80];
	int* n106_dac;
} map_namcot;


/* --- Namcot 106 (iNES: 19) ---
	Battle Fleet, *Digital Devil Monogatari - Megami Tensei II, Dokuganryuu Masamune, Dragon Ninja,
	*Erika to Satoru no Yume Bouken, Famista '90, Famista '91, *Final Lap, Hydlide 3 - Yami Kara no Houmonsha,
	Juvei Quest, Kaijuu Monogatari, *King of Kings, *Mappy Kids, Mindseeker, Namco Classic, *Namco Classic 2,
	*Rolling Thunder, *Sangokushi - Chuugen no Hasha, *Sangokushi 2 - Haou no Tairiku, Splatter House - Wanpaku Graffiti,
	Star Wars, *Youkai Douchuuki *** all Japanese
	
	*: games using the add-on sound
	
   --- Namcot 106 (stripped down) (iNES: 210) ---
	Chibi Maruko-Chan - Uki Uki Shopping, Dream Master, Family Circuit '91, Famista '92, Famista '93, Famista '94, 
	Heisei Tensai Bakabon, Top Striker, Wagyan Land 2, Wagyan Land 3 *** all Japanese
*/

/* TODO:
	- sound is correct? not much documentation was available
	- sound emulation is slow, everywhere actually, but with 8 channels it gets very noticeable
	- IRQ handling is correct? (Hydlide 3 looks ugly, bug in Erika to Satoru no Yume Bouken at the book after the password screen)
*/

static void __fastcall n106_nothing(u8 address,u8 data) { return; }
static void __fastcall n106_prg0bs(u8,u8);		static void __fastcall n106_prg1bs(u8,u8);	static void __fastcall n106_prg2bs(u8,u8);
static void __fastcall n106_chrsel(u8,u8);
static void __fastcall n106_nt(u8,u8);
static void __fastcall n106_irqlow(u8,u8);		static void __fastcall n106_irqhigh(u8,u8);
static void __fastcall n106_sounddata(u8,u8);	static void __fastcall n106_soundaddress(u8,u8);

static __inline__ void n106_chrbs(void);

typedef void __fastcall (*fpt_n106_io)(u8,u8);
static const fpt_n106_io n106_io_std[0x20]={
	n106_nothing, n106_nothing, n106_nothing, n106_nothing, /* $0000-$1fff, - */
	n106_nothing, n106_nothing, n106_nothing, n106_nothing, /* $2000-$3fff, - */
	n106_nothing,											/* $4000-$47ff, - */
	n106_sounddata,											/* $4800-$4fff, sound data */
	n106_irqlow, n106_irqhigh,								/* $5000-$5fff, irq */
	n106_nothing, n106_nothing, n106_nothing, n106_nothing,	/* $6000-$7fff, - */
	n106_chrsel, n106_chrsel, n106_chrsel, n106_chrsel,		/* $8000-$9fff, chr */
	n106_chrsel, n106_chrsel, n106_chrsel, n106_chrsel,		/* $a000-$bfff, chr */
	n106_nt, n106_nt, n106_nt, n106_nt,						/* $c000-$dfff, nametables */
	n106_prg0bs, n106_prg1bs, n106_prg2bs,					/* $e000-$f7ff, prg */
	n106_soundaddress										/* $f800-$ffff, sound address */
};
fpt_n106_io n106_io[0x20];


/* io shortcut */
_MAPIO(namcot_106) { address>>=8; n106_io[address>>3](address&0xf8,data); }


/* prg bankswitch */
static void __fastcall n106_prg0bs(u8 address,u8 data) { PRG8_BS(0,data&MAXPRG); }
static void __fastcall n106_prg1bs(u8 address,u8 data) { PRG8_BS(1,data&MAXPRG); map_namcot.n106_chr_isrom[0]=data&BIT(6); map_namcot.n106_chr_isrom[1]=data&BIT(7); n106_chrbs(); } /* bits 6 and 7 also specify if lower/upper chr bank is rom or ram */
static void __fastcall n106_prg2bs(u8 address,u8 data) { PRG8_BS(2,data&MAXPRG); }


/* chr bankswitch */
static void __fastcall n106_chrsel(u8 address,u8 data) { map_namcot.n106_chr[address>>3&7]=data; n106_chrbs(); } /* chr select */
static __inline__ void n106_chrbs(void)
{
	int i=8;
	while (i--) {
		if (map_namcot.n106_chr_isrom[i>>2]) { CHR1_BS_V(i,map_namcot.n106_chr[i]&MAXVROM); cartridge->cur_patbank_isram[i]=FALSE; } /* rom */
		else { /* rom/ram */
			if (map_namcot.n106_chr[i]<0xe0) { CHR1_BS_V(i,map_namcot.n106_chr[i]&MAXVROM); cartridge->cur_patbank_isram[i]=FALSE; }
			else { CHR1_BS(i,map_namcot.n106_chr[i]&0x1f,vram); cartridge->cur_patbank_isram[i]=map_namcot.n106_has_vram=TRUE; }
		}
	}
}


/* nametables */
static void __fastcall n106_nt(u8 address,u8 data)
{
	ppu_force_update();
	address=address>>3&3;
	if (data<0xe0) { map_namcot.n106_nt_isram[address]=FALSE; mapper.ppu_memmap_cur_name_ptr[address]=cartridge->vrom[data&MAXVROM]; } /* rom */
	else { map_namcot.n106_nt_isram[address]=TRUE; mapper.ppu_memmap_cur_name_ptr[address]=mapper.ppu_memmap_name_ram_ptr[data&1]; } /* ram */
}
#define N106_SET_NAMEDATA(x) if (map_namcot.n106_nt_isram[x]) ppu_write_name##x(address,data)
_MAPPPU_W(namcot_106,name0) { N106_SET_NAMEDATA(0); }
_MAPPPU_W(namcot_106,name1) { N106_SET_NAMEDATA(1); }
_MAPPPU_W(namcot_106,name2) { N106_SET_NAMEDATA(2); }
_MAPPPU_W(namcot_106,name3) { N106_SET_NAMEDATA(3); }


/* irq, based on cpu cycles */
static void __fastcall n106_irqlow(u8 address,u8 data) { map_namcot.n106_irq_isenabled|=(map_namcot.n106_irq_counter==0xffff); map_namcot.n106_irq_counter=data|(map_namcot.n106_irq_counter&0x7f00); } /* counter runs after expired */
static void __fastcall n106_irqhigh(u8 address,u8 data) { map_namcot.n106_irq_counter=(data<<8&0x7f00)|(map_namcot.n106_irq_counter&0xff); map_namcot.n106_irq_isenabled=data>>7; cpu_acknowledge_interrupt(~IRQ_MAPPER); }
_MAPIO_R(namcot_106)
{
	u8 ret=address>>8;
	switch (address&0xf800) {
		
		/* sound */
		case 0x4800:
			ret=map_namcot.n106_exram[map_namcot.n106_sound_address];
			map_namcot.n106_sound_address+=map_namcot.n106_sound_address_autoinc;
			map_namcot.n106_sound_address&=0x7f;
			break;
		
		case 0x5000: ret=map_namcot.n106_irq_counter&0xff; break; /* irqlow */
		case 0x5800: ret=map_namcot.n106_irq_counter>>8&0x7f; break; /* irqhigh */
		default: break;
	}
	return ret;
}
_MAPIRQ(namcot_106)
{
	if (map_namcot.n106_irq_isenabled) {
		
		int cpu_cycles;
		
		for (cpu_cycles=*mapper.cpu_cycles_ptr;mapper.cpu_cycles_old>cpu_cycles;mapper.cpu_cycles_old-=crystal->rel_cycle) {
			if (++map_namcot.n106_irq_counter&0x8000) {
				map_namcot.n106_irq_counter=0xffff; /* msb indicates counter stopped */
				map_namcot.n106_irq_isenabled=FALSE;
				
				cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(mapper.cpu_cycles_old-cpu_cycles?0:INT_EDGE_IRQ_EX));
				break;
			}
		}
	}
	
	else mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}


/* battery (some games use battery backed on-board exram) */
_MAPCLEAN(namcot_106)
{
	MEM_CLEAN(map_namcot.n106_battery_data);
	cartridge->battery_data=NULL;
}
_MAPBATLOAD(namcot_106)
{
	memcpy(cartridge->wram,map_namcot.n106_battery_data,0x2000);
	memcpy(map_namcot.n106_exram,map_namcot.n106_battery_data+0x2000,0x80);
	return FALSE;
}
_MAPBATSAVE(namcot_106)
{
	memcpy(map_namcot.n106_battery_data,cartridge->wram,0x2000);
	memcpy(map_namcot.n106_battery_data+0x2000,map_namcot.n106_exram,0x80);
	return FALSE;
}


/* sound: 8 identical channels, sharing 128 bytes of waveform data */
static struct {
	int* buffer;
	int buffer_counter;
	int cycles;
	
	int unken;
	int enabled;
	int output;
	u32 timer;
	u32 timer_raw;
	u8 volume;
	u8 length;
	u8 pos;
	u8 offset;
} n106_channel[8];

/* channel clocking */
static __inline__ void n106_channel_update(const int c)
{
	const int cpu_cycles=*cpu_get_cycles_ptr();
	
	for (;n106_channel[c].cycles>cpu_cycles;n106_channel[c].cycles-=crystal->rel_cycle) {
		if (!--n106_channel[c].timer) {
			u8 offset;
			
			if (n106_channel[c].timer_raw>0xff) n106_channel[c].timer=(0xf0000*map_namcot.n106_numchans)/n106_channel[c].timer_raw;
			n106_channel[c].timer++;
			
			if (n106_channel[c].pos>=n106_channel[c].length) n106_channel[c].pos=0;
			offset=n106_channel[c].pos+n106_channel[c].offset;
			n106_channel[c].pos++;
			
			if (offset&1) n106_channel[c].output=(map_namcot.n106_exram[offset>>1]>>4)*n106_channel[c].volume*n106_channel[c].enabled;
			else n106_channel[c].output=(map_namcot.n106_exram[offset>>1]&0xf)*n106_channel[c].volume*n106_channel[c].enabled;
		}
		n106_channel[c].buffer[n106_channel[c].buffer_counter++]=n106_channel[c].output;
	}
}

/* set wave address */
static void __fastcall n106_soundaddress(u8 address,u8 data)
{
	map_namcot.n106_sound_address=data&0x7f;
	map_namcot.n106_sound_address_autoinc=data>>7;
}

static __inline__ void n106_enable_channels(void)
{
	int i=8;
	while (i--) n106_channel[i].enabled=(n106_channel[i].unken!=0)&(n106_channel[i].timer_raw>0xff)&(i>=(8-map_namcot.n106_numchans));
}

/* set data at wave address + sound IO */
static void __fastcall n106_sounddata(u8 address,u8 data)
{
	int i=8;
	const int c=map_namcot.n106_sound_address>>3&7; /* channel */
	while (i--) { n106_channel_update(i); }
	
	map_namcot.n106_exram[map_namcot.n106_sound_address]=data; /* < $40? */
	
	/* $40-$7f */
	switch (map_namcot.n106_sound_address&0xc7) {
		/* timer low */
		case 0x40:
			n106_channel[c].timer_raw=(n106_channel[c].timer_raw&0xffff00)|data;
			n106_enable_channels();
			break;
		
		/* timer mid */
		case 0x42:
			n106_channel[c].timer_raw=(n106_channel[c].timer_raw&0xff00ff)|(data<<8);
			n106_enable_channels();
			break;
		
		/* bits 0,1: timer high. bits 2,3,4: wave length. bits 5,6,7==0: channel disabled */
		case 0x44:
			n106_channel[c].timer_raw=(n106_channel[c].timer_raw&0x00ffff)|(data<<16&0x30000);
			n106_channel[c].unken=data>>5;
			n106_enable_channels();
			n106_channel[c].length=0x20-(data&0x1c); /* 8-len << 2 */
			break;
		
		/* wave offset (bit 0=low/high 4 bits) */
		case 0x46:
			n106_channel[c].offset=data;
			break;
		
		/* bits 0,1,2,3: volume. bits 4,5,6: number of channels used ($7f only) */
		case 0x47:
			n106_channel[c].volume=data&0xf;
			if (c==7) map_namcot.n106_numchans=1+(data>>4&7);
			n106_enable_channels();
			break;
		
		default: break;
	}
	
	/* auto increment, also on read (see above at irq) */
	map_namcot.n106_sound_address+=map_namcot.n106_sound_address_autoinc;
	map_namcot.n106_sound_address&=0x7f;
}

/* new frame */
_MAPNF(namcot_106)
{
	int i=8;
	
	mapper_new_frame_updateoldcycles(); /* std */
	while (i--) n106_channel[i].cycles=*cpu_get_cycles_ptr();
}

/* end frame */
_MAPEF(namcot_106)
{
	int i=8;
	while (i--) { n106_channel_update(i); }
	
	i=n106_channel[0].buffer_counter;
	while (i--) map_namcot.n106_dac[i]=19*(n106_channel[0].buffer[i]+n106_channel[1].buffer[i]+n106_channel[2].buffer[i]+n106_channel[3].buffer[i]+n106_channel[4].buffer[i]+n106_channel[5].buffer[i]+n106_channel[6].buffer[i]+n106_channel[7].buffer[i]);
}

_MAPKILLCYCLE(namcot_106)
{
	int i=8;
	while (i--) n106_channel[i].cycles-=crystal->ppu_cycle;
}

_MAPSOUNDBC(namcot_106)
{
	int i=8;
	while (i--) n106_channel[i].buffer_counter=bc;
}

/* sound init */
static void n106_sound_init(void)
{
	int i=8;
	while (i--) {
		memset(&n106_channel[i],0,sizeof(n106_channel[i]));
		n106_channel[i].buffer=sound_create_channel();
		n106_channel[i].length=0x20;
		n106_channel[i].timer=1;
	}
	map_namcot.n106_dac=sound_create_dac();
	
	map_namcot.n106_sound_address_autoinc=TRUE;
	map_namcot.n106_numchans=1;
	SET_MAPNF(namcot_106);
	SET_MAPEF(namcot_106);
	SET_MAPKILLCYCLE(namcot_106);
	SET_MAPSOUNDBC(namcot_106);
}


/* init */
_MAPINIT(namcot_106)
{
	int i;
	
	for (i=0;i<0x20;i++) { n106_io[i]=n106_io_std[i]; }
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	for (i=0;i<8;i++) map_namcot.n106_chr[i]=i;
	for (i=0;i<0x20;i++) if (cartridge->vram[i]==NULL) { MEM_CREATE(cartridge->vram[i],0x400); } /* up to 32k vram */
	SET_MAPIO8_ALL(namcot_106);
	
	switch (maptype) {
		case MAPTYPE_NAMCOT_106:
			ppu_set_fpt(PPU_FPT_NT,FALSE,NULL,NULL,NULL,NULL,mapper_namcot_106_ppu_write_name0,mapper_namcot_106_ppu_write_name1,mapper_namcot_106_ppu_write_name2,mapper_namcot_106_ppu_write_name3);
			for (i=0;i<4;i++) map_namcot.n106_nt_isram[i]=TRUE;
			SET_MAPIO4(namcot_106);
			SET_MAPIO4_R(namcot_106);
			SET_MAPIRQ(namcot_106);
			n106_sound_init();
			
			if (cartridge->battery) {
				MEM_CREATE(map_namcot.n106_battery_data,0x2080);
				memcpy(map_namcot.n106_battery_data,cartridge->wram,0x2000);
				cartridge->battery_size=0x2080;
				cartridge->battery_data=map_namcot.n106_battery_data;
				SET_MAPBATSAVE(namcot_106);
				SET_MAPBATLOAD(namcot_106);
				SET_MAPCLEAN(namcot_106);
			}
			break;
		
		case MAPTYPE_NAMCOT_106_STRIPPED_DOWN:
			/* no nt, no irq, no sound */
			n106_io[0x18]=n106_io[0x19]=n106_io[0x1a]=n106_io[0x1b]=n106_io[0x1f]=n106_nothing;
			break;
		
		default: break;
	}
}
