/* Pirate mappers */

static struct {
	/* mariobaby */
	int mariobaby_irq_isenabled;
	int mariobaby_irq_counter;
	
	/* smb2j fds */
	int smb2j_irq_isenabled;
	int smb2j_irq_counter;
	
} map_pirate;


/* --- Contra Function 16 (iNES: 15) ---
	100-in-1 Contra Function 16,
		Contra, Arkanoid, Tetris (BPS), Tetris (Tengen), Super Mario Bros, Bomber Man, Mario Bros,
		Popeye, Donkey Kong, Donkey Kong 3, Wild Gunman, Duck Hunt, Hogan's Alley, Pinball,
		F-1 Race, Urban Champion, Balloon Fight, Baseball, Ice Climber, Mappy, Galaxian, Galaga,
		Antarctic Adventure, Chack 'n Pop, Star Force, Pac Man, Warp Man, Battle City, Field Combat,
		Pooyan, Sky Destroyer, Binary Land, Choujikuu Yousai - Macross, Championship Lode Runner *** 34
	Bao Xiao Tien Guo, Bio Hazard, A Chinese Odyssey, Phantasy Star IV(or 2), Pokemon Gold (Mars), ShangHai Tycoon, Yuefei,
	
	(and some unmentioned hacks)
	
	Bio Hazard locks up sometimes: rts->invalid address, bad dump?
*/

_MAPIO(contraf16)
{
	int highbank=data<<1&BIN8(01111110)&MAXPRG; /* bits 0-5: high bank */
	int lowbank=data>>7; /* bit 7: low bank, weird on non-8k modes */
	
	/* addressbus bits 0,1: bankswitch mode (not sure about bits 2,3) */
	switch (address&0xf) {
		case 0:
			/* 32k */
			highbank&=~2;
			PRG8_BS(0,highbank|lowbank);
			PRG8_BS(1,(highbank|1)^lowbank);
			PRG8_BS(2,highbank|2|lowbank);
			PRG8_BS(3,(highbank|3)^lowbank);
			break;
		case 1:
			/* 128k */
			PRG8_BS(0,(highbank|lowbank)&MAXPRG);
			PRG8_BS(1,((highbank|1)^lowbank)&MAXPRG);
			PRG8_BS(2,(MAXPRG&~1)|lowbank);
			PRG8_BS(3,MAXPRG^lowbank);
			break;
		case 2: {
			/* 8k */
			int bank=highbank|lowbank,i=4;
			while (i--) { PRG8_BS(i,bank); }
			break;
		}
		case 3:
			/* 16k */
			PRG8_BS(0,highbank|lowbank); PRG8_BS(2,highbank|lowbank);
			PRG8_BS(1,(highbank|1)^lowbank); PRG8_BS(3,(highbank|1)^lowbank);
			break;
		
		default: break;
	}
	
	/* bit 6: set mirroring */
	MIR_VH(data&BIT(6));
}

_MAPINIT(contraf16)
{
	SET_MAPIO8_ALL(contraf16);
	INIT_MIR_V(); /* register=0 */
	INIT_PRG_FIRST(); /* register=0 */
	INIT_CHR_FIRST(); /* std vram (some hacks use 8k vrom though) */
	DEF_6000_WRAM();
}





/* --- Mario Baby (iNES: 42(/254?)) ---
	Bio Miracle Bokutte Upa, Ai Senshi Nicol.. (both: FDS->NES ROM bootlegs)
*/

/* $8000: chr bankswitch, only for Ai Senshi Nicol, Bio Miracle has VRAM */
_MAPIO(mariobaby_8)
{
	if (MAXVROM) {
		int bank=data<<3&MAXVROM;
		CHR8_BS_V(bank);
	}
}

/* $e000: IO */
_MAPIO(mariobaby_e)
{
	switch (address&3) {
		
		/* select 8k prgbank */
		case 0: PRG_BS_6000(data&MAXPRG); break;
		
		/* set mirroring (bit 3) */
		case 1: MIR_VH(data&BIT(3)) break;
		
		/* enable/disable irq */
		case 2:
			map_pirate.mariobaby_irq_isenabled=data>>1&1;
			if (map_pirate.mariobaby_irq_isenabled) map_pirate.mariobaby_irq_counter=0x6000*crystal->rel_cycle;
			else cpu_acknowledge_interrupt(~IRQ_MAPPER);
			break;
		
		case 3: break;
	}
}

/* irq handler, based on cpu cycles, irq generated after 24576 cycles */
_MAPIRQ(mariobaby)
{
	cpu_acknowledge_interrupt(~IRQ_MAPPER); /* it seems to expect the irqline to go down after an irq */
	if (map_pirate.mariobaby_irq_isenabled) {
		map_pirate.mariobaby_irq_counter-=(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr);
		if (map_pirate.mariobaby_irq_counter<=0) {
			cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(map_pirate.mariobaby_irq_counter==0?INT_EDGE_IRQ_EX:0));
			map_pirate.mariobaby_irq_counter+=(0x6000*crystal->rel_cycle);
		}
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

/* init */
_MAPINIT(mariobaby)
{
	SET_MAPIO8(mariobaby_8,nothing,nothing,mariobaby_e);
	SET_MAPIRQ(mariobaby);
	SET_MAPNF(updateoldcycles);
	INIT_MIR_V();
	INIT_CHR_FIRST();
	
	/* prgbanks: 0=c, 1=d, 2=e, 3=f, $6000-$7fff=switchable */
	DEF_6000_PRG();
	INIT_PRG_LAST();
}





/* --- Pirate SMB2J FDS (iNES: 50) ---
   --- Pirate SMB2J FDS alternate (iNES: 40) (bit different hardware, same functions) ---
	Super Mario Bros 2. (J) (FDS->NES ROM bootleg port)
	
	in GoodNES 2.01: 40: Mario and Luigi (SMB1 Hack) [a1],			50: Super Mario Bros. (JU) (Alt Levels) [p2][!]
	in GoodNES 3.10: 40: Super Mario Bros 2 (Lost Levels) (Unl),	50: Super Mario Bros. (Alt Levels) [p1][!]
*/

/* shared */
#define SMB2J_DISABLE_IRQ()	map_pirate.smb2j_irq_isenabled=FALSE; cpu_acknowledge_interrupt(~IRQ_MAPPER)
#define SMB2J_ENABLE_IRQ()	map_pirate.smb2j_irq_isenabled=TRUE; map_pirate.smb2j_irq_counter=0x1000*crystal->rel_cycle
#define SMB2J_PRG_BS()		PRG8_BS(2,data&MAXPRG)

/* IO for alternate type */
_MAPIO(smb2j_8) { SMB2J_DISABLE_IRQ(); }	/* $8000: disable irq */
_MAPIO(smb2j_a) { SMB2J_ENABLE_IRQ(); }		/* $a000: enable irq */
_MAPIO(smb2j_e) { SMB2J_PRG_BS(); }			/* $e000: prg bankswitch */

/* IO for normal type */
_MAPIO(smb2j)
{
	address&=BIN16(11100001,01100000);
	if ((address&~0x100)==BIN16(01000000,00100000)) {
		if (address&0x100) {
			/* bit 0 enable/disable irq */
			if (data&1) { SMB2J_ENABLE_IRQ(); }
			else { SMB2J_DISABLE_IRQ(); }
		}
		else {
			/* prg bankswitch */
			data=(data&~7)|(data>>1&3)|(data<<2&BIT(2));
			SMB2J_PRG_BS();
		}
	}
}

/* irq handler, based on cpu cycles, irq generated after 4096 cycles */
_MAPIRQ(smb2j)
{
	if (map_pirate.smb2j_irq_isenabled) {
		map_pirate.smb2j_irq_counter-=(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr);
		if (map_pirate.smb2j_irq_counter<=0) {
			cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(map_pirate.smb2j_irq_counter==0?INT_EDGE_IRQ_EX:0));
			map_pirate.smb2j_irq_counter+=(0x1000*crystal->rel_cycle);
		}
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

/* init */
_MAPINIT(smb2j)
{
	int bank;
	
	SET_MAPIRQ(smb2j);
	SET_MAPNF(updateoldcycles);
	INIT_CHR_FIRST(); /* 40: 8k vrom, 50: 8k vram */
	DEF_6000_PRG();
	
	switch (maptype) {
		case MAPTYPE_PIRATE_SMB2J_FDS:
			SET_MAPIO4(smb2j);
			/* prgbanks: 0=8, 1=9, 2=switchable, 3=11, $6000-$7fff=last */
			bank=8; PRG32_BS(bank);
			PRG_BS_6000(MAXPRG);
			break;
		
		case MAPTYPE_PIRATE_SMB2J_FDS_ALTERNATE:
			SET_MAPIO8(smb2j_8,smb2j_a,nothing,smb2j_e);
			/* prgbanks: 0=4, 1=5, 2=switchable, 3=7, $6000-$7fff=6 */
			INIT_PRG_LAST();
			PRG_BS_6000(6);
			break;
		
		default: break;
	}
}





/* --- Wai Xing Zhan Shi (iNES: 242) ---
	Wai Xing Zhan Shi (part 2 of this game is known as 'Phantasy Star IV' on the Contra Function 16 mapper, almost the same)
*/

_MAPIO(waixing)
{
	/* addressbus bits 2-: select 32k prg */
	int bank=address>>1&~3&MAXPRG;
	PRG32_BS(bank);
	
	/* databus bit 0: set mirroring */
	MIR_VH(data&1);
}

_MAPINIT(waixing)
{
	SET_MAPIO8_ALL(waixing);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST(); /* 8k vram */
	INIT_MIR_V();
}
