/* Sunsoft mappers */
#include "ym2149.h" /* Sunsoft 5 sound */

static struct {
	/* 2 */
	int s2_mirroring;
	
	/* 3 */
	int s3_irq_counter;
	int s3_irq_isenabled;
	int s3_irq_write;
	
	/* 4 */
	int s4_ntvrom_isenabled;
	int s4_ntvrom_bank[2];
	int s4_mirroring;
	
	/* 5 */
	int s5_irq_counter;
	int s5_irq_isenabled;
	int s5_irqcounter_isenabled;
	int s5_index;
	
} map_sunsoft;


/* --- Sunsoft 1 (iNES: 184/122) ---
	Fantasy Zone, Atlantis no Nazo, Wing of Madoola, Kanshakudama Nage Kantarou no Toukaidou Gojuusan Tsugi *** all Japanese
*/

_MAPIO(sunsoft1)
{
	if (MAXVROM) {
		/* cartridge with vrom: */
		/* low nibble: swap 4k vrombank in $0000 */
		int bank=data<<2&MAXVROM;
		CHR4_BS_V(0,bank);
		
		/* high nibble: swap 4k vrombank in $1000 */
		bank=(data>>4)<<2&MAXVROM;
		CHR4_BS_V(4,bank);
	}
	else {
		/* cartridge with vram: swap 16k prgbank in $8000 */
		int bank=data<<1&MAXPRG;
		PRG16_BS(0,bank);
	}
}

_MAPINIT(sunsoft1)
{
	SET_MAPIO6(sunsoft1);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
}





/* --- Sunsoft 2 (iNES: 89/93) ---
	Mito Koumon, Shanghai *** both Japanese
*/

_MAPIO(sunsoft2)
{
	/* bits 4,5,6: swap 16k prgbank in $8000 */
	int bank=(data&BIN8(01110000))>>3&MAXPRG;
	PRG16_BS(0,bank);
	
	/* bit 3: set mirroring (not if it's hardwired) */
	if (map_sunsoft.s2_mirroring) {
		MIR_01(data&BIT(3))
	}
	
	/* bits 0,1,2 and 7: swap 8k vrombank */
	if (MAXVROM) {
		bank=((data&BIN8(00000111))|(data>>4&BIT(3)))<<3&MAXVROM;
		CHR8_BS_V(bank);
	}
}

_MAPINIT(sunsoft2)
{
	SET_MAPIO8_ALL(sunsoft2);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	
	switch (maptype) {
		/* normal, one-screen mirroring */
		case MAPTYPE_SUNSOFT_2:
			map_sunsoft.s2_mirroring=TRUE;
			INIT_MIR_0();
			break;
		
		/* hardwired mirroring */
		case MAPTYPE_SUNSOFT_2_HARDWIRED_MIRRORING:
			map_sunsoft.s2_mirroring=FALSE;
			break;
		
		default: break;
	}
}





/* --- Sunsoft 3 (iNES: 67) ---
	Fantasy Zone 2, Mito Koumon - Sekai Manyuu Ki *** both Japanese
*/

static void __fastcall sunsoft3_nothing(u8 bank,u8 data) { return; } /* invalid IO address */
typedef void __fastcall (*fpt_sunsoft3_io)(u8,u8); static void __fastcall sunsoft3_chrbs(u8,u8); static void __fastcall sunsoft3_irqcounter(u8,u8); static void __fastcall sunsoft3_irqcontrol(u8,u8); static void __fastcall sunsoft3_mirroring(u8,u8); static void __fastcall sunsoft3_prgbs(u8,u8);
static const fpt_sunsoft3_io sunsoft3_io[0x10]= { sunsoft3_nothing, sunsoft3_chrbs, sunsoft3_nothing, sunsoft3_chrbs, sunsoft3_nothing, sunsoft3_chrbs, sunsoft3_nothing, sunsoft3_chrbs, sunsoft3_nothing, sunsoft3_irqcounter, sunsoft3_nothing, sunsoft3_irqcontrol, sunsoft3_nothing, sunsoft3_mirroring, sunsoft3_nothing, sunsoft3_prgbs };

/* shortcuts: call one of the underneath functions */
_MAPIO(sunsoft3) { u8 bank=address>>11&0xf; sunsoft3_io[bank](bank,data); }

/*	$8800: 2K VROM in PPU $0000
	$9800: 2K VROM in PPU $0800
	$a800: 2K VROM in PPU $1000
	$b800: 2K VROM in PPU $1800 */
static void __fastcall sunsoft3_chrbs(u8 bank,u8 data) { int i=data<<1&MAXVROM; CHR1_BS_V(bank-1,i); CHR1_BS_V(bank,i|1); } 

/* $c800: set 16bit IRQ counter. first write=bits 8-15, second write=bits 0-7 */
static void __fastcall sunsoft3_irqcounter(u8 bank,u8 data)
{
	u16 count=map_sunsoft.s3_irq_counter/crystal->rel_cycle;
	map_sunsoft.s3_irq_write^=1;
	if (map_sunsoft.s3_irq_write) count=(count&0xff00)|data;
	else count=(count&0xff)|(data<<8);
	map_sunsoft.s3_irq_counter=(count*crystal->rel_cycle)+(map_sunsoft.s3_irq_counter%crystal->rel_cycle);
}

/* $d800, bit 4: enable(1)/disable(0) IRQ */
static void __fastcall sunsoft3_irqcontrol(u8 bank,u8 data) { map_sunsoft.s3_irq_isenabled=data>>4&1; if (!map_sunsoft.s3_irq_isenabled) cpu_acknowledge_interrupt(~IRQ_MAPPER); }

/* $e800, bits 0,1: mirroring */
static void __fastcall sunsoft3_mirroring(u8 bank,u8 data) { MIR_VH01(data&3) }

/* $f800: 16K PRG in $8000 */
static void __fastcall sunsoft3_prgbs(u8 bank,u8 data) { int i=data<<1&MAXPRG; PRG16_BS(0,i); }

/* irq, based on CPU cycles, fired when the counter reaches 0 */
_MAPIRQ(sunsoft3)
{
	map_sunsoft.s3_irq_counter-=(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr);
	if (map_sunsoft.s3_irq_counter<0) {
		if (map_sunsoft.s3_irq_isenabled) cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER);
		map_sunsoft.s3_irq_counter+=(0x10000*crystal->rel_cycle);
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}

/* init */
_MAPINIT(sunsoft3)
{
	SET_MAPIO8_ALL(sunsoft3);
	SET_MAPIRQ(sunsoft3);
	SET_MAPNF(updateoldcycles);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	INIT_MIR_V();
}





/* --- Sunsoft 4 (iNES: 68) ---
	After Burner (USA)/2 (J), Maharaja, Nantettatte!! Baseball
*/

/* TODO:
	Kogame Cassette (optional add-on cart) for Nantettatte!! Baseball
*/

typedef void __fastcall (*fpt_sunsoft4_io)(u8); static void __fastcall sunsoft4_pat00bs(u8); static void __fastcall sunsoft4_pat08bs(u8); static void __fastcall sunsoft4_pat10bs(u8); static void __fastcall sunsoft4_pat18bs(u8); static void __fastcall sunsoft4_nt0(u8); static void __fastcall sunsoft4_nt1(u8); static void __fastcall sunsoft4_mirroring(u8); static void __fastcall sunsoft4_prgbs(u8);
static const fpt_sunsoft4_io sunsoft4_io[8]= { sunsoft4_pat00bs, sunsoft4_pat08bs, sunsoft4_pat10bs, sunsoft4_pat18bs, sunsoft4_nt0, sunsoft4_nt1, sunsoft4_mirroring, sunsoft4_prgbs };
static __inline__ void sunsoft4_nt(void);

/* ppu nametable writes */
#define SUNSOFT4_SET_NAMEDATA(x) if (!map_sunsoft.s4_ntvrom_isenabled) ppu_write_name##x(address,data)
_MAPPPU_W(sunsoft4,name0) { SUNSOFT4_SET_NAMEDATA(0); }
_MAPPPU_W(sunsoft4,name1) { SUNSOFT4_SET_NAMEDATA(1); }
_MAPPPU_W(sunsoft4,name2) { SUNSOFT4_SET_NAMEDATA(2); }
_MAPPPU_W(sunsoft4,name3) { SUNSOFT4_SET_NAMEDATA(3); }

/* shortcuts: call one of the underneath functions */
_MAPIO(sunsoft4) { sunsoft4_io[address>>12&7](data); }

/* chr bankswitch */
/* normal */
#define SUNSOFT4_PATBS(x) int i=data<<1&MAXVROM; CHR2_BS_V(x,i)
static void __fastcall sunsoft4_pat00bs(u8 data) { SUNSOFT4_PATBS(0); } /* $8000: select 2k chr bank in ppu $0000 */
static void __fastcall sunsoft4_pat08bs(u8 data) { SUNSOFT4_PATBS(2); } /* $9000: select 2k chr bank in ppu $0800 */
static void __fastcall sunsoft4_pat10bs(u8 data) { SUNSOFT4_PATBS(4); } /* $a000: select 2k chr bank in ppu $1000 */
static void __fastcall sunsoft4_pat18bs(u8 data) { SUNSOFT4_PATBS(6); } /* $b000: select 2k chr bank in ppu $1800 */
/* in nametables */
#define SUNSOFT4_NT(x) map_sunsoft.s4_ntvrom_bank[x]=(data|0x80)&MAXVROM; sunsoft4_nt()
static void __fastcall sunsoft4_nt0(u8 data) { SUNSOFT4_NT(0); } /* $c000: select nametable 0 chr bank (from last 128) */
static void __fastcall sunsoft4_nt1(u8 data) { SUNSOFT4_NT(1); } /* $d000: select nametable 1 chr bank (from last 128 ) */

/* $e000: set mirroring/nametable vram/vrom */
static void __fastcall sunsoft4_mirroring(u8 data)
{
	map_sunsoft.s4_mirroring=data&3;
	map_sunsoft.s4_ntvrom_isenabled=data>>4&1;
	if (map_sunsoft.s4_ntvrom_isenabled) { sunsoft4_nt(); }
	
	/* else, standard (ppu internal vram) mirroring */
	else { MIR_VH01(map_sunsoft.s4_mirroring) }
}

/* $f000: 16K PRG in $8000 */
static void __fastcall sunsoft4_prgbs(u8 data) { int i=data<<1&MAXPRG; PRG16_BS(0,i); } /* select 16k prg bank in $8000 */

/* set mirroring/nametables */
static __inline__ void sunsoft4_nt(void)
{
	ppu_force_update();
	if (map_sunsoft.s4_ntvrom_isenabled) { /* set chr rom to nametable bank */
		switch (map_sunsoft.s4_mirroring) {
			case 0: mapper.ppu_memmap_cur_name_ptr[0]=mapper.ppu_memmap_cur_name_ptr[2]=cartridge->vrom[map_sunsoft.s4_ntvrom_bank[0]]; mapper.ppu_memmap_cur_name_ptr[1]=mapper.ppu_memmap_cur_name_ptr[3]=cartridge->vrom[map_sunsoft.s4_ntvrom_bank[1]]; break;
			case 1: mapper.ppu_memmap_cur_name_ptr[0]=mapper.ppu_memmap_cur_name_ptr[1]=cartridge->vrom[map_sunsoft.s4_ntvrom_bank[0]]; mapper.ppu_memmap_cur_name_ptr[2]=mapper.ppu_memmap_cur_name_ptr[3]=cartridge->vrom[map_sunsoft.s4_ntvrom_bank[1]]; break;
			case 2: mapper.ppu_memmap_cur_name_ptr[0]=mapper.ppu_memmap_cur_name_ptr[1]=mapper.ppu_memmap_cur_name_ptr[2]=mapper.ppu_memmap_cur_name_ptr[3]=cartridge->vrom[map_sunsoft.s4_ntvrom_bank[0]]; break;
			case 3: mapper.ppu_memmap_cur_name_ptr[0]=mapper.ppu_memmap_cur_name_ptr[1]=mapper.ppu_memmap_cur_name_ptr[2]=mapper.ppu_memmap_cur_name_ptr[3]=cartridge->vrom[map_sunsoft.s4_ntvrom_bank[1]]; break;
		}
	}
}

/* init */
_MAPINIT(sunsoft4)
{
	SET_MAPIO8_ALL(sunsoft4);
	ppu_set_fpt(PPU_FPT_NT,FALSE,NULL,NULL,NULL,NULL,mapper_sunsoft4_ppu_write_name0,mapper_sunsoft4_ppu_write_name1,mapper_sunsoft4_ppu_write_name2,mapper_sunsoft4_ppu_write_name3);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	INIT_MIR_V();
	DEF_6000_WRAM();
}





/* --- Sunsoft 5/FME-7 (iNES: 69) ---
	Gimmick/Mr. Gimmick, Hebereke, Batman (J), Dynamite Batman/Batman - Return of the Joker, Barcode World,
	Gremlin 2 - Shinshu Tanjou, Honoo no Doukyuuji - Dodge Danpei 1,2, Pyokotan no Dai Meiro
*/

/* sound: YM2149, or clone; 3 squarewaves channels, optionally can be substituted for noise. AFAIK only used by Gimmick (J), and noise/envelopes/IO are unused */
/* (sound test in Gimmick=on titlescreen hold select, then press start) */

_MAPIO(sunsoft5_sound_address) { ym2149_write_address(data); } /* $a000: set sound index */
_MAPIO(sunsoft5_sound_data) { ym2149_write_data(data); } /* $e000: set sound data */

_MAPNF(sunsoft5) { mapper_new_frame_updateoldcycles(); ym2149_new_frame(); }
_MAPEF(sunsoft5) { ym2149_end_frame(); }
_MAPKILLCYCLE(sunsoft5) { ym2149_kill_cycle(); }
_MAPSOUNDBC(sunsoft5) { ym2149_set_buffer_counter(bc); }

/* sound init */
static void sunsoft5_sound_init(void)
{
	SET_MAPNF(sunsoft5);
	SET_MAPEF(sunsoft5);
	SET_MAPKILLCYCLE(sunsoft5);
	SET_MAPSOUNDBC(sunsoft5);
	
	ym2149_init();
}


/* general IO */

/* $8000: set index */
_MAPIO(sunsoft5_index) { map_sunsoft.s5_index=data&0xf; }

/* $c000: set data */
_MAPIO(sunsoft5_data)
{
	switch (map_sunsoft.s5_index) {
		/* 1K vrom switch */
		case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
			CHR1_BS(map_sunsoft.s5_index,data&MAXVROM,vrom); break;
		
		/* 8K prg/ram switch ($6000-$dfff) */
		case 8:
			if (data&0x40) {
				DEF_6000_WRAM();
				cartridge->wram_isenabled=data>>7;
			}
			else {
				DEF_6000_PRG();
				PRG_BS_6000((data&BIN8(00111111))&MAXPRG);
			}
			break;
		case 9: case 0xa: case 0xb:
			PRG8_BS(map_sunsoft.s5_index-9,data&MAXPRG); break;
		
		/* mirroring */
		case 0xc: MIR_VH01(data&3) break;
		
		/* irq */
		case 0xd: /* control */
			map_sunsoft.s5_irqcounter_isenabled=data&1;
			map_sunsoft.s5_irq_isenabled=data>>7&map_sunsoft.s5_irqcounter_isenabled;
			if (!map_sunsoft.s5_irq_isenabled) cpu_acknowledge_interrupt(~IRQ_MAPPER);
			break;
		case 0xe: case 0xf: { /* 16 bit counter */
			u16 count=map_sunsoft.s5_irq_counter/crystal->rel_cycle;
			if (map_sunsoft.s5_index&1) count=(count&0xff)|(data<<8); /* f: high byte */
			else count=(count&0xff00)|data; /* e: low byte */
			map_sunsoft.s5_irq_counter=(count*crystal->rel_cycle)+(map_sunsoft.s5_irq_counter%crystal->rel_cycle);
			break;
		}
	}
}

/* irq, based on CPU cycles, fired when the counter reaches 0: almost the same as with Sunsoft 3 */
_MAPIRQ(sunsoft5)
{
	if (map_sunsoft.s5_irqcounter_isenabled) map_sunsoft.s5_irq_counter-=(mapper.cpu_cycles_old-*mapper.cpu_cycles_ptr);
	if (map_sunsoft.s5_irq_counter<0) {
		if (map_sunsoft.s5_irq_isenabled) cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER);
		map_sunsoft.s5_irq_counter+=(0x10000*crystal->rel_cycle);
	}
	mapper.cpu_cycles_old=*mapper.cpu_cycles_ptr;
}


/* main init */
_MAPINIT(sunsoft5)
{
	SET_MAPIO8(sunsoft5_index,sunsoft5_data,sunsoft5_sound_address,sunsoft5_sound_data);
	SET_MAPIRQ(sunsoft5);
	INIT_PRG_FIRSTLAST();
	DEF_6000_PRG(); PRG_BS_6000(0);
	INIT_CHR_FIRST();
	INIT_MIR_V();
	
	sunsoft5_sound_init();
}
