/* Taito mappers */

static struct {
	/* tc */
	int tc_irq_isenabled;
	u8 tc_irq_latch;
	u8 tc_irq_counter;
	
	/* x */
	int x_chr[8];
	int x_xor;
	int x_type;
	int x_one;
	
} map_taito;


/* --- Taito TC0190/0350 (0350 being a stripped down 0190?: don't use irq/mirroring) (iNES: 33/48) ---
	Akira, Bakushou!! Jinsei Gekijou 1,2,3, Bubble Bobble 2, Captain Saver, Don Doko Don 1,2, Golf Ko Open, Insector X,
	Flintstones - The Rescue of Dino & Hoppy, Jetsons, The - Cogswell's Caper!, Operation Wolf, Power Blazer,
	Takeshi no Sengoku Fuuunji *** all Japanese
*/

/* TODO:
	- Supposedly, bit 6 of 8000.0 is a mirroring switch on TC0350, I doubt that
	- IRQ handling is probably incorrect, I've tried hooking up an A12 handler (like MMC3), but that caused problems
	in the Jetsons intro(use IRQs: Flintstones, Jetsons, Captain Saver):
		- Flintstones: glitch at the 'skeleton cave' level (see demo)
		- Jetsons: small glitch at the titlescreen
*/

/* $8000-$9fff: bankswitch */
_MAPIO(taito_tc_8)
{
	switch(address&3) {
		case 0: PRG8_BS(0,data&MAXPRG); break; /* select 8k prg bank at $8000 */
		case 1: PRG8_BS(1,data&MAXPRG); break; /* select 8k prg bank at $c000 */
		
		/* select 2k chr bank at 2:$0000/$1000,3:$0800/$1800 */
		case 2: case 3:
			CHR1_BS_V(address<<1&2,data<<1&MAXVROM);
			CHR1_BS_V((address<<1&2)|1,(data<<1&MAXVROM)|1);
			break;
	}
}

/* $a000-$bfff: upper chr bankswitch */
_MAPIO(taito_tc_a)
{
	/* select 1k chr bank at 0:$1000/$0000,1:$1400/$0400,2:$1800/$0800,3:$1c00/$0c00 */
	CHR1_BS_V((address&3)|4,data&MAXVROM);
}


/* $c000-$dfff: irq */
_MAPIO(taito_tc_c)
{
	switch (address&3) {
		case 0: map_taito.tc_irq_latch=map_taito.tc_irq_counter=data; break;
		case 1: map_taito.tc_irq_counter=map_taito.tc_irq_latch; break;
		case 2: map_taito.tc_irq_isenabled=TRUE; break;
		case 3: cpu_acknowledge_interrupt(~IRQ_MAPPER); map_taito.tc_irq_isenabled=FALSE; break;
	}
}

/* $e000-$ffff: mirroring */
_MAPIO(taito_tc_e)
{
	switch (address&3) {
		case 0: MIR_VH(data&0x40) break;
		default: break;
	}
}

/* irq handler, based on scanlines */
_MAPIRQ(taito_tc)
{
	for (;;) {
		if (*mapper.cpu_cycles_ptr<crystal->sc[mapper.sc+1]) { /* new scanline */
			if (++mapper.sc>239) return;
			if (*mapper.ppu_enabled_ptr&map_taito.tc_irq_isenabled) {
				if (!++map_taito.tc_irq_counter) {
					map_taito.tc_irq_isenabled=FALSE;
					cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER);
				}
			}
		}
		else break;
	}
}

/* init */
_MAPINIT(taito_tc)
{
	INIT_PRG_FIRSTLAST();
	/* no chrbanks init */
	INIT_MIR_V();
	SET_MAPIO8(taito_tc_8,taito_tc_a,taito_tc_c,taito_tc_e);
	SET_MAPIRQ(taito_tc);
	SET_MAPNF(resetsc);
}





/* --- X-005 (iNES: 80) ---
	Kyonshiizu 2, Kyuukyoku Harikiri Stadium '88, Kyuukyoku Harikiri Stadium, Minelvaton Saga, Mirai Shinwa Jarvas,
	Taito Grand Prix - Eikou heno License, Yamamura Misa Suspense - Kyouto Ryuu no Tera Satsujin *** all Japanese
	
   --- X-005 one-screen mirroring (iNES: 207) ---
	Fudou Myouou Den (Japanese)
	
   --- X1-017 (almost like the X-005, but with chr xor, and different prg bankswitch) (iNES: 82) ---
	Kyuukyoku Harikiri Koushien, Kyuukyoku Harikiri Stadium - Heisei Gannen Han, Kyuukyoku Harikiri Stadium 3,
	SD Keiji - Blader *** all Japanese
*/

/* TODO:
	- $7ef7-$7ef9?
	- Kyuukyoku Harikiri Koushien will lock up with a message, the first time you start it,
	  probably saying that the battery backed SRAM is broken (this is normal I guess, but kinda neat)
*/

static __inline__ void taito_x_chr_bs(void);
void __fastcall (*taito_x_prg_bs)(u16,u8);
static __fastcall void taito_x_005_prg_bs(u16,u8);
static __fastcall void taito_x1_17_prg_bs(u16,u8);

/* io */
_MAPIO(taito_x)
{
	WRAM_WRITE(data);
	
	switch (address) {
		/* select 2k chr, normally for the lower half of the ppu address space */
		case 0x7ef0: case 0x7ef1:
			map_taito.x_chr[address<<1&2]=data&~1&MAXVROM;
			map_taito.x_chr[(address<<1&2)|1]=(data|1)&MAXVROM;
			map_taito.x_one=data&0x80;
			taito_x_chr_bs();
			break;
		
		/* select 1k chr, normally for the upper half of the ppu address space */
		case 0x7ef2: case 0x7ef3: case 0x7ef4: case 0x7ef5:
			map_taito.x_chr[(address&7)+2]=data&MAXVROM;
			map_taito.x_one=data&0x80;
			taito_x_chr_bs();
			break;
		
		/* mirroring/xor */
		case 0x7ef6:
			/* bit 0: set mirroring */
			MIR_HV(data&1)
			
			/* bit 1: set chr xor, only for X1-17 */
			map_taito.x_xor=data<<1&map_taito.x_type;
			taito_x_chr_bs();
			break;
		
		/* enable/disable WRAM? */
		case 0x7ef7: case 0x7ef8: case 0x7ef9: break;
		
		/* select prg */
		default: (*taito_x_prg_bs)(address,data); break;
	}
}

/* chr bankswitch */
static __inline__ void taito_x_chr_bs(void)
{
	int i=8;
	while (i--) { CHR1_BS_V(i^map_taito.x_xor,map_taito.x_chr[i]); }
	
	/* one-screen mirroring for Taito X-005 type 2 */
	if (map_taito.x_type&1) { MIR_01(map_taito.x_one) }
}

/* prg bankswitch for Taito X-005 */
static __fastcall void taito_x_005_prg_bs(u16 address,u8 data)
{
	switch (address) {
		/* select 8k prg in $8000, $a000, or $c000 (a,b=0, c,d=1, e,f=2) */
		case 0x7efa: case 0x7efb: case 0x7efc: case 0x7efd: case 0x7efe: case 0x7eff:
			PRG8_BS(((address-2)>>1)&3,data&MAXPRG); break;
		
		default: break;
	}
}

/* prg bankswitch for Taito X1-17 */
static __fastcall void taito_x1_17_prg_bs(u16 address,u8 data)
{
	switch (address) {
		/* select 8k prg in $8000, $a000, or $c000 (bits 2 and up) */
		case 0x7efa: case 0x7efb: case 0x7efc:
			PRG8_BS((address&7)-2,data>>2&MAXPRG); break;
		
		default: break;
	}
}

/* init */
_MAPINIT(taito_x)
{
	int i;
	
	SET_MAPIO6(taito_x);
	
	for (i=0;i<8;i++) map_taito.x_chr[i]=i;
	INIT_CHR_FIRST();
	INIT_PRG_FIRSTLAST();
	DEF_6000_WRAM();
	INIT_MIR_H();
	
	switch (maptype) {
		
		case MAPTYPE_TAITO_X_005:
			taito_x_prg_bs=&taito_x_005_prg_bs;
			break;
		
		case MAPTYPE_TAITO_X_005_ONE_SCREEN_MIRRORING:
			INIT_MIR_0();
			taito_x_prg_bs=&taito_x_005_prg_bs;
			map_taito.x_type=1;
			break;
		
		case MAPTYPE_TAITO_X1_017:
			taito_x_prg_bs=&taito_x1_17_prg_bs;
			map_taito.x_type=4;
			break;
		
		default: break;
	}
}
