/* Tengen mappers */

/* (MIMIC-1 is like MMC3) */

static struct {
	/* rambo-1 */
	int r1_control;
	int r1_chr[8];
	int r1_prg[3];
	int r1_irq_isenabled;
	int r1_irq_latch;
	int r1_irq_counter;
	int r1_irq_counter_old;
	int r1_irq_reload;
	int r1_irq_type;
	int r1_irq_ccc;
	
} map_tengen;


/* --- RAMBO-1 (still quite similar to MMC3) (iNES: 64) ---
	Klax, Rolling Thunder, Skull & Crossbones, Shinobi
*/

/* TODO:
	Shinobi: the statusbar is missing at 2nd boss (helicopter), the game doesn't use Rambo-1 IRQs
*/

/* irq */
#define RAMBO_IRQCLOCK(x) \
if ((map_tengen.r1_irq_counter==0)|map_tengen.r1_irq_reload) { \
	map_tengen.r1_irq_counter=map_tengen.r1_irq_latch+map_tengen.r1_irq_reload; \
	map_tengen.r1_irq_counter_old=0; \
	map_tengen.r1_irq_reload=FALSE; \
} \
else map_tengen.r1_irq_counter_old=map_tengen.r1_irq_counter--; \
if ((map_tengen.r1_irq_counter==0)&(map_tengen.r1_irq_counter_old!=0)&map_tengen.r1_irq_isenabled) { \
	map_tengen.r1_irq_counter_old=0; \
	cpu_set_interrupt(INTERRUPT_IRQ|IRQ_MAPPER|(x?0:INT_EDGE_IRQ_EX)); \
}

/* $c001.0=0 (a12) */
_MAPA12(tengen_rambo1)
{
	if (address&0x1000) {
		if (((*mapper.ppu_cycles_ptr+mapper.a12_rise_wait)<=mapper.ppu_cycles_old)&((mapper.ppu_address_old|map_tengen.r1_irq_type)==0)) { RAMBO_IRQCLOCK((*mapper.ppu_cycles_ptr-*mapper.cpu_cycles_ptr)>crystal->rel_cycle) }
		mapper.ppu_cycles_old=*mapper.ppu_cycles_ptr;
	}
	mapper.ppu_address_old=address&0x1000;
}
/* $c001.0=1 (clock every 4 cpu clockcycles(ccc)) */
_MAPIRQ(tengen_rambo1)
{
	int cpu_cycles;
	
	ppu_force_update(); /* for a12 */
	
	for (cpu_cycles=*mapper.cpu_cycles_ptr;mapper.cpu_cycles_old>cpu_cycles;mapper.cpu_cycles_old-=crystal->rel_cycle) {
		if (((map_tengen.r1_irq_ccc=(map_tengen.r1_irq_ccc+1)&3)==0)&map_tengen.r1_irq_type) { RAMBO_IRQCLOCK(mapper.cpu_cycles_old-cpu_cycles) }
	}
}

/* bankswitch */
static __inline__ void rambo1_bs(void)
{
	int i,bit;
	
	/* swap prg at: control bit 6 is 0: $8000, $a000, $c000, 1: $a000, $c000, $8000 */
	bit=map_tengen.r1_control>>6&1;
	i=3; while (i--) { PRG8_BS((i+bit)%3,map_tengen.r1_prg[i]&MAXPRG); }
	
	/* swap chr */
	bit=map_tengen.r1_control>>5&4; /* xor */
	if (map_tengen.r1_control&BIT(5)) { i=4; while (i--) { CHR1_BS_V(i^bit,map_tengen.r1_chr[i]&MAXVROM); } }
	else { i=4; while (i--) { CHR1_BS_V(i^bit,((map_tengen.r1_chr[i&2]&~1)|(i&1))&MAXVROM); } }
	for (i=4;i<8;i++) { CHR1_BS_V(i^bit,map_tengen.r1_chr[i]&MAXVROM); }
}

/* io */
_MAPIO(tengen_rambo1)
{
	switch (address&0xe001) {
		
		/* set control register */
		case 0x8000: map_tengen.r1_control=data; rambo1_bs(); break;
		
		/* set bank */
		case 0x8001: {
			int s=map_tengen.r1_control&0xf;
			switch (s) {
				case 0: case 1: map_tengen.r1_chr[s<<1]=data; break; /* 1/2k chr0/2 */
				case 2: case 3: case 4: case 5: map_tengen.r1_chr[s+2]=data; break; /* 1k chr4/5/6/7 */
				case 6: case 7: map_tengen.r1_prg[s&1]=data; break; /* 8k prg0/1 */
				case 8: case 9: map_tengen.r1_chr[(s<<1&2)|1]=data; break; /* 1k chr1/3 */
				case 0xf: map_tengen.r1_prg[2]=data; break; /* 8k prg2 */
				default: break;
			}
			rambo1_bs();
			break;
		}
		
		/* mirroring */
		case 0xa000: MIR_VH(data&1) break;
		
		/* irq, as mmc3 */
		case 0xc000: map_tengen.r1_irq_latch=data; break;
		case 0xc001: map_tengen.r1_irq_reload=TRUE; map_tengen.r1_irq_type=data&1; break; /* bit 0: select source */
		case 0xe000: map_tengen.r1_irq_isenabled=FALSE; cpu_acknowledge_interrupt(~IRQ_MAPPER); break;
		case 0xe001: map_tengen.r1_irq_isenabled=TRUE; break;
		
		default: break;
	}
}

/* init */
_MAPINIT(tengen_rambo1)
{
	int i=8; while (i--) map_tengen.r1_chr[i]=i; INIT_CHR_FIRST();
	map_tengen.r1_prg[1]=1; map_tengen.r1_prg[2]=MAXPRG-1; INIT_PRG_FIRSTLAST();
	
	INIT_MIR_V();
	SET_MAPIO8_ALL(tengen_rambo1);
	SET_MAPA12(tengen_rambo1);
	SET_MAPIRQ(tengen_rambo1); /* override mapa12 irq handler */
}





/* --- simple P32/C8(xxxxPCCC) (iNES: no) ---
	Tetris (it's only got 32k prg, so it's compatible with CNROM)
*/

_MAPIO(tengen_l1)
{
	/* bits 0,1,2: swap 8k chr */
	int bank=data<<3&BIN8(00111000)&MAXVROM;
	CHR8_BS_V(bank);
	
	/* bit 3: swap 32k prg */
	bank=data>>1&BIT(2)&MAXPRG;
	PRG32_BS(bank);
}

_MAPINIT(tengen_l1)
{
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
	SET_MAPIO8_ALL(tengen_l1);
}
