/* Unlicensed mappers */

/* (AVE mappers have their own module) */

static struct {
	/* action 52 */
	int action52_ram[4];
	
	/* camerica quattro */
	int camerica_bf9096_block;
	int camerica_bf9096_bank;
	
	/* colordreams */
	int colordreams_b0;
	
	/* rumblestation */
	int rumble_lo;
	int rumble_hi;
	
} map_unl;


/* --- Action 52 (iNES: 228) ---
	Action 52, Cheetah Men II
*/

/* $4020-$5fff: 4x4bit ram */
_MAPIO(action52_ram) { map_unl.action52_ram[address&3]=data&0xf; }
_MAPIO_R(action52_ram) { return map_unl.action52_ram[address&3]; }

/* $8000-$ffff: io */
_MAPIO(action52)
{
	/* address bits 7,8,9,10,11,12: select 32k prg */
	int bank=address>>5&BIN8(11111100)&MAXPRG;
	PRG32_BS(bank);
	
	/* address bits 0,1,2,3, data bits 0,1: select 8k chr */
	bank=((address<<5)|(data<<3&BIN8(00011000)))&MAXVROM;
	CHR8_BS_V(bank);
}

_MAPINIT(action52)
{
	SET_MAPIO8_ALL(action52);
	SET_MAPIO4(action52_ram);
	SET_MAPIO4_R(action52_ram);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
}





/* --- Camerica BF9093 (iNES: 71) ---
	Bee 52, Big Nose Freaks Out, Big Nose the Caveman, Dizzy The Adventurer, Fantastic Adventures of Dizzy,
	Linus Spacehead's Cosmic Crusade, Micro Machines, MiG 29 - Soviet Fighter, Stunt Kids, Ultimate Stuntman
	
	(plus some others that are wrongly labeled as UNROM)
	
   --- Camerica BF9097 (same as above, except for one-screen mirroring) (iNES: no/71)
	Fire Hawk
*/

_MAPIO(camerica_ce)
{
	/* swap 16k prgbank in $8000 */
	int bank=data<<1&MAXPRG;
	PRG16_BS(0,bank);
}

_MAPIO(camerica_bf9097)
{
	/* bit 3:? (sometimes set) */
	
	/* bit 4: set mirroring */
	MIR_01(data&BIT(4))
}

_MAPINIT(camerica)
{
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST();
	
	switch (maptype) {
		/* bf9093 */
		case MAPTYPE_CAMERICA_BF9093:
			SET_MAPIO8(nothing,nothing,camerica_ce,camerica_ce);
			break;
		
		/* bf9097 */
		case MAPTYPE_CAMERICA_BF9097:
			SET_MAPIO8(camerica_bf9097,nothing,camerica_ce,camerica_ce);
			INIT_MIR_0();
			break;
		
		default: break;
	}
}





/* --- Camerica BF9096 (iNES: 232) ---
   --- Camerica BF9096 Aladdin (almost the same) (newer versions of GoodNES have 'corrected' the Aladdin images) (iNES: 232/no) ---
	Quattro Adventure,
		Linus Spacehead, Super Robin Hood, Boomerang Kid, Treasure Island Dizzy
	Quattro Arcade,
		CJ's Elephant Antics, Stunt Buggies, F-16 Renegade, Go! Dizzy Go!
	Quattro Sports/Super Sports Challenge (E)
		Baseball Pro's, BMX Simulator, Soccer Simulator, Pro Tennis Simulator
*/

/* TODO:
	- Quattro Sports (U) (V3 Plug-Thru Cart) [a1] crashes, it's most likely a bad dump
*/

/* prg bankswitch */
static __inline__ void camerica_bf9096_bs(void)
{
	int prg0=(map_unl.camerica_bf9096_block|map_unl.camerica_bf9096_bank)&MAXPRG;
	int prg1=(map_unl.camerica_bf9096_block|BIN8(00000110))&MAXPRG;
	PRG16_BS(0,prg0); PRG16_BS(2,prg1);
}

/* $8000-$bfff, bits 3,4: select 64k block */
_MAPIO(camerica_bf9096_8a)
{
	map_unl.camerica_bf9096_block=data&BIN8(00011000);
	camerica_bf9096_bs();
}
_MAPIO(camerica_bf9096_aladdin)
{
	map_unl.camerica_bf9096_block=(data<<1&BIT(4))|(data>>1&BIT(3)); /* reversed databus */
	camerica_bf9096_bs();
}

/* $c000-$ffff */
_MAPIO(camerica_bf9096_ce)
{
	/* select 16k bank in 64k block for $8000-$bfff */
	map_unl.camerica_bf9096_bank=data<<1&BIN8(00000110);
	camerica_bf9096_bs();
}

/* init */
_MAPINIT(camerica_bf9096)
{
	switch (maptype) {
		/* std */
		case MAPTYPE_CAMERICA_BF9096:
			SET_MAPIO8(camerica_bf9096_8a,camerica_bf9096_8a,camerica_bf9096_ce,camerica_bf9096_ce);
			break;
		
		/* aladdin */
		case MAPTYPE_CAMERICA_BF9096_ALADDIN:
			SET_MAPIO8(camerica_bf9096_aladdin,camerica_bf9096_aladdin,camerica_bf9096_ce,camerica_bf9096_ce);
			break;
		
		default: break;
	}
	
	camerica_bf9096_bs();
	INIT_CHR_FIRST(); /* 8k vram */
}





/* --- Color Dreams(Rev A/Rev B/Rev C/AGCI 47516/196/BC6) (iNES: 11(/231?)) ---
	Crystal Mines, Menace Beach, all those funky religious games, etc.
	
   --- AGCI 50282, same as Color Dreams mapper, except for D0 awkwardness (iNES: 144) ---
	Death Race
*/

_MAPIO(colordreams)
{
	int bank;
	
	/* dataline 0 on ROM always wins for AGCI 50282 */
	data|=(bus&map_unl.colordreams_b0);
	
	/* bits 0,1: select 32k prg bank */
	bank=data<<2&MAXPRG;
	PRG32_BS(bank);
	
	/* bits 2,3: lockout chip stuff */
	
	/* high nibble: select 8k vrom bank. */
	bank=data>>1&BIN8(01111000)&MAXVROM;
	if (MAXVROM) { CHR8_BS_V(bank); }
}

_MAPINIT(colordreams)
{
	map_unl.colordreams_b0=maptype==MAPTYPE_AGCI_50282;
	
	SET_MAPIO8_ALL(colordreams);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
}





/* --- Rumble Station (legal famiclone with 15 colordreams games) (iNES: 46) ---
	Rumble Station 15-in-1
		Baby Boomer, Captain Comic, Challenge of the Dragon, Crystal Mines, Galactic Crusader,
		King Neptune's Adventure, Master Chu & The Drunkard Hu, Menace Beach, Moon Ranger,
		Operation Secret Storm, Pesterminator, Raid 2020, Robodemons, Secret Scout, The P'radikus Conflict
		
	(Galactic Crusader needs horizontal mirroring, but mirroring is hardwired on this mapper)
*/

/* bankswitch */
static __inline__ void rumblestation_bs(void)
{
	int bank=(map_unl.rumble_hi<<3|(map_unl.rumble_lo<<2&BIT(2)))&MAXPRG; PRG32_BS(bank);
	bank=((map_unl.rumble_hi<<2&0x3c0)|(map_unl.rumble_lo>>1&BIN8(00111000)))&MAXVROM; CHR8_BS_V(bank);
}

_MAPIO(rumblestation_hi) { map_unl.rumble_hi=data; rumblestation_bs(); } /* $6000-$7fff: select 64k blocks (CCCCPPPP) */
_MAPIO(rumblestation_lo) { map_unl.rumble_lo=data; rumblestation_bs(); } /* $8000-$ffff: select 32k prg/8k chr from block (similar to colordreams: (xCCCxxxP)) */

_MAPINIT(rumblestation)
{
	SET_MAPIO6(rumblestation_hi);
	SET_MAPIO8_ALL(rumblestation_lo);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
}
