#ifndef PPU_H
#define PPU_H

enum { PPU_FPT_NORMAL=0, PPU_FPT_PT0, PPU_FPT_PT1, PPU_FPT_NT, PPU_FPT_CUSTOM };
#define PPU_FPT_ALL 8

typedef __fastcall u8(*fp_ppu_read_memorymap)(u16);
typedef __fastcall void(*fp_ppu_write_memorymap)(u16,u8);
typedef __fastcall void(*fp_ppu_modechangesignal)(int);

/* write only */										/* $2000 */
/* write only */										/* $2001 */
u8 __fastcall ppu_read_status(u16);						/* $2002 */
/* write only */										/* $2003 */
u8 __fastcall ppu_read_sprite_memory_data(u16);			/* $2004 */
/* write only */										/* $2005 */
/* write only */										/* $2006 */
u8 __fastcall ppu_read_memory_data(u16);				/* $2007 */
u8 __fastcall ppu_read_openbus(u16);					/* for write-onlies */

void __fastcall ppu_write_control1(u8);					/* $2000 */
void __fastcall ppu_write_control2(u8);					/* $2001 */
/* read only */											/* $2002 */
void __fastcall ppu_write_sprite_memory_address(u8);	/* $2003 */
void __fastcall ppu_write_sprite_memory_data(u8);		/* $2004 */
void __fastcall ppu_write_scroll(u8);					/* $2005 */
void __fastcall ppu_write_memory_address(u8);			/* $2006 */
void __fastcall ppu_write_memory_data(u8);				/* $2007 */

#define PPU_WARMUP_SOFT 1
#define PPU_WARMUP_HARD 2

void __fastcall ppu_write_control1_warmup(u8);
void __fastcall ppu_write_control2_warmup(u8);
void __fastcall ppu_write_scroll_warmup(u8);
void __fastcall ppu_write_memory_address_warmup(u8);

void ppu_set_fpt (int,int,fp_ppu_read_memorymap,fp_ppu_read_memorymap,fp_ppu_read_memorymap,fp_ppu_read_memorymap,fp_ppu_write_memorymap,fp_ppu_write_memorymap,fp_ppu_write_memorymap,fp_ppu_write_memorymap);
void ppu_set_modechangesignal(fp_ppu_modechangesignal);
void ppu_set_overscan(void);
void ppu_give_screen_ptr(u16*);
void ppu_init(void);
void ppu_reset(int,int);
void ppu_warm(void);
void ppu_clean(void);
void __fastcall ppu_new_frame(void);
void __fastcall ppu_clear_nmi_enable(void);
void ppu_set_update_nothing(void);
void __fastcall ppu_force_update(void);
int __fastcall ppu_get_vblank_nmi_disallowed(void);
int __fastcall ppu_get_skipped_cycle(void);
int __fastcall ppu_get_warmup(void);
void __fastcall ppu_clear_status(void);

int* ppu_get_enabled_ptr(void);
int* ppu_get_valid_scanline_ptr(void);
u8* ppu_get_control1_ptr(void);
u8* ppu_get_control2_ptr(void);
int* ppu_get_reading_sprites_ptr(void);
int* ppu_get_rendering_ptr(void);
int* ppu_get_sc_ptr(void);
int* ppu_get_scc_ptr(void);
u16* ppu_get_address_ptr(void);
int* ppu_get_cycles_ptr(void);
int ppu_get_cycles(void);

/* memmap */
void ppu_memmap_init(void);

void ppu_memmap_printf_internal_nametable(int);

typedef __fastcall void(*fp_ppubusoverride)(u16);
void ppu_memmap_set_ppubusoverride(fp_ppubusoverride);

u8* ppu_memmap_get_palette_ptr(void);
u8** ppu_memmap_get_cur_name_ptr(void);
u8** ppu_memmap_get_name_ram_ptr(void);

/* memmap handlers */
u8 __fastcall ppu_read_pt0a(u16);
u8 __fastcall ppu_read_pt0b(u16);
u8 __fastcall ppu_read_pt0c(u16);
u8 __fastcall ppu_read_pt0d(u16);
u8 __fastcall ppu_read_pt1a(u16);
u8 __fastcall ppu_read_pt1b(u16);
u8 __fastcall ppu_read_pt1c(u16);
u8 __fastcall ppu_read_pt1d(u16);
u8 __fastcall ppu_read_name0(u16);
u8 __fastcall ppu_read_name1(u16);
u8 __fastcall ppu_read_name2(u16);
u8 __fastcall ppu_read_name3(u16);

void __fastcall ppu_write_pt0a(u16,u8);
void __fastcall ppu_write_pt0b(u16,u8);
void __fastcall ppu_write_pt0c(u16,u8);
void __fastcall ppu_write_pt0d(u16,u8);
void __fastcall ppu_write_pt1a(u16,u8);
void __fastcall ppu_write_pt1b(u16,u8);
void __fastcall ppu_write_pt1c(u16,u8);
void __fastcall ppu_write_pt1d(u16,u8);
void __fastcall ppu_write_name0(u16,u8);
void __fastcall ppu_write_name1(u16,u8);
void __fastcall ppu_write_name2(u16,u8);
void __fastcall ppu_write_name3(u16,u8);

u8 __fastcall ppu_read_palette(u16);
void __fastcall ppu_write_palette(u16,u8);

u8 __fastcall ppu_read_pt0a_override(u16);
u8 __fastcall ppu_read_pt0b_override(u16);
u8 __fastcall ppu_read_pt0c_override(u16);
u8 __fastcall ppu_read_pt0d_override(u16);
u8 __fastcall ppu_read_pt1a_override(u16);
u8 __fastcall ppu_read_pt1b_override(u16);
u8 __fastcall ppu_read_pt1c_override(u16);
u8 __fastcall ppu_read_pt1d_override(u16);
u8 __fastcall ppu_read_name0_override(u16);
u8 __fastcall ppu_read_name1_override(u16);
u8 __fastcall ppu_read_name2_override(u16);
u8 __fastcall ppu_read_name3_override(u16);

void __fastcall ppu_write_pt0a_override(u16,u8);
void __fastcall ppu_write_pt0b_override(u16,u8);
void __fastcall ppu_write_pt0c_override(u16,u8);
void __fastcall ppu_write_pt0d_override(u16,u8);
void __fastcall ppu_write_pt1a_override(u16,u8);
void __fastcall ppu_write_pt1b_override(u16,u8);
void __fastcall ppu_write_pt1c_override(u16,u8);
void __fastcall ppu_write_pt1d_override(u16,u8);
void __fastcall ppu_write_name0_override(u16,u8);
void __fastcall ppu_write_name1_override(u16,u8);
void __fastcall ppu_write_name2_override(u16,u8);
void __fastcall ppu_write_name3_override(u16,u8);

u8 __fastcall ppu_read_palette_override(u16);
void __fastcall ppu_write_palette_override(u16,u8);

#endif /* PPU_H */
